/*--------------------------------------------------------------------
 * $Id: miecalc.c,v 1.23 2007/11/06 16:53:47 ulrich Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

/* @44c@ */
/* @c44@ */

/********************************************************************************/
/* In order to use the functions provided by the mie library,          @44_10c@ */
/* #include <miecalc.h> in your source code and link with libRadtran_c.a.       */
/*                                                                              */
/* @strong{Example:}                                                            */
/* Example for a source file:                                                   */
/* @example                                                                     */
/*                                                                              */
/*   ...                                                                        */
/*   #include "../src_c/miecalc.h"                                              */
/*   ...                                                                        */
/*                                                                              */
/* @end example                                                                 */
/*                                                                              */
/* Linking of the executable, using the GNU compiler gcc:                       */
/* @example                                                                     */
/*                                                                              */
/*   gcc -o test test.c -lRadtran_c                                             */
/*                                                                              */
/* @end example                                                        @c44_10@ */
/********************************************************************************/

/****************************************************************************/
/* The Mie library provides functions for Mie calculations,       @44_20c@  */
/* interfacing the MIEV0 and BHMIE codes by Warren Wiscombe                 */
/* (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman        */
/* (ftp://astro.princeton.edu/draine/scat/bhmie/). Functions for            */
/* evaluating the phase function and the integrated phase function are      */
/* also provided.                                                 @c44_20@  */
/****************************************************************************/


#include "../src_f/f77-uscore.h"
#include "../src_c/fortran_and_c.h"
#include "./numeric.h"
#include "./miecalc.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#include <string.h>
#endif

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif



/***********************************************************************************/
/* Function: mie_calc                                                     @44_30i@ */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see src_c/miecalc.h)             */
/*  mie_out_struct *output:  mie output structure (see src_c/miecalc.h)            */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER or AEROSOL,              */
/*                           the refractive index is read from crefin              */
/*  mie_complex crefin:      Complex refractive index (both numbers positive)      */
/*  float wavelength:        Wavelength [micron]                                   */
/*  float radius:            Droplet radius [micron]                               */
/*  float temperature:       Temperature [K]                                       */
/*  mie_complex *ref:        Complex index of refraction                           */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/
int mie_calc (mie_inp_struct input, mie_out_struct *output, 
	      int program, int medium, mie_complex crefin, 
	      float wavelength, float radius, 
	      float temperature, int nstokes, mie_complex *ref)
{
  void F77_FCN (wrefice) (float *wl, float *temp, float *re, float *im); 

  void F77_FCN (wrefwat) (float *wl, float *temp, float *re, float *im); 

  void F77_FCN (bhmie) (float *xx, mie_complex *crefin, int *numang,
			mie_complex *s1, mie_complex *s2,  
			float *qext, float *qsca, float *qback, float *gsca);

  void F77_FCN (miev0) ( float *xx, mie_complex *crefin, int *perfct, float *mimcut, 
			 int *anyang, int *numang, float *xmu, int *nmom, 
			 int *ipolzn, int *momdim, int *prnt, float *qext, 
			 float *qsca, float *gqsc, float *pmom, mie_complex *sforw, 
			 mie_complex *sback, mie_complex *s1, mie_complex *s2, mie_complex *tforw, 
			 mie_complex *tback, float *spike);
  
 

  /* input.nmom+2*4, because miev0 starts counting Legendre polynomials 
     from 0 and fortran adds one element at end of string*/
  float *tmp_pmom = (float *) calloc ((input.nmom+2)*4, sizeof(float));
  float **pmom_array;
  int im=0, ip=0; 
  
  char function_name[]="mie_calc";
  char file_name[]="src_c/miecalc.c";

  pmom_array = calloc (4, sizeof(float *));
  for (ip=0; ip<4; ip++) 
    pmom_array[ip] = calloc (input.nmom+2, sizeof(float));
  
  
  if (nstokes > 1 && program !=MIEV0){
    fprintf (stderr, "Error: For nstokes > 1 only the program MIEV0 can be used.\n");
    return -1;
  }
 
  switch (medium) {
  case ICE:
    F77_FCN (wrefice) (&wavelength, &temperature, &(ref->re), &(ref->im));
    if (ref->re == 0.0 && ref->im == 0.0) {
      fprintf (stderr, "Error, during execution of 'wrefice' in %s (%s)\n", function_name, file_name);
      return -1;
    }
    break;
    
  case WATER:
    F77_FCN (wrefwat) (&wavelength, &temperature, &(ref->re), &(ref->im));
    if (ref->re == 0.0 && ref->im == 0.0) {
      fprintf (stderr, "Error, during execution of 'wrefwat' in %s (%s)\n", function_name, file_name);
      return -1;
    }
    break;

  case USER:
  case AEROSOL:
    ref->re = crefin.re;
    ref->im = crefin.im;
    break;
    
  default:
    fprintf (stderr, "Error: unknown medium\n");
    return -1;
  }
  output->crefin.re = ref->re;
  output->crefin.im = ref->im;
  output->xx        = 2.0*PI * radius / wavelength;

  switch (program) {
  case BHMIE:
    F77_FCN (bhmie) ( &output->xx, &output->crefin, &input.numang,
		      input.s1, input.s2,  
		      &output->qext, &output->qsca, &output->qback, 
		      &output->gsca );
    break;
  case MIEV0:
    
    /* Loop over number of Stokes components
       description of *ipolzn* in Mie documentation. */ 
    if (nstokes==1)
      input.ipolzn=0;
    else if (nstokes==2)
      input.ipolzn=12;
    else if (nstokes==3)
      input.ipolzn=123;
    else if (nstokes==4)
      input.ipolzn=1234;
    else{
      fprintf(stderr, "Error: Number of Stokes components (nstokes) not between 1 and 4\n");
      return -1;
    }
    
    /* This is needed to avoid segmentation fault in case miev0 is used 
       without calculating Legendre polynomials (miev0 gives segmentation
       fault if input.momdim=0.)*/
    if(input.nmom==0)
      input.momdim=1;
    
    F77_FCN (miev0) ( &output->xx, &output->crefin, &input.perfct, 
		      &input.mimcut, &input.anyang, &input.numang,  
		      input.xmu, &input.nmom, &input.ipolzn,  
		      &input.momdim, input.prnt,  
		      &output->qext, &output->qsca, &output->gqsc,  
		      tmp_pmom, &output->sforw, &output->sback,   
		      input.s1, input.s2,  
		      output->tforw, output->tback, &output->spike ); 
    
    output->gsca = output->gqsc / output->qsca;
    
    
    
    if (input.nmom > 0){
      fortran2c_2D_float_ary_noalloc (4, input.nmom+2, tmp_pmom, pmom_array);
      
      for (im=0; im<=input.nmom; im++){
        for (ip=0; ip<4; ip++){
          output->pmom[ip][im]= pmom_array[ip][im];
        }
      }
    }
    
    break;
  default:
    fprintf (stderr, "error: unknown Mie program\n");
    return -1;
  }

  free(tmp_pmom);
  
  if (input.nmom>0) {
    for (ip=0; ip<4; ip++)
      free(pmom_array[ip]);
    free(pmom_array);
  }


  return 0;

 
}


/***********************************************************************************/
/* Function: mie_calc_sizedist                                            @44_30i@ */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see src_c/miecalc.h)             */
/*  mie_out_struct *output:  mie output structure (see src_c/miecalc.h)            */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER, the refractive index    */
/*                           is read from crefin                                   */
/*  mie_complex crefin:      Complex refractive index (both numbers positive)      */
/*  float wavelength:        wavelength [micron]                                   */
/*  float temperature:       temperature                                           */
/*  double *x_size:          size distribution, radius [um]                        */
/*  double *y_size:          size distribution, n(r)                               */
/*  int n_size:              size distribution, number of radii                    */
/*  double *beta:            extinction coefficient [km-1] per unit                */
/*                           liquid water content (returned)                       */
/*  double *omega:           Single scattering albedo (returned)                   */
/*  double *g:               Asymmetry parameter (returned)                        */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int mie_calc_sizedist (mie_inp_struct input, mie_out_struct *output, 
		       int program, int medium, mie_complex crefin, 
		       float wavelength, float temperature, int nstokes,
		       double *x_size, double *y_size, int n_size,
		       double *beta, double *omega, double *g, 
		       mie_complex *ref)
{
  int status=0, is=0, ip=0, im=0;

  double xsquared=0, xcubed=0, norm=0;
  
  double ext=0, sca=0, asy=0, vol=0;

  /* allocate memory for fields */
  double *extinction = calloc (n_size, sizeof (double));
  double *scattering = calloc (n_size, sizeof (double));
  double *asymmetry  = calloc (n_size, sizeof (double));
  double *volume     = calloc (n_size, sizeof (double));

  double ***pmom=NULL;

  
  if (input.nmom>0) {
    pmom     = calloc (4, sizeof(double **));
    
    for (ip=0; ip<4; ip++) {
      pmom    [ip] = calloc (input.nmom+1, sizeof(double *));
      
      for (im=0; im<=input.nmom; im++)
	pmom[ip][im] = calloc (n_size, sizeof(double));
    }
  }


  for (is=0; is<n_size; is++) {
    
    status = mie_calc (input, output, program, medium, crefin, 
		       wavelength, x_size[is], temperature, nstokes, ref);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by mie_calc()\n", status);
      return status;
    }
    
    xsquared = x_size[is] * x_size[is] * 1e-12;
    xcubed   = xsquared   * x_size[is] * 1e-6;
    
    extinction[is] = output->qext * xsquared * y_size[is];
    scattering[is] = output->qsca * xsquared * y_size[is];
    asymmetry [is] = output->gsca * scattering[is]; 
    volume    [is] = xcubed * y_size[is];

	
    /* quote from MIEV.doc, concerning the normalization of pmom[]:   */
    /*	                                                              */
    /*	The normalized moments are                                    */
    /*      4 / ( XX**2 * QSCA ) * PMOM,  but it is PMOM itself, not  */
    /*      these normalized moments, which should be integrated over */
    /*      a size distribution.                                      */
    
    if (input.nmom>0)
      for (ip=0; ip<4; ip++)
	for (im=0; im<=input.nmom; im++)
	      pmom[ip][im][is] = output->pmom[ip][im] * y_size[is];
  }
  
  norm = integrate (x_size, y_size,     n_size);
  
  ext  = integrate (x_size, extinction, n_size) / norm;
  sca  = integrate (x_size, scattering, n_size) / norm;
  asy  = integrate (x_size, asymmetry,  n_size) / norm;
  vol  = integrate (x_size, volume,     n_size) / norm;
  
  if (input.nmom>0)
    for (ip=0; ip<4; ip++)
      for (im=0; im<=input.nmom; im++)
	output->pmom[ip][im] = integrate (x_size, pmom[ip][im], n_size) / norm;
  

  *beta  = 3.0*ext/(4.0*vol)/1000.0;
  *omega = sca/ext;
  *g     = asy/sca;



  /* free memory */
  free (extinction);
  free (scattering);
  free (asymmetry);
  free (volume);
  
  if (input.nmom>0) {
    for (ip=0; ip<4; ip++) {
      for (im=0; im<=input.nmom; im++)
	free(pmom[ip][im]);
      
      free(pmom[ip]);
    }
    
    free(pmom);
  }


  return 0;
}




/************************************************************************************/
/* Function: phase_function                                               @44_30i@  */
/* Description:                                                                     */
/*  Calculate the phase function from its moments and output to a file "phase.dat". */
/*  The phase function                                                              */
/*    @iftex                                                                        */
/*    @tex                                                                          */
/*     $p(\mu)$ is @*                                                               */
/*      $$ p (\mu) = \sum_{m=0}^{\infty} (2m+1) \cdot k_m \cdot P_m (\mu) $$        */
/*      where $k_m$ is the m'th moment and $P_m (\mu)$ is the m'th                  */
/*      Legendre polynomial.                                                        */
/*     @end tex                                                                     */
/*    @end iftex                                                                    */
/*    @ifinfo                                                                       */
/*     p(mu) is @*                                                                  */
/*     p (mu) = sum (m=0 to infinity) (2m+1) * k(m) * Pm (mu) @*                    */
/*     where k(m) is the m'th moment and Pm (mu) is the m'th Legendre polynomial.   */
/*    @end ifinfo                                                                   */
/*                                                                                  */
/* Parameters:                                                                      */
/*  float *moment: moments of the phase function, f[0, ..., L-1]                    */
/*  int L:         number of moments                                                */
/*                                                                                  */
/* Return value:                                                                    */
/*  0  if o.k., <0 if error                                                         */
/*                                                                                  */
/* Example:                                                                         */
/* Files:                                                                           */
/* Known bugs:                                                                      */
/*  Syntax and parameters of this function are subject to change.                   */
/* Author:                                                                          */
/*                                                                        @i44_30@  */
/************************************************************************************/

int phase_function (float *moment, int L)
{
  int l=0, it=0;
  double p0=0, p1=0, p2=0;
  double mu=0, dmu=0;
  double sum=0, sum2=0, factor=0;
  int Nt=0;
  FILE *outfile=NULL;

  if ((outfile = fopen("phase.dat", "w"))==NULL) {
    fprintf (stderr, "Error opening phase.dat for writing\n");
    return -1;
  }

  Nt  = 100001;
  dmu = 2.0 / (double) (Nt-1); 
  
  factor = 1.0 / 4.0 / 3.141692653590 / moment[0];

  for (it=0; it<Nt; it++) {

    mu = -1.0 + (double) it * dmu;
    
    p0=1;
    p1=mu;
    

    sum  = moment[0]*p0 + 3.0*moment[1]*p1;

    for (l=2; l<=L; l++) {
      p2 = (double) (2*l-1) / (double) l * mu * p1 - 
	(double) (l-1) / (double) l * p0;
      
      sum += (double) (2*l+1) * moment[l] * p2;
      p0=p1;
      p1=p2;
    }

    fprintf (outfile, "%g  %g  %g\n", mu, sum*factor, sum2*factor);
  }

  fclose (outfile);
  return 0;
}



/***********************************************************************************/
/* Function: read_mie_table                                               @44_30i@ */
/* Description:                                                                    */
/*  Read Frank Evans' Mie table, similar to his subroutine READ_MIE_TABLE          */
/*  provided in plotmietab.f; the Mie table can be either a single-column file     */
/*  with the moments of the phase function, or a file complying with Frank Evans'  */
/*  'plotmietab' routine, containing moments for different droplet sizes.          */
/*  IMPORTANT: First, read_mie_table() looks for a file 'filename'.cdf which,      */
/*  when available, is interpreted as the netCDF version of the file.              */
/*  For the netCDF format of the Mie table file, have a look at                    */
/*  tools/cloudprp2cdf.sh which converts the ASCII to the netCDF version.          */
/*  ATTENTION: Frank Evans' stores Legendre coefficients f(l), not moments p(l):   */
/*             f = p * (2*l + 1)                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  char *filename:   filename where data is stored                                */
/*  double *r0:       smallest effective radius found in filename                  */
/*  double *dr:       effective radius step                                        */
/*  int *nreff:       number of effective radii                                    */
/*  double *wavelen:  wavelength [nm]                                              */
/*  double *nre:      real part of the refractive index                            */
/*  double *nim:      imaginary part of the refractive index                       */
/*  double **extinc:  array[0 ... nreff-1] of extinction coefficients              */
/*  double **albedo:  array[0 ... nreff-1] of single scattering albedos            */
/*  int **nleg:       array[0 ... nreff-1] storing the number of Legendre coeffc.  */
/*  float ***legen:   array[0 ... nreff-1][0...ip][0 ... nleg]                     */
/*                                                     of Legendre coefficients    */
/*  int quiet:        'Shut up!' flag                                              */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int read_mie_table (char *filename, 
		    double *r0, double *dr, size_t *nreff, 
		    double *wavelen, double *nre, double *nim,
		    double **extinc, double **albedo, int **nleg, 
		    float ****legen, size_t *nphase, int quiet)
{
  int i=0, ip=0, im=0, status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0, temp=0;
  
  double reffmin=0, reffmax=0;

  double *reff=NULL;

  char dummy[1024];
  
  FILE *f=NULL;
  
#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_reff=0, idd_maxmom=0, idd_nphase=0;
  int id_reff=0, id_ext=0, id_ssa=0, id_nmom=0, id_pmom=0;
  int id_wavelen=0; 
  int new_format=1; 

  size_t maxmom = 0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t start3D[3] = {0,0,0}; /*reff, nstokes, nmom*/
  size_t start2D[2] = {0,0}; /*reff, nstokes, nmom*/
 
  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (cdf_filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    if (!quiet) {
      fprintf (stderr, "*** Cannot open netCDF file %s, using\n", cdf_filename);
      fprintf (stderr, "*** ASCII file %s instead.\n", filename);
      fprintf (stderr, "*** It is recommended to use the netCDF format because access is much faster.\n");
    }
  }
  
  else {

    /* read file */

    /* get dimension id for "maxmom" */
    status = nc_inq_dimid (ncid, "maxmom", &idd_maxmom);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading maxmom id from %s\n", status, cdf_filename);
      return status;
    }
    
    /* get dimension length for "maxmom" */
    status = nc_inq_dimlen (ncid, idd_maxmom, &maxmom);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading maxmom from %s\n", status, cdf_filename);
      return status;
    }

    /* get dimension id for "reff" */
    status = nc_inq_dimid (ncid, "reff", &idd_reff);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading reff id from %s\n", status, cdf_filename);
      return status;
    }
    
    /* get dimension length for "reff" */
    status = nc_inq_dimlen (ncid, idd_reff, nreff);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading reff from %s\n", status, cdf_filename);
      return status;
    }
    
    /* get dimension id for "nphase" */
    status = nc_inq_dimid (ncid, "nphase", &idd_nphase);
    if (status!=NC_NOERR) {
      if(!quiet){
      fprintf (stderr, "Warning: nphase not found in netcdf file %s. Assume old \n",cdf_filename);
      fprintf (stderr, "         dataformat, where only phase function moments are included. \n");
      }
      *nphase=1;
      new_format=0;
    }
    
    if(new_format){
      /* get dimension length for "nphase" */
      status = nc_inq_dimlen (ncid, idd_nphase, nphase);
      if (status!=NC_NOERR) {
        fprintf (stderr, "Error %d reading nphase from %s\n", status, cdf_filename);
        return status;
      }
    }
    
    /* allocate memory for fields */
    reff    = calloc (*nreff, sizeof(double));
    *extinc = calloc (*nreff, sizeof(double));
    *albedo = calloc (*nreff, sizeof(double));
    *nleg   = calloc (*nreff, sizeof(int));
    
    count[0] = *nreff;
    
    /* get variable id for "reff" */
    status = nc_inq_varid (ncid, "reff", &id_reff);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* get variable id for "ext" */
    status = nc_inq_varid (ncid, "ext", &id_ext);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }


    /* get variable id for "ssa" */
    status = nc_inq_varid (ncid, "ssa", &id_ssa);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

  
    /* get variable id for "nmom" */
    status = nc_inq_varid (ncid, "nmom", &id_nmom);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

  
    /* get variable id for "pmom" */
    status = nc_inq_varid (ncid, "pmom", &id_pmom);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

  
    /* get variable id for "wavelen" */
    status = nc_inq_varid (ncid, "wavelen", &id_wavelen);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* (CE) The refractive index is not used in uvspec. Commented this
       because for aerosols the refractive index varies with 
       relative humidity. */
    /* get variable id for "nre" */
    /*status = nc_inq_varid (ncid, "nre", &id_nre);
      if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
      }
    */

    /* get variable id for "nim" */
    /*status = nc_inq_varid (ncid, "nim", &id_nim);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */
    
    /*(CE: reffmin and reffmax can be obtained from reff */
    /* get variable id for "reffmin" */
    /*
      status = nc_inq_varid (ncid, "reffmin", &id_reffmin);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */

    /* get variable id for "reffmax" */
    /*status = nc_inq_varid (ncid, "reffmax", &id_reffmax);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */
    
    /* read "wavelen" */
    status = nc_get_var1_double (ncid, id_wavelen, start, wavelen);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* convert wavelength from micron to nm */
    *wavelen*=1000.0;

    
    /* read "nre" */
    /*status = nc_get_var1_double (ncid, id_nre, start, nre);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */
    
    /* read "nim" */
    /*status = nc_get_var1_double (ncid, id_nim, start, nim);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */
    double not_defined=-999.;
    /* not used in uvspec ????*/
    nre = &not_defined;
    nim = &not_defined;

    /* read "reffmin" */
    /*status = nc_get_var1_double (ncid, id_reffmin, start, &reffmin);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */
    
    /* read "reffmax" */
    /* status = nc_get_var1_double (ncid, id_reffmax, start, &reffmax);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    */

    /* read "reff" */
    status = nc_get_vara_double (ncid, id_reff, start, count, reff);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    
    reffmin=reff[0];
    reffmax=reff[*nreff-1];

    /* read "ext" */
    status = nc_get_vara_double (ncid, id_ext, start, count, *extinc);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* read "ssa" */
    status = nc_get_vara_double (ncid, id_ssa, start, count, *albedo);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* read "nmom" */
    status = nc_get_vara_int (ncid, id_nmom, start, count, *nleg);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    } 
    
    status = ASCII_calloc_float_3D(legen, *nreff, *nphase, maxmom);
    if (status!=0) {
      fprintf (stderr, "Error allocating variable legen in miecalc.c.s\n");
      return status;
    } 
    
    if(new_format){
     
      for (i=0; i<*nreff; i++){
        for (ip=0; ip<*nphase; ip++){
          for (im=0; im<(*nleg)[i]; im++){
            /* allocate and read Legendre polynomials */
            start3D[0]=i; 
            start3D[1]=ip;
            start3D[2]=im; 
            status = nc_get_var1_float (ncid, id_pmom, start3D, 
                                        &((*legen)[i][ip][im]));
            if (status != 0) {
            fprintf (stderr, "Error %d reading Legendre polynoms from netCDF file %s \n", status, filename);
            return status;
            }
            (*legen)[i][ip][im] /= (double) (2*im + 1);
            /*fprintf(stderr, "(*legen)[i][ip][im] %d %d %d  %g \n",i, ip, im, (*legen)[i][ip][im]); */
          }
        }
      }
    }
    
    else{
      for (i=0; i<*nreff; i++){
        for (im=0; im<(*nleg)[i]; im++){
          /* allocate and read Legendre polynomials */
          start2D[0]=i; 
          start2D[1]=im;
          status = nc_get_var1_float (ncid, id_pmom, start2D, 
                                      &((*legen)[i][0][im]));
          if (status != 0) {
            fprintf (stderr, "Error %d reading Legendre polynoms from netCDF file %s \n", status, filename);
            return status;
          }
          (*legen)[i][0][im] /= (double) (2*im + 1);
        }
      }
    } 
    
    nc_close(ncid);
    
    /* Check number of phase matrix elements. */
    if (*nphase!=1 && *nphase!=6){
      fprintf(stderr, "error, number of phase matrix elements in %s must be either 1 or 6\n", filename);
      fprintf(stderr, "       but the file includes the polynomials for %d phase matrix elements. \n", *nphase);
      return -1; 
    }
    
    /* check if reff[i] is an equidistant grid */
    if (*nreff>1) {
      *r0 = reff[0];
      *dr = reff[1] - reff[0];
      for (i=1; i<*nreff; i++)
	if ((float) (reff[i-1] + *dr) != (float) reff[i]) {
	  fprintf (stderr, "error, effective radii in %s are not equidistant\n", filename);
	  fprintf (stderr, "reff[1] - reff[0] = %g, reff[i]-reff[i-1] = %g\n",
		   *dr, reff[i] - reff[i-1]);
	  return -1;
	}
    }
    else {
      *r0 = reff[0];
      *dr = 0.0;
    }
    
    
    /* free memory */
    free(reff);
    
    return 0;
  }

#endif

  /* check if the moments file is a 1-column file */
  status = ASCII_checkfile (filename, 
			    &rows, &min_columns, &max_columns, &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  if (min_columns==1 && max_columns==1) {

    *wavelen = -999;  
    
    *nre   = -999;
    *nim   = -999;
    *nreff = 1;

    /* allocate memory for fields */
    *extinc = calloc ((size_t) *nreff, sizeof (double));
    *albedo = calloc ((size_t) *nreff, sizeof (double));
    *nleg   = calloc ((size_t) *nreff, sizeof (int));

    (*nleg)[0] = rows;

    *r0 = -999;
    *dr = 0;

    *legen  = calloc ((size_t) *nreff, sizeof (float *));
    (*legen)[0] = calloc (1, sizeof (float *));
    /* ??? do we really need the following ??? */
    (*legen)[0][0] = calloc ((size_t) (*nleg)[0], sizeof (float));
 
    
    fprintf (stderr, " ... reading one column of data from %s\n", filename);
    
    status = read_1c_file_float (filename, &((*legen)[0][0]), &temp);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    /* final check, just to make sure ... */
    if (temp!=rows) {
      fprintf (stderr, "FATAL error reading %s:\n", filename);
      fprintf (stderr, "  found %d rows earlier, %d rows now!\n", rows, temp);
      return status;
    }

    fprintf (stderr, " ... read %d coefficients from %s\n", rows, filename);
  }
  else {
    
    if ((f = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Error opening %s for reading\n", filename);
      return -1;
    }
    
    (void) fgets (dummy, 1024, f);           /* first line ignored          */

    (void) fscanf (f, "%lf", wavelen);       /* 2nd line, wavelength        */
    (void) fgets (dummy, 1024, f);           /* ignore rest of line         */

    (void) fscanf (f, "%lf %lf", nre, nim);  /* 3rd line, refractive index  */
    (void) fgets (dummy, 1024, f);           /* ignore rest of line         */
    
    (void) fgets (dummy, 1024, f);           /* 4th line, distflag, ignored */
    
    (void) fgets (dummy, 1024, f);           /* 5th line, alpha, ignored    */
    
    /* 6th line, number of radii, max and min radius */
    (void) fscanf (f, "%d %lf %lf", nreff, &reffmin, &reffmax);  
    (void) fgets (dummy, 1024, f);           /* ignore rest of line         */
    
    
    /* convert wavelength from micron to nm */
    *wavelen*=1000.0;
    
    /* allocate memory for fields */
    reff   = calloc (*nreff, sizeof (double));
    *extinc = calloc (*nreff, sizeof (double));
    *albedo = calloc (*nreff, sizeof (double));
    *nleg   = calloc (*nreff, sizeof (int));

    *legen = calloc ((size_t) *nreff, sizeof(float **));
    
    for (i=0; i<*nreff; i++) {
      (void) fscanf (f, "%lf %lf %lf %d %d", 
		     &(reff[i]), &((*extinc)[i]), &((*albedo)[i]), &((*nleg)[i]), 
                     nphase );
      (void) fgets (dummy, 1024, f);    /* ignore rest of line */
      
   
      
      /* If nphase is neither 1 of 6, it is probably the old dataformat and only
         the phase function polynomials are read.*/
      if (*nphase!=1 && *nphase!=6)
        *nphase=1;
      
      (*nleg)[i]++;
         
      /* allocate array for legendre polynomials*/
      (*legen)[i] = calloc (*nphase, sizeof(float *));
      for(ip=0; ip<*nphase; ip++){
        (*legen)[i][ip] = calloc ((*nleg)[i], sizeof(float));
      }
      
      for (ip=0; ip<*nphase; ip++){
        for (im=0; im<(*nleg)[i]; im++) {
          (void) fscanf (f, "%f", &((*legen)[i][ip][im]));
          (*legen)[i][ip][im] /= (double) (2*im + 1);
        }
      }
    }
    
    /* check if reff[i] is an equidistant grid */
    if (*nreff>1) {
      *r0 = reff[0];
      *dr = reff[1] - reff[0];
      for (i=1; i<*nreff; i++)
	if ((float) (reff[i-1] + *dr) != (float) reff[i]) {
	  fprintf (stderr, "error, effective radii in %s are not equidistant\n", filename);
	  fprintf (stderr, "reff[1] - reff[0] = %g, reff[i]-reff[i-1] = %g\n",
		   *dr, reff[i] - reff[i-1]);
	  return -1;
	}
    }
    else {
      *r0 = reff[0];
      *dr = 0.0;
    }

    /* free memory */
    free(reff);
  }

  return 0;
}


/***********************************************************************************/
/* Function: read_mie_table_lambda                                        @44_30i@ */
/* Description:                                                                    */
/*  Similar to read_mie_table, but reads only wavelength                           */
/*                                                                                 */
/* Parameters:                                                                     */
/*  char *filename:   filename where data is stored                                */
/*  double *wavelen:  wavelength [nm]                                              */
/*  int quiet:        'Shut up!' flag                                              */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int read_mie_table_lambda (char *filename, 
			   double *wavelen, int quiet)
{
  int status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;

  char dummy[1024];
  
  FILE *f=NULL;

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int id_wavelen=0;

  size_t start[1] = {0};

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");



  /* open netcdf file */
  status = nc_open (cdf_filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    if (!quiet) {
      fprintf (stderr, "*** Cannot open netCDF file %s, using\n", cdf_filename);
      fprintf (stderr, "*** ASCII file %s instead.\n", filename);
      fprintf (stderr, "*** It is recommended to use the netCDF format because access is much faster.\n");
    }
  }
  else {

    /* read wavelength from file */

    /* get variable id for "wavelen" */
    status = nc_inq_varid (ncid, "wavelen", &id_wavelen);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* read "wavelen" */
    status = nc_get_var1_double (ncid, id_wavelen, start, wavelen);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* convert wavelength from micron to nm */
    *wavelen*=1000.0;

    nc_close(ncid);

    return 0;
  }

#endif

  /* check if the moments file is a 1-column file */
  status = ASCII_checkfile (filename, 
			    &rows, &min_columns, &max_columns, &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  if (min_columns==1 && max_columns==1) {
    *wavelen = -999;  
    fprintf (stderr, " ... one-column Mie file, cannot determine wavelength from %s\n", filename);
    return -1;
  }
  else {

    if ((f = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Error opening %s for reading\n", filename);
      return -1;
    }
    
    (void) fgets (dummy, 1024, f);           /* first line ignored          */

    (void) fscanf (f, "%lf", wavelen);       /* 2nd line, wavelength        */
    (void) fgets (dummy, 1024, f);           /* ignore rest of line         */

    /* convert wavelength from micron to nm */
    *wavelen*=1000.0;
  }

  return 0;
}


