/************************************************************************/
/* mystic.h                                                             */
/*                                                                      */
/* MYSTIC - Monte Carlo code for the physically correct tracing of      */
/*          photons in cloudy atmospheres.                              */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*         DLR, bernhard.mayer@dlr.de                                   */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 2000 Bernhard Mayer                                    */
/************************************************************************/

#ifndef __mystic_h
#define __mystic_h

#if defined (__cplusplus)
extern "C" {
#endif

#include <stdio.h>  /* required for FILENAME_MAX */
#include "phasetable.h"


/* radiation source */
#define MCSRC_NONE                 0
#define MCSRC_SOLAR                1
#define MCSRC_ISOTROPIC_SURFACE    2
#define MCSRC_ISOTROPIC_TOP        3
#define MCSRC_THERMAL_SURFACE      4
#define MCSRC_THERMAL_ATMOSPHERE   5
#define MCSRC_THERMAL_BACKWARD     6/*TZ bt*/

/* albedo method */
#define MCALB_NONE       0
#define MCALB_LAM        1
#define MCALB_LAM2D      2
#define MCALB_RPV        3
#define MCALB_RPV2D      4
#define MCALB_COXANDMUNK 5
#define MCALB_AMBRALS2D  6


/* boundary conditions */
#define MCBCOND_NONE       0
#define MCBCOND_PERIODIC   1
#define MCBCOND_MIRROR     2
#define MCBCOND_ABSORB     3

/* calculate 3D absorption field? */
#define MCABS_NONE        0    /* no                                */
#define MCABS_ABSORPTION  1    /* calculate absorption              */
#define MCABS_EMISSION    2    /* calculate emission                */
#define MCABS_HEATING     3    /* calculate (absorption - emission) */
#define MCABS_ACTINIC     4    /* calculate actinic flux            */

#define MCBACKWARD_NONE       0 /* forward calculation           */
#define MCBACKWARD_RADIANCE   1 /* backward radiance             */
#define MCBACKWARD_EDIR       2 /* direct irradiance             */
#define MCBACKWARD_EDN        3 /* downward diffuse irradiance   */
#define MCBACKWARD_EUP        4 /* upward diffuse irradiance     */ 
#define MCBACKWARD_ABS        5 /* absorption                    */
#define MCBACKWARD_ACT        6 /* actinic flux                  */


/* flag what is defined in the wcloud3D.dat, 1,2,3 */
#define CLD_OPTPROP           1 /* explicit optical properties: ext, asy, ssa */
#define CLD_EXTREFF           2 /* extinction + effective radius              */
#define CLD_LWCREFF           3 /* liquid water content + effective radius    */


/* azimuth conventions */
#define MCAZIMUTH_NEW 0
#define MCAZIMUTH_OLD 1

/* small number for floating point comparisons */
#define MC_EPSILON 1E-6

typedef enum {NONE, UPPER, LOWER} BOUNDARY;

typedef struct {
  double a;
  double b;
  double c;
  double d;
  double upper;
} surface;


typedef struct {
  int n;
  double *air;
  double *aer;
  double *ozo;
  double *wc;
  double *ic;
  double *tot;
} profile;


/* profile3D is required to store the profiles of  */
/* absorption and scattering coefficients; float   */
/* should be accurate enough                       */

typedef struct {
  int Nz;
  int Nx;
  int Ny;
  int *threed;     /* flag if a layer is 3D */
  int nthreed;     /* number of 3D layers   */
  float ***wc;
  float ***ic;
  float ***tot;
} profile3D;


/* ddprofile3D is required to store the profile of */
/* the absorbed energy; as all output quantities,  */
/* this parameter is also stored in double rather  */
/* than float                                      */

typedef struct {
  int Nz;
  int Nx;
  int Ny;
  int *threed;
  int nthreed;
  /* double ***aer; */
  /* double ***ozo; */
  /* double ***ice; */
  double ***cld;
  double ***tot;
} ddprofile3D;


typedef struct {
  double air;
  double aer;
  double ozo;
  double cld;
  double ice;
  double tot;
} optical_depth;


typedef struct {
  int    **ndir;    /* number of photons     */
  int    **ndn;
  int    **nup;

  double **edir;     /* irradiance           */
  double **edn; 
  double **eup;  

  double **fdir;     /* actinic flux         */
  double **fdn; 
  double **fup;  

  double ***raddir;  /* radiance             */
  double ***raddif;
  double ***radesc;

  /* introduced for I3RC case7: radiance binned in radial and pathlength bins */
  double ***radpat; /* cone sampling  */
  double ***radpes; /* local estimate */

  /* standard deviations */
  int std_alloc;     /* flag if allocated */
  double **edir2;    /* irradiance        */
  double **edn2; 
  double **eup2;  
  
  double **fdir2;    /* actinic flux      */
  double **fdn2; 
  double **fup2;  

  double ***raddir2; /* radiance          */
  double ***raddif2;
  double ***radesc2;
  double ***radpat2; 
  double ***radpes2; 

  int ***nraddir;    /* number of photons */
  int ***nraddif;    /* (radiance)        */
  int ***nradesc;
  int ***nradpat;
  int ***nradpes;
} radiation_field;


typedef struct {
  radiation_field *surf;
  radiation_field **alt;
  double **back;
  double **back2;
  ddprofile3D *absorption3D;    /* average            */
  ddprofile3D *absorption3D2;   /* standard deviation */
} result_struct;


typedef struct {
  double dx[3];
  int hit[3];
  double cotheta; 
} direction;


/* Structure to store a direction and opening angle */
/* for radiance calculations; convention:           */ 
/*    phi=0,   South                                */
/*    phi=90,  West                                 */
/*    phi=180, North                                */
/*    phi=270, East                                 */
/*                                                  */
/*    theta > 0 upward                              */
/*    theta < 0 downward                            */

typedef struct {
  double theta;         /* polar angle                      */
  double phi;           /* azimuth, internal                */
  double externalphi;   /* azimuth, as defined by the user  */
  double alpha; 
  double cosalpha; 
  double omega; 
  direction dir;
  direction dirsur;     /* direction in surface coordinates */ 
} radang;


/* structure to store the sum of photon weights for an individual photon */
struct tnode {
  double *loc;	  /* pointer to storage    */
  double weight;  /* sum of photon weights */
  struct tnode *left;       /* left child  */
  struct tnode *right;      /* right child */
};


/* linked list element to store a photon path */
struct photon_path {
  double x[3];
  struct photon_path *next;
};



typedef struct {
  double x[3];          /* location                                 */
  direction dir;        /* direction                                */
  direction dirsur;     /* direction, in surface coordinates        */
  int direct;           /* direct or diffuse                        */
  int scattercounter;   /* number of scattering + reflection events */ 

  double pathlength;    /* photon pathlength                      */
  optical_depth tauabs; /* absortion optical depth along the path */

  BOUNDARY boundary;    /* flag: hit lower 1D boundary yes/no  */
  int boundary2D;       /* flag: hit lower 3D boundary yes/no  */
  int crossingboundary; /* flag: left current box yes/no       */
  int outofdomain;      /* flag: left domain yes/no            */
  int reallyhit[3];     /* flag: crossed box boundary          */
  int absorption_flag;  /* TZ bt flag: real absorption for backward thermal yes/no */    
  int ic;               /* box index, x */
  int jc;               /* box index, y */
  int kc;               /* box index, z */

  double *pdir;         /* probability for direction radang[]  */

  int ipa;              /* independent pixel approximation     */

  struct tnode *wtree;  /* photon weight tree                  */
  double weight;

  struct photon_path *path;         /* photon path linked list */
} photon_struct;


typedef struct {
  double k;
  double rho0;
  double theta;
} rpv_struct;


typedef struct {
  double iso;
  double vol;
  double geo;
} ambrals_struct;


typedef struct {
  int method;
  char filename[FILENAME_MAX];

  double **albedo2D;
  double albedo;
  double delX;
  double delY;
  int Nx;
  int Ny;
  double Wsurf;

  int reflectalways;

  char rpv_filename [FILENAME_MAX];
  rpv_struct *rpv;   /* rpv structure                              */
  int *rpv_count;    /* number of occurrence of each RPV index     */
  char **rpv_index;  /* xy-field of RPV indices                    */

  /* Cox and Munk */
  float u10;
  float pcl;
  float xsal;

  char ambrals_filename [FILENAME_MAX];
  ambrals_struct **ambrals;  /* xy-field of Ambrals BRDF structure */

} albedo_struct;


typedef struct {
  int elev2D;
  char filename[FILENAME_MAX];

  surface **surf;
  double surfmax;
  double delX;
  double delY;
  int Nx;
  int Ny;

  int surfaceparallel;
} elevation_struct;

typedef struct {
  double delX;
  double delY;
  double *X;
  double *Y;
  double *Z;
  int Nx;
  int Ny;
  int Nz;
  double xmax;
  double ymax;
  int *threed;
  int nthreed;

  profile *dtsca;
  profile *dtabs;
  profile *ksca;
  profile *kscao;     /*TZ ds*/
  profile *kabs;
  profile *g1;
  profile *g2;
  profile *ff;
  profile *lwc;
  profile *reff;
  profile *Bplanck;
 
  profile3D *ksca3D;
  profile3D *ksca3Do; /*TZ ds*/
  profile3D *kabs3D;

  profile3D *g1cld3D;
  profile3D *g2cld3D;
  profile3D *ffcld3D;

  profile3D *reff3D;
  
  double rayleigh_depol; /* Rayleigh depolarisation */

  double rmin;           /* minimum effective radius in 1D/3D cloud structure */
  double rmax;           /* maximum effective radius in 1D/3D cloud structure */
  int nonHGcld;          /* Henyey-Greenstein or phase function table */

  optical_depth tabstot;
  optical_depth tscatot;

  double Watm;
  double maxemis;
  double delta_scale_mucut;/* delta scaling TZ ds*/
} atmosphere_struct;


typedef struct {
  int *sample;
  int nzout;
  radang *rad;
  int Nd;         /* number of angular directions */
  int Nr;         /* number of radii              */
  int Nt;         /* number of time intervals     */
  float dr;       /* radius interval              */
  float dt;       /* time interval                */
  double delX;
  double delY;
  int Nx;
  int Ny;
  int surface;
  int escape;
  int passback3D;
  int std;        /* calculate standard deviation */
  double delta_scale_mucut;/* TZ ds*/

  int spherical;

  int bcond;     /* boundary conditions */

  int backward; 
  int backward_is;
  int backward_js;

  int backward_islower;
  int backward_jslower;
  int backward_isupper;
  int backward_jsupper;

  int backward_writeallpixels;

  double forward_vza;
  double forward_phi;
  double original_sza;
  double original_phi0;

  double zstart;              /* photon start poistion */

  char backward_altstr[255];  /* store the altitude level for the output filename */
} sample_struct;

/* 3d cloud structure; included in uvspec cld_out_struct; */
/* used by MYSTIC and IPA                                 */
typedef struct {
  int Nx;             /* number of grid points in x-direction */
  int Ny;             /* number of grid points in y-direction */
  int nlyr;           /* number of layers                     */
  
  double delX;        /* grid size in x-direction    [m] */
  double delY;        /* grid size in y-direction    [m] */
  
  double *zd;         /* level altitudes, 0 ... nlyr [m] */

  float ***lwc;      /* liquid water content    [g/m3]   */
  float ***reff;     /* effective radius        [micron] */

  float ***ext;      /* extinction coefficient  [1/m]       */
  float ***ssa;      /* single scattering albedo            */
  float ***g1;       /* asymmetry parameter 1 for double-HG */
  float ***g2;       /* asymmetry parameter 2 for double-HG */
  float ***ff;       /* forward fraction for  double-HG     */
  float ***Dscale;   /* delta-scaling factor                */ 

  float reffmin;     /* minimum effective radius in cloud   */
  float reffmax;     /* maximum effective radius in cloud   */

  phase_function_table phase_cld;  /* complete single scattering properties: */
                                   /* tables of phase function and           */
                                   /* cumulative phase function              */            

  int cldproperties; /* as defined in the wcloud3D header                  */
  int nonHGcld;      /* flag: Henyey-Greenstein or explicit phase function */ 
  
  int *threed;       /* flag: 3D layer or 1D layer */
  int nthreed;       /* number of 3D layers        */

} cld3d_out_struct;


/* prototypes */
int mystic (int *nlyr,
	    float *dtair, float *dto3, 
	    float *dtaer, float *omaer, float *g1aer, float *g2aer, float *faer,
	    float *dtcld, float *omcld, float *g1cld, float *g2cld, float *fcld,
	    float *dtice, float *omice, float *g1ice, float *g2ice, float *fice,
	    float *refind,
	    float *rayleigh_depol,
	    cld3d_out_struct *wcld3d,
	    cld3d_out_struct *icld3d,
	    float *recld, float *reice,
	    float *temper, 
	    float *zprof,
	    float *btemp,
	    float **momaer, int *nmomaer,
	    float *alb, float *sza, float *phi0, int nphotons,
	    int *source, float *wvnmlo, float *wvnmhi, float *wavelength, 
	    float *zout, int *nzout,
	    float *rpv_rho0, float *rpv_k, float *rpv_theta, 
	    float *u10, float *pcl, float *xsal,
	    int *ipa, int *absorption, int *loaddata,
	    sample_struct *sample,
	    char *basename, 
	    char *albedofilename,
	    char *rpvfilename, char **rpv_labels, int *rpv_nlabels,
	    char *ambralsfilename,
	    char *elevationfilename,
	    float *delta_scaling_mucut,/*TZ ds*/
	    float *truncate,
	    int *reflectalways,
	    int surfaceparallel,
	    int *quiet,
	    float *rfldir, float *rfldn,  float *flup,
	    float *uavgso, float *uavgdn, float *uavgup,
	    float *rfldir3d, float *rfldn3d, float *flup3d, 
	    float *uavgso3d, float *uavgdn3d, float *uavgup3d, 
	    float *radiance3d,
	    float ***abs3d, int nxcld, int nycld, int nzcld,
	    char *datapath, int write_files);

  
double sc_Isotropic_phi();
double sc_Rayleigh_theta();
double sc_Rayleigh_theta_depol (double depol);
double sc_Isotropic_theta();
double sc_Lambertian_theta();
double sc_HG_theta(double g);
double sc_HG2_theta(double g1, double g2, double ff);
double random_tau();
  

double rpv_brdf (double rho0, double k, double theta, 
		 double mu1, double mu2, double phi);

double rpv_albedo (double rho0, double k, double theta, double mu1);

int random_rpv_direction (double rho0, double k, double theta, double mu1,
                            double *mu2, double *phi);
  
 
/* CE: Moved these functions here because they are used in sofi.c */
photon_struct *calloc_photon (sample_struct *sample);
void init_direction (double sinsza, double cossza, double sinphi,
                     double cosphi, 
                     direction *dir);

int random_ambrals_direction (double iso, double vol, double geo, 
			      double mu1, double *mu2, double *phi);



#if defined (__cplusplus)
}
#endif

#endif
