/*--------------------------------------------------------------------
 * $Id: phasetable.c,v 1.1 2007/07/06 23:01:07 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

/* @45c@ */
/* @c45@ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "phasetable.h"
#include "numeric.h"


#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* number of entries in the cumulative phase function lookup table */
#define NPHASE 20000


/* prototypes of internal functions */
static void free_pft (pft *p);


/***********************************************************************************/
/* Function: setup_Legendre_table                                         @45_30i@ */
/* Description:                                                                    */
/*  Calculate lookup-tables of phase function and cumulative phase function        */
/*  for a given wavelength index iv from the single scattering properties         */
/*  stored in ssprop                                                               */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

int setup_Legendre_table (phase_function_table *phase,
			  cld_ssprop_struct *ssprop,
			  int iv,
			  double rmin, double rmax, 
			  double delta_scaling_mucut, double truncate, int quiet) /*TZ ds*/
{
  int i=0, l=0, status=0;
  int imin=0, imax=0;

  
  double r0=0, r0_old, dr=0;
  size_t nreff=0, nreffact=0, nphase=0;

  double *mu=NULL, *p=NULL;
  int n=0;

  double *legen=NULL;

  if (rmin > rmax) {
    fprintf (stderr, "Error, found rmin > rmax; this message indicates that you told MYSTIC\n");
    fprintf (stderr, "to use a Mie phase function for a cloud of zero optical thickness;\n");
    fprintf (stderr, "Please change either of these two because this causes trouble internally!\n");
    return -1;
  }

  /* get radius intervals */
  r0     = ssprop->r0[iv];
  dr     = ssprop->dr[iv];
  nreff  = ssprop->nreff[iv];
  nphase = ssprop->nphase;


  /* check if the range of radii specified in the 3D input file is covered; */
  /* if only one radius is provided and its value is negative, the phase    */
  /* function is assumed to be constant with location.                      */
     

  if (nreff==1 && r0<0) { 
    fprintf (stderr, " ... found only one (negative) effective radius; assuming\n");
    fprintf (stderr, " ... phase function is constant with location.\n");
  }
  else {
    if (rmin < r0 || rmax > r0 + (double) (nreff-1) * dr) {
      fprintf (stderr, "Error, range of effective radii provided is not sufficient:\n");
      fprintf (stderr, " provided: %g - %g, required: %g - %g\n", 
	       r0, r0 + (double) (nreff-1) * dr, rmin, rmax);
      return -1;
    }
  }
  

  /* determine the required range of effective radii */
  imin=0;
  while (r0 + (double) imin * dr < rmin) {
    imin++;
    if (imin==nreff) {
      imin = nreff-1;
      break;
    }
  }

  imin = (imin>0?imin-1:0);
  
  imax=nreff-1;
  while (r0 + (double) imax * dr > rmax) {
    imax--;
    if (imax==-1) {
      imax = 0;
      break;
    }
  }
  
  imax = (imax<nreff-1?imax+1:nreff-1);

  r0_old = r0;
  r0 = r0_old + (double) imin * dr;

  nreffact = (imax-imin+1);

  if (!quiet)
    fprintf (stderr, " ... using %d radii between %g and %g\n", 
	     nreffact, r0_old + (double) imin * dr, r0_old + (double) imax * dr);


  calloc_phase_function_table (phase, nreffact, r0, dr);
  
  for (i=0; i<nreffact; i++) {   /* calculate only those tables which are really needed */
    
    if (!quiet)
      fprintf (stderr, " ... calculating Legendre table for reff = %g\n", 
	       r0_old + (double) (imin+i) * dr);
    
    switch (ssprop->type) {
    case PHASE_MOMENTS:

      /* first evaluate phase function at a predefined grid */
      legen = calloc (ssprop->nleg[iv][imin+i], sizeof(double));
      for (l=0; l<ssprop->nleg[iv][imin+i]; l++)
	legen[l] = (double) ssprop->legen[iv][imin+i][0][l];

      /* calculate table of phase function p(mu) */
      status = calc_Legendre_phase (legen, ssprop->nleg[iv][imin+i], &mu, &p, &n);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by calc_Legendre_phase()\n", status);
	return status;
      }
      
      status = calc_cumulative_table (mu, p, n, truncate, phase->iphase[i], delta_scaling_mucut);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d calculating cumulative probability table\n", status);
	return status;
      }

      free(legen); free(mu); free(p);
      break;

    case PHASE_EXPLICIT:
      /* calculate table of phase function p(mu)     */
      /* ??? only for first phase matrix element ??? */
      status = calc_cumulative_table (ssprop->mu[iv][imin+i], ssprop->p[iv][imin+i][0], ssprop->nmu[iv][imin+i], 
				      truncate, phase->iphase[i], delta_scaling_mucut);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d calculating cumulative probability table\n", status);
	return status;
      }
      break;
      
    default:
      fprintf (stderr, "Error, unknown phase function specification %d\n", ssprop->type);
      return -1;
    }

    /* copy extinction coefficient to final destination */
    phase->ext[i]   = ssprop->extinc[iv][imin+i];

    /* copy single scattering albedo to final destination */
    phase->omega[i] = ssprop->albedo[iv][imin+i];

    /* copy single scattering albedo to final destination */
    phase->Dscale[i] = phase->iphase[i]->Dscale;

    fprintf (stderr, "DDSSCALE %f\n", phase->Dscale[i]);

  }
  
  /* calculate extinction and single scattering albedo increments */
  for (i=0; i<nreffact-1; i++) {
    phase->dext   [i] = (phase->ext   [i+1] - phase->ext   [i]) / dr;
    phase->domega [i] = (phase->omega [i+1] - phase->omega [i]) / dr;
    phase->dDscale[i] = (phase->Dscale[i+1] - phase->Dscale[i]) / dr;
  }

  return 0;
}


/***********************************************************************************/
/* Function: free_pft                                                     @45_30i@ */
/* Description:                                                                    */
/*  Free memory of struct pft; used by free_phase_function_table().                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

static void free_pft (pft *p)
{
  free (p->mu);
  free (p->p);
  free (p->F);

  free (p->A);
  free (p->B);
  free (p->C);

  /*TZ ds neu*/ 
  free (p->po);
  free (p->Fo);

  free (p->Ao);
  free (p->Bo);
  free (p->Co);
}


/***********************************************************************************/
/* Function: calloc_phase_function_table                                  @45_30i@ */
/* Description:                                                                    */
/*  Allocate memory for struct phase_function_table                                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

void calloc_phase_function_table (phase_function_table *p, int n, double r0, double dr)
{
  int i=0;

  p->n  = n;
  p->r0 = r0;
  p->dr = dr;

  p->iphase = calloc ((size_t) n, sizeof(pft *));

  p->ext  = calloc ((size_t) n, sizeof(double));
  p->dext = calloc ((size_t) n, sizeof(double));

  p->omega  = calloc ((size_t) n, sizeof(double));
  p->domega = calloc ((size_t) n, sizeof(double));

  p->Dscale  = calloc ((size_t) n, sizeof(double));
  p->dDscale = calloc ((size_t) n, sizeof(double));

  /* initialize Dscale with 1 */
  for (i=0; i<n; i++)
    p->Dscale[i]=1.0;

  for (i=0; i<n; i++)
    p->iphase[i] = calloc (1, sizeof(pft));
}


/***********************************************************************************/
/* Function: free_phase_function_table                                    @45_30i@ */
/* Description:                                                                    */
/*  Free memory of struct phase_function_table                                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

void free_phase_function_table (phase_function_table *p)
{
  int i=0;

  for (i=0; i<p->n; i++)
    free_pft (p->iphase[i]);

  free (p->iphase);

  free (p->ext);
  free (p->dext);

  free (p->omega);
  free (p->domega);

  free (p->Dscale);
  free (p->dDscale);
}




/***********************************************************************************/
/* Function: calc_Legendre_phase                                          @45_30i@ */
/* Description:                                                                    */
/*  Calculate the phase function from Legendre coefficients                        */ 
/*  for a set of polar angles.                                                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i45_30@ */
/***********************************************************************************/

int calc_Legendre_phase (double *pmom, int nmom, double **mu, double **p, int *n)
{
  int l=0, it=0;

  double pl=0, plm1=0, plm2=0;
  double factor=0, sum=0, increment=0;

  int L=nmom;

  double *f = calloc ((size_t) nmom, sizeof(double));


  /* multiply by (2l+1) to convert from 'moments of the phase function' */
  /* to Legendre coefficients;                                          */
  for (l=0; l<L; l++)
    f[l] = pmom[l] * (double) (2*l+1);

  /* normalize to 1 */
  factor = 1.0 / f[0];

  for (l=0; l<L; l++)
    f[l] *= factor;

  *n = NPHASE;

  *mu = calloc ((size_t) (*n), sizeof(double));
  *p  = calloc ((size_t) (*n), sizeof(double));

  increment = PI/(double) (*n-1);

  for (it=0; it<*n; it++) {
    
    (*mu)[it] = -cos((double) it * increment);

    plm2 = 1;
    plm1 = (*mu)[it];
  
    sum = plm2*f[0] + plm1*f[1];
 
    for (l=2; l<L; l++) {
      pl = ( (double) (2*l-1) * (*mu)[it] * plm1 - (double) (l-1) * plm2) / (double) l;
      
      sum += f[l]*pl;

      plm2=plm1;
      plm1=pl;
    }

    (*p)[it] = sum;

    if ((*p)[it]<0) {
      fprintf (stderr, "Warning, phase function %g negative at mu = %g!\n", 
	       (*p)[it], (*mu)[it]);
    }

  }

  /* free memory */
  free(f);

  return 0;
}


/***********************************************************************************/
/* Function: calc_cumulative_table                                        @62_30i@ */
/* Description:                                                                    */
/*  Calculate a table of cumulative probabilities F1[], directions mu1[],          */
/*  and slopes D1[] for use with sc_mu(); the table is calculated from the         */
/*  phase function itself, not from its Legendre coefficients.                     */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int calc_cumulative_table (double *mu, double *p, int n, double truncate, 
			   pft *iphase, double delta_scaling_mucut) /*TZ ds*/
{
  int i=0, it=0;
  double *F=NULL;
  double alpha=0, beta=0, factor=0;
  int itcut=0; /*TZ ds*/
  double delta_scaling_factor=0;/*TZ ds*/

  if (mu[0]!=-1.0) {
    fprintf (stderr, "Error, first mu in phase function must be -1.0!\n");
    fprintf (stderr, "Found %.10f\n", mu[0]);
    return -1;
  }

  if (mu[n-1]!=1.0) {
    fprintf (stderr, "Error, last mu in phase function must be +1.0!\n");
    fprintf (stderr, "Found %.10f\n", mu[n-1]);
    return -1;
  }

  /* calculate cumulative phase function by linear integration */
  F = calloc (n, sizeof(double));
  F[0] = 0;
  for (i=1; i<n; i++)
    F[i] = F[i-1] + (mu[i]-mu[i-1])*(p[i]+p[i-1])/2.0;

  /* normalize to 2.0 */
  if (F[n-1]!=2.0) {
    fprintf (stderr, " ... renormalizing phase function by factor %f\n", 2.0/F[n-1]);
    factor = 2.0/F[n-1];

    for (i=0; i<n; i++) {
      F[i] *= factor;
      p[i] *= factor;
    }
  }

  /* copy to final destination */
  iphase->n = n;
  
  iphase->mu = calloc ((size_t) iphase->n, sizeof(double));
  iphase->p  = calloc ((size_t) iphase->n, sizeof(double));
  iphase->F  = calloc ((size_t) iphase->n, sizeof(double));
  iphase->A  = calloc ((size_t) iphase->n, sizeof(double));
  iphase->B  = calloc ((size_t) iphase->n, sizeof(double));
  iphase->C  = calloc ((size_t) iphase->n, sizeof(double));
  iphase->po  = calloc ((size_t) iphase->n, sizeof(double)); /*TZ ds*/
  iphase->Fo  = calloc ((size_t) iphase->n, sizeof(double)); /*TZ ds*/
  iphase->Ao  = calloc ((size_t) iphase->n, sizeof(double)); /*TZ ds*/
  iphase->Bo  = calloc ((size_t) iphase->n, sizeof(double)); /*TZ ds*/
  iphase->Co  = calloc ((size_t) iphase->n, sizeof(double)); /*TZ ds*/

  for (i=0; i<iphase->n; i++) {
    iphase->mu[i] = mu[i];
    iphase->p[i]  = p[i];
    iphase->F[i]  = F[i];

    iphase->po[i]  = p[i];/*TZ ds*/
    iphase->Fo[i]  = F[i];/*TZ ds*/
  }  

  /* calculate coefficients for linear interpolation of the table */
  for (i=0; i<iphase->n-1; i++) {
    beta  = (p[i+1]-p[i]) / (mu[i+1]-mu[i]);
    alpha = p[i] - beta * mu[i];

    iphase->C[i] = F[i] - alpha*mu[i] - 0.5*beta*mu[i]*mu[i];
    iphase->B[i] = alpha;
    iphase->A[i] = beta/2.0;

    iphase->Co[i] = F[i] - alpha*mu[i] - 0.5*beta*mu[i]*mu[i]; /*TZ ds*/
    iphase->Bo[i] = alpha; /*TZ ds*/
    iphase->Ao[i] = beta/2.0; /*TZ ds*/
  }


  /**************************************************************************/
  /* von hier TZ ds ... */
  itcut=n;

  /* initialize delta-scaling factor to 1 = no delta-scaling */
  iphase->Dscale = 1.0;

  if (delta_scaling_mucut>-1.0) {
    double dif=99999.;
    for (i=0; i<iphase->n; i++) {
      if (fabs(mu[i]-delta_scaling_mucut) < dif) { /*find closest mu in tabulated values*/
	dif=fabs(mu[i]-delta_scaling_mucut);
	itcut=i;
      }
    }

    fprintf (stderr, " ... delta-scaling ... truncating phase function at mucut = %g\n", mu[itcut]);
        
    delta_scaling_factor=2./F[itcut];/* F, the integral of p, is 2 */
    iphase->Dscale=delta_scaling_factor;
    
    for (i=0; i<n; i++) {
      F[i] =  F[i]*delta_scaling_factor; /*TZ ds: delta-scaled integrated phase function*/
      p[i] =  p[i]*delta_scaling_factor; /*TZ ds: delta-scaled phase function*/
      if (i>itcut) F[i] = 2.;            /*TZ ds: delta-scaled integrated phase function*/
      if (i>itcut) p[i] = 0;             /*TZ ds: delta-scaled phase function*/
    }    

    for (i=0; i<iphase->n; i++) {
      iphase->p[i]  = p[i];
      iphase->F[i]  = F[i];
    }  

    /* calculate coefficients for linear interpolation of the table  */
    /* for delta-scaled version of the phase function */
    for (i=0; i<iphase->n-1; i++) {
      beta  = (p[i+1]-p[i]) / (mu[i+1]-mu[i]);
      alpha = p[i] - beta * mu[i];
      
      iphase->C[i] = F[i] - alpha*mu[i] - 0.5*beta*mu[i]*mu[i];
      iphase->B[i] = alpha;
      iphase->A[i] = beta/2.0;
    } 
  }
  /* ... bis hier TZ ds*/  
  /**************************************************************************/


  if (truncate>-1.0) {
    iphase->truncate = 1;
    iphase->mutrnc = truncate;
    
    it = locate (iphase->mu, iphase->n, iphase->mutrnc);

    if (it==iphase->n-1 || iphase->mutrnc >= 1.0) {
      fprintf (stderr, " ... no need to truncate at %f\n", iphase->mutrnc);
      iphase->truncate = 0;
    }
    else {
      /* value of the phase function at the truncation angle */
      iphase->ptrnc = iphase->p[it] + (iphase->mutrnc - iphase->mu[it]) / 
	(iphase->mu[it+1] - iphase->mu[it]) * (iphase->p[it+1] - iphase->p[it]);

      iphase->Ftrnc = iphase->F[it] + (iphase->mutrnc - mu[it]) * (iphase->ptrnc+p[it])/2.0;
      iphase->dtrnc = 2.0/(1.0-iphase->mutrnc) * ((2.0-iphase->Ftrnc)/(1.0-iphase->mutrnc) - iphase->ptrnc);
    }
  }
  

  /* free memory */
  free (F);

  return 0;
}


/***********************************************************************************/
/* Function: read_phase_function                                          @62_30i@ */
/* Description:                                                                    */
/*  Read phase function from a 2-column file.                                      */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int read_phase_function (char *filename, 
			 double *r0, double *dr, size_t *nreff, size_t *nphase,
			 double **extinc, double **albedo, int **nmu, 
			 double ***mu, double ****phase, int *type, int quiet)
{
  int ir=0, imu=0, status=0;

  *nreff  = 1;  /* number of effective radii       */ 
  *nphase = 1;  /* number of phase matrix elements */
  
  /* allocate memory for fields */
  *extinc = calloc ((size_t) *nreff, sizeof (double));
  *albedo = calloc ((size_t) *nreff, sizeof (double));
  
  *r0 = -999;
  *dr = -999;
  
  *nmu    = calloc ((size_t) *nreff, sizeof (int));
  *mu     = calloc ((size_t) *nreff, sizeof (double *));
  *phase  = calloc ((size_t) *nreff, sizeof (double **));

  for (ir=0; ir<*nreff; ir++)
    *phase[ir] = calloc (*nphase, sizeof (double *));
 
  if (!quiet) 
    fprintf (stderr, " ... reading two columns of data from %s\n", filename);

  status = read_2c_file (filename, &((*mu)[0]), &((*phase)[0][0]), &((*nmu)[0]));
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* check if the user defined mu or theta */
  for (imu=0; imu<(*nmu)[0]; imu++) {
    if ((*mu)[0][imu]<-1 || (*mu)[0][imu]>1) {
      fprintf (stderr, "Error in %s: Found mu=%f, expecting mu=cos(theta), p(mu)\n", filename, (*mu)[0][imu]);
      return -1;
    }
  }

  *type = PHASE_EXPLICIT;

  if (!quiet)
    fprintf (stderr, " ... read %d coefficients from %s\n", (*nmu)[0], filename);

  return 0;
}


