/*--------------------------------------------------------------------
 * $Id: phasetable.h,v 1.1 2007/07/06 23:01:07 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#ifndef __phasetable_h
#define __phasetable_h

#include <stdio.h>


/* phase function defined by moments or explicitely */
#define PHASE_MOMENTS  1
#define PHASE_EXPLICIT 2


/* table of phase function and cumulative phase function;  */
/* between the specified data points the phase function    */
/* is assumed to vary linearely in mu; in consequence, the */
/* cumulative phase function varies quadratic in mu; the   */
/* coefficients A, B, and C are used for the interpolation */
/* of mu between two adjacent cumulative phase function    */
/* values (needed for sampling of a random direction)      */
typedef struct {
  double *mu;   /* cosine of scattering angle;                     */ 
                /* arbitrary values but sorted in increasing order */

  double *p;    /* scattering phase function, normalized to 2      */
  double *po;   /*TZ ds neu*/ 

  double *F;    /* cumulative phase function, 0-2!                 */
  double *Fo;   /*TZ ds neu*/ 

  double *A;    /* coefficients for the interpolation of mu        */
  double *B;    /* between two adjancent values of the cumulative  */
  double *C;    /* phase function F                                */
  double *Ao;   /*TZ ds neu*/ 
  double *Bo;   /*TZ ds neu*/
  double *Co;   /*TZ ds neu*/

  int n;        /* number of data points                           */

  int truncate;   /* truncation yes/no                                         */ 
  double mutrnc;  /* truncation angle                                          */
  double ptrnc;   /* phase function at truncation angle                        */
  double Ftrnc;   /* cumulative phase function at truncation angle             */
  double dtrnc;   /* increment for truncation: p = ptrnc + dtrnc * (mu-mutrnc) */

  double Dscale;  /*TZ ds*/
} pft;

typedef struct {
  pft **iphase;    /* phase function table for each effective radius reff[i] */
  double *ext;     /* extinction coefficient for each effective radius       */
  double *dext;    /* extinction coefficient increment                       */
  double *omega;   /* single scattering albedo for each effective radius     */
  double *domega;  /* single scattering albedo increment                     */
  double *Dscale;  /* delta-scaling factor                              TZ ds*/
  double *dDscale; /* delta-scaling factor increment                    TZ ds*/

  double r0;       /* reff[i] = r0 + i * dr; i = 0 ... n-1                   */
  double dr;
  int    n;
} phase_function_table;

typedef struct {
  /* the following quantities are defined for each wavelength, */
  /* on an equidistant grid of effective radii,                */
  /* r0, r0+dr, r0+2*dr, ..., r0 + (nreff-1)*dr                */

  double **extinc;    /* extinction coefficient           [iv][ireff]               */
  double **albedo;    /* single scattering albedo         [iv][ireff]               */

  /* either define Legendre moments of the phase function                           */
  float  ****legen;   /* Legendre moments                 [iv][ireff][iphase][ileg] */
  int    **nleg;      /* number of Legendre moments       [iv][ireff]               */

  /* or define the phase function itself at a set of angles                         */
  int    **nmu;       /* Number of phase function angles  [iv][ireff]               */
  double ***mu;       /* Phase function angles            [iv][ireff][imu]          */
  double ****p;       /* Phase function                   [iv][ireff][iphase][imu]  */

  double  *r0;        /* start radius for each wavelength [iv]                      */
  double  *dr;        /* wavelength step                  [iv]                      */
  size_t  *nreff;     /* number of effective radii        [iv]                      */ 
  size_t   nphase;    /* number of phase matrix elements                            */

  int   alloc;          /* flag, if basic memory is allocated */
  int   alloc_explicit; /* flag, if memory for explicit phase function is allocated  */
  int   alloc_moments;  /* flag, if memory for phase function moments is allocated   */

  int type;            /* phase function or Legendre moments */
} cld_ssprop_struct;





int calc_Legendre_phase (double *pmom, int nmom, double **mu, double **p, int *n);

int read_mie_table (char *filename, 
		    double *r0, double *dr, size_t *nreff, 
		    double *wavelen, double *nre, double *nim,
		    double **extinc, double **albedo, int **nleg, 
		    float ****legen, size_t *nphase, int quiet);

int read_mie_table_lambda (char *filename, 
			   double *wavelen, int quiet);

int setup_Legendre_table (phase_function_table *phase,
			  cld_ssprop_struct *ssprop,
			  int iv,
			  double rmin, double rmax, 
			  double delta_scaling_mucut, double truncate, int quiet); /*TZ ds*/

int calc_cumulative_table (double *mu, double *p, int n, double truncate, 
			   pft *iphase, double delta_scaling_mucut); /*TZ ds*/

int read_phase_function (char *filename, 
			 double *r0, double *dr, size_t *nreff, size_t *nphase,
			 double **extinc, double **albedo, int **nmu, 
			 double ***mu, double ****phase, int *type, int quiet);

void calloc_phase_function_table (phase_function_table * phase, int n, double r0, double dr);
void free_phase_function_table (phase_function_table *phase);

#endif



