/*--------------------------------------------------------------------
 * $Id: cloud3d.c,v 1.18 2008/01/09 15:21:10 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <float.h>

#include "uvspec.h"
#include "ckdfu.h"
#include "cloud3d.h"
#include "ipa.h"
#include "wcloud3d.h"
#include "mystic.h"
#include "f77-uscore.h"


#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


/* prototypes of internal functions */
/* static int compare_levels_df (double *z1, int n1, float  *z2, int n2); */

static int read_and_check_sample2D (char *samplefilename,
				    float *zd, int nlev,
				    int nxout, int nyout, 
				    double dxout, double dyout, 
				    float *zout, int nzout,
				    int radpat_Nr, int radpat_Nt, float radpat_dt, 
				    float *umu, int numu, float *phi, int nphi,
				    int aziconv,
				    sample_struct *sample,
				    int escape, int calcstd,
				    int backward, int output,
				    int backward_islower, int backward_jslower,
				    int backward_isupper, int backward_jsupper,
				    int backward_writeallpixels,
				    int spherical, int bcond,
				    int quiet);


static int read_wc3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output);


/***********************************************************************************/
/* Function: equal                                                        @62_30i@ */
/* Description:                                                                    */
/*  Two numbers equal?                                                             */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

static int equal (double x, double y)
{
  if ((float) x != (float) y)
    return 0;

  return 1;
}


/***********************************************************************************/
/* Function: read_and_check_sample2D                                      @62_30i@ */
/* Description:                                                                    */
/*  Initialize sample_struct with data read from samplefilename and passed as      */
/*  function arguments.                                                            */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  float *zout:  array of altitudes to be sampled; overwrites data from           */
/*                samplefilename if nzout>0                                        */
/*  int   nzout:  number of altitudes to be sampled; if nzout<=0, data from        */
/*                samplefilename are used.                                         */
/*                                                                                 */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

static int read_and_check_sample2D (char *samplefilename,
				    float *zd, int nlev,
				    int nxout, int nyout, 
				    double dxout, double dyout, 
				    float *zout, int nzout,
				    int radpat_Nr, int radpat_Nt, float radpat_dt, 
				    float *umu, int numu, float *phi, int nphi,
				    int aziconv,
				    sample_struct *sample,
				    int escape, int calcstd,
				    int backward, int output,
				    int backward_islower, int backward_jslower,
				    int backward_isupper, int backward_jsupper,
				    int backward_writeallpixels,
				    int spherical, int bcond,
				    int quiet)
{
  int status=0;
  int lu=0, iu=0, j=0, id=0, kc=0;
  double *z_hist=NULL;
  int Nd=0;

  double xr=0, yr=0;

  double *stheta=NULL, *sphi=NULL, *salpha=NULL;

  FILE *ftmp=NULL;

  if (!quiet)
    fprintf (stderr, " ... reading sampling information\n");

  
  /* check if sample file exists and can be opened for reading */
  ftmp = fopen (samplefilename, "r");
  if (ftmp==NULL) {
    /* if all 4 parameters have been defined in the input file, */
    /* then we don't need sample2D.dat                          */
    if (nxout>0&&nyout>0 && dxout>0 && dyout>0) {
      if (!quiet)
	fprintf (stderr, " ... %s not found and not required\n", samplefilename);
    }
    else {
      fprintf (stderr, "Error! Either define the sample grid in the input file or provide sample2D.dat\n");
      return -1;
    }
  }
  else {
    fclose(ftmp);
  
    /* read sample2D.dat */
    status = read_2D_sample (samplefilename,
			     &(sample->Nx), &(sample->Ny),
			     &(sample->delX), &(sample->delY),
			     &stheta, &sphi, &salpha, &Nd, aziconv, 
			     quiet);
    
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, samplefilename);
      return status;
    }
  }

  /* allocate memory for sample flag */
  sample->sample = calloc ((size_t) nlev, sizeof(int));
  
  /* over-write grid size, if nxout > 0 and nyout > 0 */
  if (nxout>0 && nyout>0) {

    if (dxout>0 && dyout>0) {
      sample->delX = dxout * 1000.0;  /* convert to m */
      sample->delY = dyout * 1000.0;  /* convert to m */
    }
    else {   /* need to recalculate */
      sample->delX = (double) (sample->Nx) * (sample->delX) / (double) nxout;
      sample->delY = (double) (sample->Ny) * (sample->delY) / (double) nyout;
    }

    sample->Nx = nxout;
    sample->Ny = nyout;

    if (!quiet)
      fprintf (stderr, " ... overwriting sample grid, now %d x %d pixels with area %f x %f\n",
	       sample->Nx, sample->Ny, sample->delX, sample->delY);

    /* check if we really want that always */
    sample->passback3D = 1;
  }

  /* over-write altitude profile, if nzout > 0 */
  if (nzout>0) {   /* use data passed as function arguments */

    /* first check if zout is sorted in ascending order */
    for (lu=1; lu<nzout; lu++)
      if (zout[lu] - zout[lu-1] < 0) {
	fprintf (stderr, "Error, zout not sorted in ascending order\n");
	status--;
      }
    
    /* ??? special treatment of -999, surface ??? */

    if (!quiet)
      fprintf (stderr, " ... using %d altitudes from the uvspec input file:\n", nzout);
    
    for (lu=0; lu<nzout; lu++) {
      if (zout[lu]>=0) {
	for (kc=0; kc<nlev; kc++) {
	  if (fabs(zd[nlev-1-kc] - zout[lu]) < 1E-3) {
	    sample->sample[kc]=1;
	    break;
	  }
	}
	if (kc==nlev) {
	  fprintf (stderr, "\nFATAL error: altitude grid does not contain level %g\n", zout[lu]);
	  fprintf (stderr, "which has been specified as output altitude in %s\n", samplefilename);
	  return -1;
	}
      }
    }
  }

  if (!quiet) {
    fprintf (stderr, " ... sampling result at the following altitudes:\n");

    for (kc=0; kc<nlev; kc++)
      if (sample->sample[kc])
	fprintf (stderr, " ...  %8.3f km\n", zd[nlev-1-kc]);

    fprintf (stderr, "\n");
  }

  if (zout[0]<0)
    sample->surface = 1;
    
  /* over-write radiance angles, if numu>0 and nphi>0 */
  if (numu>0 && nphi>0) {   /* use data passed as function arguments */

    if (!quiet)
      fprintf (stderr, " ... using %d radiance angles from the uvspec input file:\n", numu * nphi);

    /* allocate memory for sample radiances */
    Nd = numu * nphi;
    sample->Nd = Nd;
    if (Nd>0)
      sample->rad  = calloc (Nd, sizeof (radang));

    for (iu=0; iu<numu; iu++) {
      for (j=0; j<nphi; j++) {

	/* copy uvspec radiance to sample array */
	sample->rad[id].theta = acos(umu[iu])/PI*180.0;
	sample->rad[id].phi   = phi[j];
	sample->rad[id].alpha = 5.0;   /* default; should change that by adding */
	                               /* another libRadtran input option       */

	if (!quiet)
	  fprintf (stderr, " ...  (%g, %g)\n", sample->rad[id].theta, sample->rad[id].phi);
	
	id++;
      }
    }
  }
  else {   /* use data from sample2D.dat */
    
    /* allocate memory for sample radiances */
    sample->Nd = Nd;
    if (Nd>0)
      sample->rad  = calloc (Nd, sizeof (radang));
    
    for (id=0; id<Nd; id++) {
      sample->rad[id].theta = stheta [id];
      sample->rad[id].phi   = sphi   [id];
      sample->rad[id].alpha = salpha [id];
    }
  }  

  /****************************************************/
  /* Changed MYSTIC azimuth convention March 1, 2004! */
  /* From then on, the MYSTIC azimuth convention      */
  /* follows the DISORT convention, that is, as seen  */
  /* from the photon rather than from the detector.   */
  /* azimuth 0 = coming from the direction of the sun */
  /* (looking towards the sun) while azimuth 180 is   */
  /* going into the direction of the sun (or for the  */
  /* observer, sun in the back. If aziconv is set to  */
  /* MCAZIMUTH_OLD then the old convention is used.   */
  /*                                                  */
  /* Inside MYSTIC nothing is changed; azimuths       */
  /* still refer to the observing instrument.         */
  /*                                                  */
  /* To achieve this, we need to reverse the azimuths */
  /* at this location and store the original values   */
  /* for the output.                                  */
  /****************************************************/

  for (id=0; id<sample->Nd; id++) {
    sample->rad[id].externalphi = sample->rad[id].phi;

    switch (aziconv) {
    case MCAZIMUTH_NEW:
      sample->rad[id].phi = 180.0 + sample->rad[id].phi;
      break;
    case MCAZIMUTH_OLD:
      sample->rad[id].phi = sample->rad[id].phi;
      break;
    default: 
      fprintf (stderr, "Error, unknown azimuth convention %d\n", aziconv);
      return -1;
    }

    /* shift to interval [0 ... 360] */
    while (sample->rad[id].phi < 0.0)
      sample->rad[id].phi += 360.0;

    while (sample->rad[id].phi > 360.0)
      sample->rad[id].phi -= 360.0;
  }


  if (stheta != NULL) free (stheta); 
  if (sphi != NULL)   free(sphi); 
  if (salpha != NULL) free(salpha);


  if (sample->Nd>0) {
    sample->escape = escape;
    
    if (!quiet)
      fprintf (stderr, " ... calculating radiances for %d directions\n", sample->Nd);
    
    fprintf (stderr, "**********************************************************\n");
    fprintf (stderr, "**********************************************************\n");
    fprintf (stderr, "** Attention! On March 1, 2004, the MYSTIC azimuth      **\n");
    fprintf (stderr, "** was made consistent to the disort convention         **\n");
    fprintf (stderr, "** for compatibility reasons. That is, the              **\n");
    fprintf (stderr, "** user-defined azimuth is now interpreted as seen from **\n");    
    fprintf (stderr, "** the photon rather than from the detector.            **\n");
    fprintf (stderr, "** azimuth = 0 means coming from the direction of the   **\n");
    fprintf (stderr, "** sun (detector looking towards the sun) while         **\n");
    fprintf (stderr, "** azimuth = 180 means going towards the direction      **\n");
    fprintf (stderr, "** of the sun (or, from the detector: sun in the back). **\n");
    fprintf (stderr, "** If you don't like that, please use mc_azimuth_old    **\n");
    fprintf (stderr, "** in the uvspec input file.                            **\n");
    fprintf (stderr, "**********************************************************\n");
    fprintf (stderr, "**********************************************************\n");
  }


  if (calcstd)
    sample->std = 1;
  else 
    sample->std = 0;
  

  /* radial pathlength sampling */
  sample->Nr = radpat_Nr;
  sample->Nt = radpat_Nt;
  sample->dt = radpat_dt;

  /* fit circle into domain */
  xr = 0.5 * (double) sample->Nx * sample->delX;
  yr = 0.5 * (double) sample->Ny * sample->delY;

  sample->dr = (xr < yr ? xr : yr) / (double) sample->Nr;

  if (!quiet && sample->Nr > 0)
    fprintf (stderr, " ... radial pathlength sampling with intervals dr = %.6e and dt = %.6e\n",
	     sample->dr, sample->dt);

  free (z_hist);

  /* number of levels to be sampled */
  for (kc=0; kc<nlev; kc++)
    sample->nzout += sample->sample[kc];




  /* calculate zenith and azimuth angles */ 
  if (backward) {
    if (Nd>0)
      sample->backward = MCBACKWARD_RADIANCE;
    else 
      sample->backward = output;  /* currently only downward diffuse is implemented */

    sample->backward_writeallpixels = backward_writeallpixels;

    sample->backward_islower = backward_islower;
    sample->backward_jslower = backward_jslower;
    sample->backward_isupper = backward_isupper;
    sample->backward_jsupper = backward_jsupper;

    if (sample->backward_islower < 0) sample->backward_islower=0;
    if (sample->backward_jslower < 0) sample->backward_jslower=0;
    if (sample->backward_isupper < 0) sample->backward_isupper=0;
    if (sample->backward_jsupper < 0) sample->backward_jsupper=0;

    if (sample->backward_islower >= sample->Nx) sample->backward_islower = sample->Nx-1;
    if (sample->backward_jslower >= sample->Ny) sample->backward_jslower = sample->Ny-1;
    if (sample->backward_isupper >= sample->Nx) sample->backward_isupper = sample->Nx-1;
    if (sample->backward_jsupper >= sample->Ny) sample->backward_jsupper = sample->Ny-1;
  }

  if (sample->backward && !sample->escape) {
    sample->escape = 1;
    if (!quiet)
      fprintf (stderr, " ... switching on local estimate (default for backward calculations)\n");
  }

  if (sample->backward && (sample->Nr>0 && sample->dt>0)) {
    fprintf (stderr, "Error, radial/pathlength and backward cannot be combined!\n");
    return -1;
  }

  /* boundary conditions */
  sample->bcond = bcond;

  /* spherical geometry - always use backwards and absorbing boundary conditions! */
  sample->spherical = spherical;
  if (spherical) {
    if (Nd>0)
      sample->backward = MCBACKWARD_RADIANCE;
    else 
      sample->backward = output;  /* currently only downward diffuse is implemented */

    sample->escape = 1;
    sample->backward_writeallpixels = 0;

    if (sample->backward_islower!=0 || 
	sample->backward_jslower!=0 ||
	sample->backward_isupper!=0 || 
	sample->backward_jsupper!=0) {
      fprintf (stderr, " ... backward sampling area ignored in spherical geometry!\n");
      fprintf (stderr, " ... (photons are always started at the center of the domain)\n");
    }
    
    sample->backward_islower = 0;
    sample->backward_jslower = 0;
    sample->backward_isupper = 0;
    sample->backward_jsupper = 0;

    if (sample->bcond!=MCBCOND_ABSORB)
      fprintf (stderr, " ... using absorbing boundary conditions for spherical\n");

    sample->bcond = MCBCOND_ABSORB;
  }

  return 0;
}


/***********************************************************************************/
/* Function:                                                              @62_30i@ */
/* Description:                                                                    */
/*  Setup a 3D atmosphere structure, to be used as input to MYSTIC.                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int setup_atmosphere3D (input_struct input, output_struct *output)
{
  int kc=0, status=0;
  char cloud3d_filename[FILENAME_MAX];
  char samplefilename[FILENAME_MAX]="sample2D.dat";
  FILE *f=NULL;
  int readwc=0;

  /* reset counters */
  output->wc.threed.nthreed = 0;
  output->ic.threed.nthreed = 0;

  /* don't need to read cloud properties if not a 3D calculation */
  /* 3D cloud file defined or no - need to re-think once 3D and  */
  /* IPA are homogenized ???                                     */
  if (input.rte.solver != MONTECARLO)
    return 0;
  

  /*******************************/
  /* first, read the sample file */
  /*******************************/


  status = read_and_check_sample2D (samplefilename,
				    output->atm.zd, output->atm.nlev,
				    input.rte.mc.Nx_sample, input.rte.mc.Ny_sample,
				    input.rte.mc.dx_sample, input.rte.mc.dy_sample,
				    output->atm.zout_sur, output->atm.nzout,
				    input.rte.mc.Nr, input.rte.mc.Nt, input.rte.mc.dt, 
				    input.rte.umu, input.rte.numu, 
				    input.rte.phi, input.rte.nphi, 
				    input.rte.mc.azimuth,
				    &(output->mc.sample), 
				    input.rte.mc.escape, input.rte.mc.std, 
				    input.rte.mc.backward.yes, input.rte.mc.backward.output,  
				    input.rte.mc.backward.islower, input.rte.mc.backward.jslower,
				    input.rte.mc.backward.isupper, input.rte.mc.backward.jsupper,
				    input.rte.mc.backward.writeallpixels,
				    input.rte.mc.spherical, input.rte.mc.bcond,
				    input.quiet);
  

  if (status!=0) {
    fprintf (stderr, "Error %d returned by read_and_check_sample2D()\n", status);
    return status;
  }



  /*******************************/
  /* second, read the cloud file */
  /*******************************/

  readwc=0;
  if (*input.rte.mc.filename[FN_MC_WCLOUD]) {
    strcpy (cloud3d_filename,  input.rte.mc.filename[FN_MC_WCLOUD]);
    readwc=1;
  }
  else {
    strcpy (cloud3d_filename,  "./wcloud3D.dat");
    
    /* test if file exists and is readable */
    if ((f=fopen(cloud3d_filename,"r"))==NULL) {
      fprintf (stderr, " ... did not find 3D water cloud file - trying without!\n");
      readwc=0;
    }
    else {
      readwc=1;
      fclose (f);
    }
  }

  /* copy cldproperties/nonHGcld information to 3D structure */
  /* even if no 3D cloud but only a 1D cloud is defined      */ 
  output->wc.threed.nonHGcld      = output->wc.nonHGcld;
  output->wc.threed.cldproperties = output->wc.cldproperties;

  if (readwc) {
    if (!input.quiet)
      fprintf (stderr, " ... reading 3D cloud data from %s\n", cloud3d_filename);
    

    /* read 3D cloud data */
    status = read_wc3D_file (cloud3d_filename, 
			     &(output->wc.threed), 
			     input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d reading 3D cloud water file %s\n", status, cloud3d_filename);
      return status;
    }
    
    if (!input.quiet)
    fprintf (stderr, " ... read %d x %d x %d cloud data set\n", 
	     output->wc.threed.Nx, output->wc.threed.Ny, output->wc.threed.nlyr);
    
    /* we got the 3D cloud structure in output->wc.threed now; */
    /* this structure is passed to mystic()                    */
    
    
    for (kc=0; kc<output->wc.threed.nlyr; kc++)
      if (output->wc.threed.threed[kc])
	output->wc.threed.nthreed++;
    
    if (!input.quiet)
      fprintf (stderr, " ... found %d 3D water cloud layers\n", output->wc.threed.nthreed);
    

    if (!input.quiet) {
      switch (output->wc.threed.cldproperties) {
      case CLD_OPTPROP:
	fprintf (stderr, " ... reading extinction coefficient, asymmetry parameter,\n");
	fprintf (stderr, " ... and single scattering albedo 3D cloud file\n");
	break;
      case CLD_EXTREFF:
	fprintf (stderr, " ... reading extinction coefficient and effective droplet radius from 3D cloud file\n");
	break;
      case CLD_LWCREFF:
	fprintf (stderr, " ... reading liquid water content and effective droplet radius from 3D cloud file\n");
	break;
      default:
	fprintf (stderr, "Error, unknown cloud properties %d\n", output->wc.threed.cldproperties);
	return -1;
      }
    }

    if (!input.quiet)
      fprintf (stderr, " ... checking 3D levels\n");
    
    /* make sure that 1D and 3D properties are defined on the same altitude grid */
    if (output->wc.threed.nlyr != output->atm.nlyr) {
      fprintf (stderr, "Error, number of layers in 3D water cloud different from atmospheric profile!\n");
      return -1;
    }
    
    /* ??? more testing required - this is currently done in mystic(), after re-gridding ??? */
    
    if (!input.quiet) {
      fprintf (stderr, " ... 3D cloud definition seems to be o.k.\n");
      
      for (kc=0; kc<output->wc.threed.nlyr; kc++)
	if (output->wc.threed.threed[kc])
	  fprintf (stderr, " ... found 3D clouds at level %d, altitude range %g - %g\n", 
		   kc, output->wc.threed.zd[kc], output->wc.threed.zd[kc+1]);
    }
  }
  else {  /* no 3D water cloud data */
    output->wc.threed.nthreed=0;
    output->wc.threed.nlyr = output->atm.nlyr;
  }

  /************************/
  /* ice cloud properties */
  /************************/

  /* don't need to read cloud properties if not a 3D calculation */
  /* 3D cloud file defined or no - need to re-think once 3D and  */
  /* IPA are homogenized ???                                     */
  if (strlen(input.rte.mc.filename[FN_MC_ICLOUD])>0) {
    strcpy (cloud3d_filename,  input.rte.mc.filename[FN_MC_ICLOUD]);
    
    if (!input.quiet)
      fprintf (stderr, " ... reading 3D ice cloud data from %s\n", cloud3d_filename);
    
    /* read 3D cloud data */
    status = read_3D_cloud (cloud3d_filename,
			    &(output->ic.threed.Nx), 
			    &(output->ic.threed.Ny), 
			    &(output->ic.threed.nlyr), 
			    &(output->ic.threed.delX), &(output->ic.threed.delY),
			    &(output->ic.threed.zd), 
			    &(output->ic.threed.lwc),
			    &(output->ic.threed.reff),
			    &(output->ic.threed.ext),
			    &(output->ic.threed.g1),
			    &(output->ic.threed.g2),
			    &(output->ic.threed.ff),
			    &(output->ic.threed.ssa),
			    &(output->ic.threed.Dscale),
			    &(output->ic.threed.reffmin), &(output->ic.threed.reffmax),
			    &(output->ic.threed.cldproperties),
			    &(output->ic.threed.threed),
			    input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading 3D ice cloud file %s\n", status, cloud3d_filename);
      return status;
    }
    
    if (!input.quiet)
    fprintf (stderr, " ... read %d x %d x %d cloud data set\n", 
	     output->ic.threed.Nx, output->ic.threed.Ny, output->ic.threed.nlyr);

    /* we got the ice 3D cloud structure in output->ic.threed now; */
    /* this structure is passed to mystic()                        */
    
    for (kc=0; kc<output->ic.threed.nlyr; kc++)
      if (output->ic.threed.threed[kc])
	output->ic.threed.nthreed++;
    
    if (!input.quiet)
      fprintf (stderr, " ... found %d 3D ice cloud layers\n", output->ic.threed.nthreed);
    
    fprintf (stderr, "*** ATTENTION: gspec and wspec for ice clouds are ignored\n");
    fprintf (stderr, "*** because Yang/Key optical properties for hexagonal columns\n");
    fprintf (stderr, "*** are currently assumed for ice particles! Column 4 is interpreted\n");
    fprintf (stderr, "*** as ice water content and column 5 as effective particle radius.\n");
    
    if (!input.quiet)
      fprintf (stderr, " ... checking 3D levels\n");
    
    /* make sure that 1D and 3D properties are defined on the same altitude grid */
    if (output->ic.threed.nlyr != output->atm.nlyr) {
      fprintf (stderr, "Error, number of layers in 3D ice cloud different from atmospheric profile!\n");
      return -1;
    }

    /* ??? more testing required - this is currently done in mystic(), after re-gridding ??? */

    if (!input.quiet) {
      fprintf (stderr, " ... 3D ice cloud definition seems to be o.k.\n");
      
      for (kc=0; kc<output->ic.threed.nlyr; kc++)
	if (output->ic.threed.threed[kc])
	  fprintf (stderr, " ... found 3D clouds at level %d, altitude range %g - %g\n", 
		   kc, output->ic.threed.zd[kc], output->ic.threed.zd[kc+1]);
    }
  }
  else {
    output->ic.threed.nthreed=0;
    output->ic.threed.nlyr=output->atm.nlyr;
  }    


  /* we require that water and ice clouds are defined */
  /* on identical horizontal grids                    */
  if (output->wc.threed.nthreed>0 && output->ic.threed.nthreed>0) {
    if (output->wc.threed.delX != output->ic.threed.delX || 
	output->wc.threed.delY != output->ic.threed.delY || 
	output->wc.threed.Nx != output->ic.threed.Nx     || 
	output->wc.threed.Ny != output->ic.threed.Ny) {
      fprintf (stderr, "Error: 3D water and ice clouds need to be defined on identical horizontal grids!\n");
      return -1;
    }
    
    fprintf (stderr, " ... horizontal water and ice cloud grids agree\n");
  }

  /* if only water clouds defined, copy water cloud grid to ice cloud grid */
  if (output->wc.threed.nthreed>0 && output->ic.threed.nthreed==0) {
    output->ic.threed.Nx   = output->wc.threed.Nx;
    output->ic.threed.Ny   = output->wc.threed.Ny;
    output->ic.threed.delX = output->wc.threed.delX;
    output->ic.threed.delY = output->wc.threed.delY;
  }

  /* if only ice clouds defined, copy ice cloud grid to water cloud grid */
  if (output->wc.threed.nthreed==0 && output->ic.threed.nthreed>0) {
    output->wc.threed.Nx   = output->ic.threed.Nx;
    output->wc.threed.Ny   = output->ic.threed.Ny;
    output->wc.threed.delX = output->ic.threed.delX;
    output->wc.threed.delY = output->ic.threed.delY;
  }

  /* if neither water nor ice clouds are defined, use the simplest possible grid */
  if (output->wc.threed.nthreed==0 && output->ic.threed.nthreed==0) {
    output->wc.threed.Nx   = 1;
    output->wc.threed.Ny   = 1;
    output->ic.threed.Nx   = 1;
    output->ic.threed.Ny   = 1;

    output->wc.threed.delX = 1000000; /* ???? need to define domain size in uvspec input ???? */
    output->wc.threed.delY = 1000000; /* ???? need to define domain size in uvspec input ???? */
    output->ic.threed.delX = 1000000; /* ???? need to define domain size in uvspec input ???? */
    output->ic.threed.delY = 1000000; /* ???? need to define domain size in uvspec input ???? */
  }

  /* copy to atmosphere structure */
  output->atm.Nxcld = output->wc.threed.Nx;
  output->atm.Nycld = output->wc.threed.Ny;
  output->atm.Nzcld = output->wc.threed.nlyr;

  /* finally determine the combined water and ice cloud 3D levels */
  output->atm.nthreed=0;
  output->atm.threed = calloc (output->atm.nlyr, sizeof(int));

  for (kc=0; kc<output->atm.nlyr; kc++) {
    output->atm.threed[kc] = 0;
    
    /* water cloud layers */
    if (output->wc.threed.nthreed > 0)
      output->atm.threed[kc] += output->wc.threed.threed[kc];

    /* ice cloud layers */
    if (output->ic.threed.nthreed > 0)
      output->atm.threed[kc] += output->ic.threed.threed[kc];

    if (output->atm.threed[kc] > 0)
      output->atm.nthreed++;
  }

  if (!input.quiet)
    fprintf (stderr, " ... found %d 3D layers (water and ice clouds combined) on a %d x %d x %dgrid\n", 
	     output->atm.nthreed, output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld);
  

  /* if no 3D cloud defined, copy sampling area size to cloud area size */
  if (output->atm.nthreed<1) {
    fprintf (stderr, " ... no 3D clouds defined; copying sample area size to cloud area size\n");
    
    output->atm.Nxcld    = 1;
    output->atm.Nycld    = 1;
    output->wc.threed.Nx = 1;
    output->wc.threed.Ny = 1;
    output->ic.threed.Nx = 1;
    output->ic.threed.Ny = 1;

    output->wc.threed.delX = (double) (output->mc.sample.Nx) * (output->mc.sample.delX);
    output->wc.threed.delY = (double) (output->mc.sample.Ny) * (output->mc.sample.delY);
    output->ic.threed.delX = (double) (output->mc.sample.Nx) * (output->mc.sample.delX);
    output->ic.threed.delY = (double) (output->mc.sample.Ny) * (output->mc.sample.delY);
  }
  
  /* check if cloud grid and sample grids are compatible */
  if (!equal ((double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
	      (double) output->wc.threed.Nx * output->wc.threed.delX)) {
    fprintf (stderr, "Error, x-size of sampling area (%.10e) does not equal x-size of cloud area (%.10e)\n",
	     (double) (output->mc.sample.Nx) * (output->mc.sample.delX), 
	     (double) output->wc.threed.Nx * output->wc.threed.delX);
    return -1;
  }

  if (!equal ((double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
	      (double) output->wc.threed.Ny * output->wc.threed.delY)) {
    fprintf (stderr, "Error, y-size of sampling area (%.10e) does not equal y-size of cloud area (%.10e)\n",
	     (double) (output->mc.sample.Ny) * (output->mc.sample.delY), 
	     (double) output->wc.threed.Ny * output->wc.threed.delY);
    return -1;
  }
  
  return 0;
}


/***********************************************************************************/
/* Function: compare_levels_df                                            @62_30i@ */
/* Description:                                                                    */
/*  Compare two altitude grids. First is expected to be double, second float       */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

/*
static int compare_levels_df (double *z1, int n1,
			      float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    return -1;
  }

  for (kc=0; kc<=n1; kc++)
    if (fabs((float) z1[kc] - z2[kc]) > MC_EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;
}
*/      




/* Read 3D water cloud data from file */

static int read_wc3D_file (char *filename, 
			   cld3d_out_struct *cld3d, 
			   input_struct input, output_struct *output)
{
  int status=0;

  void  F77_FCN (wcloud) (float *lambda_t, int *newsiz, int *nlyr,
			  char *filepath, int *nstring, float *lwc, float *wceffr, 
			  float *tmp_wc_dtau, float *tmp_wc_gg, float *tmp_wc_ssa, float *zd,
			  int *wclyr);

  /* read wc_file */
  status = read_3D_cloud (filename,
			  &(cld3d->Nx), 
			  &(cld3d->Ny), 
			  &(cld3d->nlyr), 
			  &(cld3d->delX), &(cld3d->delY),
			  &(cld3d->zd), 
			  &(cld3d->lwc),
			  &(cld3d->reff),
			  &(cld3d->ext),
			  &(cld3d->g1),
			  &(cld3d->g2),
			  &(cld3d->ff),
			  &(cld3d->ssa),
			  &(cld3d->Dscale),
			  &(cld3d->reffmin), &(cld3d->reffmax),
			  &(cld3d->cldproperties),
			  &(cld3d->threed),
			  input.quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  /* default: Henyey-Greenstein - flag needed by MYSTIC */
  output->wc.threed.nonHGcld=0;

  switch (input.wc.properties) {
  case WCPROP_HU:
  case WCPROP_ECHAM4:
    break;  

  case WCPROP_MIE:
  case WCPROP_FILE:

    if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
      
      /* need to read cloud property file only if not yet already done */
      if (!(output->wc.ssprop.alloc && output->wc.ssprop.alloc_moments)) {
	/* Read optical properties file */
	status = read_cloud_prop (input.filename[FN_WCLOUD_PROP], 
				  output->wl.lambda_t, output->wl.nlambda_t, 
				  output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
				  input.wc.interpolate, input.verbose, input.quiet,
				  &(output->wc), &(output->atm.nmom));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d reading water cloud property file %s\n", 
		   status, input.filename[FN_WCLOUD_PROP]);
	return status;
	}

	/* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	output->wc.threed.nonHGcld=1;
      }
      else {
	fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		 input.filename[FN_WCLOUD_PROP]);
      }
    }
    else {
      fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
	       input.filename[FN_WCLOUD_PROP]);
    }
    break;
    
  case WCPROP_PHASE:

    if (cld3d->cldproperties!=CLD_OPTPROP) { /* no need to read optical properies if HG anyway */
      
      /* need to read cloud property file only if not yet already done */
      if (!(output->wc.ssprop.alloc && output->wc.ssprop.alloc_explicit)) {
	/* read optical properties file */
	
	fprintf (stderr, "read_cloud_prop_phase (%s)\n", input.rte.mc.filename[FN_MC_PCLD]);
	
	status = read_cloud_prop_phase (input.rte.mc.filename[FN_MC_PCLD], 
					output->wl.lambda_t, output->wl.nlambda_t, 
					output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
					input.wc.interpolate, input.verbose, input.quiet,
					&(output->wc), &(output->atm.nmom));
	
	if (status!=0) {
	  fprintf (stderr, "Error %d reading water cloud property file %s\n", 
		   status, input.rte.mc.filename[FN_MC_PCLD]);
	  return status;
	}

	/* set flag for non-Henyey-Greenstein, that is, tabulated phase function */
	output->wc.threed.nonHGcld=1;
      }
      else {
	fprintf (stderr, " ... no need to read cloud property file %s again\n", 
		 input.rte.mc.filename[FN_MC_PCLD]);
      }
    }
    else {
      fprintf (stderr, " ... ignoring cloud property file %s, using Henyey-Greenstein\n", 
	       input.filename[FN_WCLOUD_PROP]);
    }
    
    break;

  default:
    fprintf (stderr, "Error, unknown wc_properties %d. This is not\n", input.wc.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }

  return 0;
}



/* Convert 1D and 3D water cloud data to MYSTIC optical properties for a given wavelength */

int convert_wc_mystic (input_struct input, output_struct *output,
		       cld3d_out_struct *wcld3d, int iv)
{
  int status=0;
  int nstring = (int) strlen(input.filename[FN_PATH]);

  float lwc [2] = {0,0};
  float reff[2] = {0,0};
  float dtau[2] = {0,0}, g1[2]={0,0}, ssa[2]={0,0};
  float zd  [2] = {1,0};
  float rmin=0, rmax=0;
  int nlyr=1, nlev=2;
  int newsiz=1;
  int wclyr=1;
  int ix=0, iy=0, iz=0;

  int i1=0;
  double extinc=0, omega=0, dscale=0;
  void  F77_FCN (wcloud) (float *lambda_t, int *newsiz, int *nlyr,
			  char *filepath, int *nstring, float *lwc, float *wceffr, 
			  float *tmp_wc_dtau, float *tmp_wc_gg, float *tmp_wc_ssa, float *zd,
			  int *wclyr);

  switch (wcld3d->cldproperties) {
  case CLD_OPTPROP:

    /* ?????????????????????????????????????????????????????????????????????????????? */
    /* currently we have a special treatment for explicitely defined phase functions; */
    /* this is subject to change, as we want to make the explicit phase function      */
    /* compatible to WCPROP_MIE/WCPROP_FILE; that is, specify phase function in the   */
    /* same format as in the momcld_files for different effective radii plus          */
    /* extinction coefficient and single scattering albedo                            */

    if (input.wc.properties == WCPROP_PHASE) {

      rmin = 10; rmax = 10; /* doesn't matter anyway */

      /* generate pgase function table */
      status = setup_Legendre_table (&(wcld3d->phase_cld), 
				     &(output->wc.ssprop),
				     iv,
				     rmin, rmax, 
				     input.rte.mc.delta_scaling_mucut, 
				     input.rte.mc.truncate, 
				     input.quiet);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d returned by setup_Legendre_table()\n", 
		 status);
	return status;
      }
    }

    break;

  case CLD_EXTREFF:
  case CLD_LWCREFF:

    /* first create phase function table */
    switch (input.wc.properties) {
    case WCPROP_HU:
    case WCPROP_ECHAM4:
      break;  
      
    case WCPROP_MIE:
    case WCPROP_FILE:
    case WCPROP_PHASE:
      
      /* combine 1D and 3D rmin/rmax */
      rmax = (wcld3d->reffmax > output->wc.microphys.effrmax ? wcld3d->reffmax : output->wc.microphys.effrmax);

      /* take the smaller of both which is larger than zero */
      if (wcld3d->reffmin > 0) {
	if (output->wc.microphys.effrmin > 0)
	  rmin = (wcld3d->reffmin < output->wc.microphys.effrmin ? wcld3d->reffmin : output->wc.microphys.effrmin);
	else
	  rmin = wcld3d->reffmin;
      }
      else 
	rmin = output->wc.microphys.effrmin;
      
      if (rmin<=0) /* no need to do anything */
	break;
      
      fprintf (stderr, " ... generating phase function tables between rmin=%f and rmax=%f\n", rmin, rmax);

      /* generate pgase function table */
      status = setup_Legendre_table (&(wcld3d->phase_cld), 
				     &(output->wc.ssprop),
				     iv,
				     rmin, rmax, 
				     input.rte.mc.delta_scaling_mucut, 
				     input.rte.mc.truncate, 
				     input.quiet);/*TZ ds*/
      if (status!=0) {
	fprintf (stderr, "Error %d returned by setup_Legendre_table()\n", 
		 status);
	return status;
      }
      break;
      
    default:
      fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }

    for (iz=0; iz<wcld3d->nlyr; iz++)
      if (wcld3d->nthreed>0)
	if (wcld3d->threed[iz])  /* if a 3D layer */
	  for (ix=0; ix<wcld3d->Nx; ix++)
	    for (iy=0; iy<wcld3d->Ny; iy++) {
	      
	      if ((wcld3d->cldproperties==CLD_EXTREFF && wcld3d->ext[iz][ix][iy]>0) ||
		  (wcld3d->cldproperties==CLD_LWCREFF && wcld3d->lwc[iz][ix][iy]>0)) {
		
		/* use dummy LWC - it doesn't matter anyway */
		if (wcld3d->cldproperties==CLD_EXTREFF) {
		  lwc[0] = 1.0; 
		  lwc[1] = 1.0;
		}
		
		if (wcld3d->cldproperties==CLD_LWCREFF) {
		  lwc [0] = wcld3d->lwc[iz][ix][iy];
		  lwc [1] = wcld3d->lwc[iz][ix][iy];
		}
		
		reff[0] = wcld3d->reff[iz][ix][iy];
		reff[1] = wcld3d->reff[iz][ix][iy];

		switch (input.wc.properties) {
		case WCPROP_HU:
		
		  switch (input.ck_scheme) {
		  case CK_FU:
		  
		    /* call Fu and Liou code by Fred Rose */
		  
#if HAVE_FULIOU
		    /*
		      status = ckdfucld (zd, 
		      reff, 
		      lwc, 
		      nlev,
		      dtau, g1, ssa);
		    */
		  
		    /* currently not supported because ckdfucld loops over */
		    /* all 18 bands which is a waste of time and should be */
		    /* changed                                             */
		    fprintf (stderr, "Error, correlated_k fu in combination with wc_properties fu\n");
		    fprintf (stderr, "is currently not supported!\n"); 
		  
#else
		    fprintf (stderr, "Error, Fu and Liou not supported!\n");
		    return -1;
#endif
		  
		    if (status!=0) {
		      fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
		      return status;
		    }
		  
		    break;
		  
		  default:   /* old uvspec wcloud.f for Hu and Stamnes */
		  
		    /* wcloud calculates lwc_layer and reff_layer internally */
		    F77_FCN (wcloud) (&output->wl.lambda_t[iv], &newsiz, &nlyr,
				      input.filename[FN_PATH], &nstring, 
				      lwc, reff, 
				      dtau, g1, ssa, 
				      zd, &wclyr);
		  
		    break;
		  }
	    
		  /* write results to final destination */
		  wcld3d->ssa[iz][ix][iy] = ssa[0];
		  wcld3d->g1 [iz][ix][iy] = g1 [0];
		  wcld3d->g2 [iz][ix][iy] = 0;
		  wcld3d->ff [iz][ix][iy] = 1.0;
		
		  /*
		  fprintf (stderr, "WCLOUD3D %d %d %d  %f %f %f %f\n",
			   iz, ix, iy,
			   wcld3d->ssa[iz][ix][iy],
			   wcld3d->g1 [iz][ix][iy],
			   wcld3d->g2 [iz][ix][iy],
			   wcld3d->ff [iz][ix][iy]);
		  */
 
		  if (wcld3d->cldproperties==CLD_LWCREFF)
		    wcld3d->ext[iz][ix][iy] = dtau[0] / 1000.0; 
		
		  /* need to convert again from 1/km to 1/m */		
		  /* because wcloud() thinks in km rather   */
		  /* than m                                 */

		  break;
		
		case WCPROP_ECHAM4:
		
		  status = wc_echam4 (output->wl.lambda_t[iv], nlev-1,
				      lwc, reff, 
				      dtau, g1, ssa,
				      zd, wclyr);
		
		  if (status!=0) {
		    fprintf (stderr, "Error %d returned by wc_echam4()\n", status);
		    return status;
		  }

		  /* write results to final destination */
		  wcld3d->ssa[iz][ix][iy] = ssa[0];
		  wcld3d->g1 [iz][ix][iy] = g1 [0];
		  wcld3d->g2 [iz][ix][iy] = 0;
		  wcld3d->ff [iz][ix][iy] = 1.0;
		
		  if (wcld3d->cldproperties==CLD_LWCREFF)
		    wcld3d->ext[iz][ix][iy] = dtau[0] / 1000.0;
		
		  /* need to convert again from 1/km to 1/m */		
		  /* because wcloud() thinks in km rather   */
		  /* than m                                 */

		  break;  
		
		case WCPROP_MIE:
		case WCPROP_FILE:
		case WCPROP_PHASE:

		  /* interpolate extinction and absorption coefficient */
		  /* of each box to given effective radius             */
		  i1 = (int) ((wcld3d->reff[iz][ix][iy] - wcld3d->phase_cld.r0) / wcld3d->phase_cld.dr);
		
		  /* careful with rounding errors at the array boundaries */
		  i1 = (i1>wcld3d->phase_cld.n-1?wcld3d->phase_cld.n-1:i1);
		  i1 = (i1<0?0:i1);
	      
		  extinc  = wcld3d->phase_cld.ext[i1] + wcld3d->phase_cld.dext[i1] * 
		    (wcld3d->reff[iz][ix][iy] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr));
		
		  omega   = wcld3d->phase_cld.omega[i1] + wcld3d->phase_cld.domega[i1] * 
		    (wcld3d->reff[iz][ix][iy] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr));
		
		  dscale  = wcld3d->phase_cld.Dscale[i1] + wcld3d->phase_cld.dDscale[i1] * 
		    (wcld3d->reff[iz][ix][iy] - (wcld3d->phase_cld.r0 + (double) i1 * wcld3d->phase_cld.dr));
		
		  /* extinction coefficient */     
		  if (wcld3d->cldproperties==CLD_LWCREFF)
		    wcld3d->ext[iz][ix][iy] = wcld3d->lwc[iz][ix][iy] * extinc / 1000.0;
		 
		  /* single scattering albedo */
		  wcld3d->ssa[iz][ix][iy] = omega;
		
		  /* delta-scaling factor */
		  wcld3d->Dscale[iz][ix][iy] = dscale;
		
		  fprintf (stderr, "DSCALE %f  %d %f\n", dscale, i1, wcld3d->phase_cld.Dscale[i1]);


		  /* ???? Attention: need to calculate the delta-scaling factor also in the 1D case ???? */ 
		  break;
		
		default:
		  fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
		  fprintf (stderr, "supposed to happen and indicates a coding error.\n");
		
		  return -1;
		  break;
		}
	    }
	  }
    break;

  default:
    fprintf (stderr, "Error, unknown cloud properties %d\n", wcld3d->cldproperties);
    return -1;
  }


  return 0;
}
  

/* Free 3D water cloud optical properties */

int free_wc_mystic (input_struct input, cld3d_out_struct *wcld3d)
{
  switch (wcld3d->cldproperties) {
  case CLD_OPTPROP:
    break;
    
  case CLD_EXTREFF:
  case CLD_LWCREFF:

    /* first create phase function table */
    switch (input.wc.properties) {
    case WCPROP_HU:
    case WCPROP_ECHAM4:
      break;  
      
    case WCPROP_MIE:
    case WCPROP_FILE:
    case WCPROP_PHASE:
      
      free_phase_function_table (&(wcld3d->phase_cld));
      break;
      
    default:
      fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
      fprintf (stderr, "supposed to happen and indicates a coding error.\n");
      
      return -1;
      break;
    }
  }

  return 0;
}
  
