#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <getopt.h>

#include "ascii.h"
#include "numeric.h"


#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


#define PROGRAM "pmom"
#define VERSION "0.1"

/*************************************************************/
/* print usage information                                   */
/*************************************************************/


static void print_usage (char *filename)
{
  fprintf (stderr, "%s %s - Calculate Legendre moments\n\n", PROGRAM, VERSION);
  fprintf (stderr, "written by Claudia Emde,\n");
  fprintf (stderr, "           DLR, e-Mail claudia.emde@dlr.de\n");
  fprintf (stderr, "Version %s finished January 26, 2006\n\n", VERSION);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "-h            display help message\n");
  fprintf (stderr, "-l <number>   number of legendre moments\n");
  fprintf (stderr, "-r <grid no>  scattering angle grid resolution (1-5) \n");
  fprintf (stderr, "-c            calculate coeffcients instead of polynomials,\n");
  fprintf (stderr, "              required to generate input files for uvspec.\n");
  fprintf (stderr, "-n            normalize phase function \n\n");
 
  fprintf (stderr, "USAGE: %s [options] <filename>\n\n", filename);
  fprintf (stderr, "%s calculates the Legendre moments of the phase function.\n\n", PROGRAM);
  fprintf (stderr, "The input must be provided as 2-column file, containing \n");
  fprintf (stderr, "the scattering angle grid in the first column and the \n");
  fprintf (stderr, "phase function value in the second column.\n");
  fprintf (stderr, "Output are the Legendre moments. \n\n");
  fprintf (stderr, "You may specify the number of moments using the option -l.\n\n");
  fprintf (stderr, "Different scattering angle grid resolutions can be \n");
  fprintf (stderr, "chosen. For moderate forward peaks, the standard grid \n");
  fprintf (stderr, "  (1 - equidistant, 0.01 degrees step width) \n");
  fprintf (stderr, "should be sufficient. \n");
  fprintf (stderr, "For phase functions with a very strong forward peak, \n");
  fprintf (stderr, "e.g. ice particle phase functions, the finest grid resolution \n");
  fprintf (stderr, "  (2 - equidistant, 0.001 degrees step width) \n");
  fprintf (stderr, "should be specified. \n\n");
  fprintf (stderr, "Grid 3 uses the scattering angle grid provided by the \n");
  fprintf (stderr, "input file. Grids 4 and 5 have a finer resolution around\n");
  fprintf (stderr, "the forward peak. \n \n");
  fprintf (stderr, "The accuracy of the Legendre decomposition may be tested \n");
  fprintf (stderr, "using the tool phase: \n");
  fprintf (stderr, "phase -c -d -s 1 pmom_outfile.dat \n"); 
  fprintf (stderr, "\n");
}

static int get_options (int argc, char **argv,
                        char *programname,
                        char *infilename,
                        int *n_legendre,
                        int *grid_res, 
                        int *coeff,
                        int *norm)
{
  int c=0;
  char *dummy=NULL;
  
  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);
  
  /* set defaults */
  strcpy (infilename, "");
  *n_legendre = 0;
  *grid_res = 1;
  *norm = 0;

  while ((c=getopt (argc, argv, "hl:r:cn")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'l':  /* number of legendre polynomials */
      *n_legendre = strtol (optarg, &dummy, 10);
      break;
    case 'r':
      *grid_res = strtol (optarg, &dummy, 10);
      break;
    case 'c':  /*Legendre coefficients */
      *coeff=1;
      break;
    case 'n':
      *norm=1;
      break; 
    default:
      print_usage (programname);
      return (-1);
    }
  }

  /* check number of remaining command line arguments */
  if (argc - optind != 1)  {
    print_usage (programname);
    return -1;
  }
  
  strncpy (infilename, argv[optind+0], FILENAME_MAX);
  
  return 0;  /* if o.k. */
}


int main(int argc, char **argv)
{
  int status=0;
  char programname[FILENAME_MAX], infilename[FILENAME_MAX]="";
  
  int n_legendre=0;
  int grid_res=0;
  int coeff=0; 
  int norm=0;
  double *u=NULL;
  double *Xi=NULL;
  double pint=0; 
  double *ph_int=NULL; 
  double p0_1=0, p0_2=0, p1_1=0, p1_2=0, p2_1=0, p2_2=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *f=NULL;
  double *s=NULL;
  int L=0;
  int l=0, i=0, N_t=0;
  double *theta=NULL;

  status = get_options (argc, argv, programname, infilename, &n_legendre,
                        &grid_res, &coeff, &norm);

  if (status!=0)
    return status;
  
    
  /* Read phase function data */
  read_2c_file (infilename, &f, &s, &L); 
  fprintf (stderr, " ... read %d data points from %s\n", L, infilename);
  
  if (status!=0) {
    fprintf (stderr, "error %d reading %s\n", status, infilename);
    return status;
  }
  
  fprintf (stderr, "... calculate %d legendre polynomials\n", n_legendre);
  
  switch(grid_res)
    {
    case 1:
      fprintf (stderr, "... use grid resolution %d\n", grid_res);
      N_t = 18000;
      theta = (double*) malloc((N_t+1) * sizeof(f[0])); 
      theta[0] = 0.0;
      for (i=1; i <= N_t; i++)
        theta[i] = theta[i-1] + 0.01;
      break;

    case 2:
      fprintf (stderr, "... use grid resolution %d.\n", grid_res);
      N_t = 180000;
      theta = (double*) malloc((N_t+1) * sizeof(f[0]));
      theta[0] = 0.0;
      for (i=1; i <= N_t; i++)
        theta[i] = theta[i-1] + 0.001;
      break; 
     
    case 3:
      fprintf (stderr, "... use grid resolution %d.\n", grid_res);
      fprintf (stderr, "warning: this is probably inaccurate.\n");
      theta = (double*) malloc(L*sizeof(f[0]));
      N_t = L-1;
      for (i=0; i<L; i++)
        theta[i] = f[i];
      break;
      
    case 4:
      fprintf (stderr, "... use grid resolution %d\n", grid_res);
      N_t = 27000; 
      theta = (double*) malloc((N_t+1) * sizeof(f[0]));
      theta[0] = 0.0;
      for (i=1; i <= 10000; i++)
        theta[i] = theta[i-1] + 0.001;
      for (i=10001; i <= N_t; i++)
        theta[i] = theta[i-1] + 0.01;
      break;
      
    case 5:
      fprintf (stderr, "... use grid resolution %d.\n", grid_res);
      fprintf (stderr, "warning: this is probably inaccurate.\n");
      N_t = 2700; 
      theta = (double*) malloc((N_t+1) * sizeof(f[0]));
      theta[0] = 0.0;
      for (i=1; i <= 1000; i++)
        theta[i] = theta[i-1] + 0.01;
      for (i=1001; i <= N_t; i++)
        theta[i] = theta[i-1] + 0.1;
      break;
      
    default: 
      fprintf (stderr, "Grid resolution not valid. \n");
      fprintf (stderr, "Use resolution 1 \n");
      N_t = 180000;
      theta = (double*) malloc((N_t+1) * sizeof(f[0]));
      theta[0] = 0.0;
      for (i=1; i <= N_t; i++)
        theta[i] = theta[i-1] + 0.001;
      break;
      
    }               

  u  = calloc (N_t+1, sizeof(double));
  Xi = calloc (n_legendre, sizeof(double));
  ph_int = calloc (N_t+1, sizeof(double));

  status = linear_coeffc (f, s, L, &a0, &a1, &a2, &a3);
  if (status!=0)  {
    fprintf (stderr, "Sorry cannot do linear interpolation\n");
    fprintf (stderr, "linear_coeffc() returned status %d\n", status);
    return status;
  }

  
  if(grid_res!=3)
    {
      for (i = 0; i<=N_t; i++)
        {
        status = calc_splined_value (theta[i], &(ph_int[i]), f, L, a0, a1, a2, a3);
        /*ph_int[i] = gsl_interp_eval(interp, f, s, theta[i], acc);
          fprintf(stdout, "%f %g \n", theta[i], ph_int[i]); */
        }
    }

 
  
  free(a0); free(a1); free(a2); free(a3); 


  for (l=0; l<=N_t; l++)
    u[l] = cos(theta[l]*PI/180.);

  /*Check if the phase function is normalized to 2.0 If not quit. */
  for (l=0; l<N_t; l++)
    pint += 0.5*(ph_int[l]+ph_int[l+1])*fabs(u[l+1] - u[l]);
  
  if (fabs(2.0 - pint) > 1e-2) {
    fprintf (stderr, "*** Warning, phase function not normalized to 2 but to %.8f\n", pint);
    fprintf (stderr, "*** Use option -n to normalize it.\n");
  }
  
  if(norm){
    for (l=0; l<N_t; l++)
      ph_int[l] *= 2/pint;
  }
  
  pint=0.0;
  /* Check if the phase function is normalized to 2.0 If not quit. */
  for (l=0; l<N_t; l++)
    pint += 0.5*(ph_int[l]+ph_int[l+1])*fabs(u[l+1] - u[l]);
  
  if (fabs(2.0 - pint) > 1e-2)
    fprintf (stderr, "*** Warning, phase function not normalized to 2 but to %.8f\n", pint);
  

  for (i=0; i<N_t;i++) {
    p0_1 = 1.0;
    p0_2 = 1.0;
    
    Xi[0]+=0.5*0.5*(ph_int[i]+ph_int[i+1])*fabs(u[i+1]-u[i]); 
    
    p1_1=u[i];
    p1_2=u[i+1];
    
    if (coeff)
      Xi[1]+=0.5*0.5*(p1_1*ph_int[i]+p1_2*ph_int[i+1])*fabs(u[i+1]-u[i])*3; /* Legendre coefficients *(2l+1)=3*/
    else
      Xi[1]+=0.5*0.5*(p1_1*ph_int[i]+p1_2*ph_int[i+1])*fabs(u[i+1]-u[i]);  
    
    for (l=2; l<n_legendre; l++) {
      p2_1=(2*(double)l-1)/(double)l*u[i]*p1_1-((double)l-1)/(double)l*p0_1; 
      p2_2=(2*(double)l-1)/(double)l*u[i+1]*p1_2-((double)l-1)/(double)l*p0_2;
      
      if (coeff)
	Xi[l]+=0.5*0.5*(p2_1*ph_int[i]+p2_2*ph_int[i+1])*fabs(u[i+1]-u[i])
	  *(2*(double)l + 1); /* Factor to obtain Legendre coefficients,*/ 
      else
	Xi[l]+=0.5*0.5*(p2_1*ph_int[i]+p2_2*ph_int[i+1])*fabs(u[i+1]-u[i]);
      
      p0_1 = p1_1;
      p0_2 = p1_2;
      p1_1 = p2_1;
      p1_2 = p2_2;
    }
  }
  
  /* Print out the Legendre coefficients */
  for (l=0; l<n_legendre; l++) {
    if (coeff)
      fprintf(stdout,  "%g " , Xi[l] ); 
    else
      fprintf(stdout,  "%g \n " , Xi[l]);
  }
  

  free (u);
  free (Xi);
  free (ph_int);

  return 0;
}  
