/*--------------------------------------------------------------------
 * $Id: solve_rte.c,v 1.157 2008/01/09 15:21:10 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <math.h>
#include <string.h>
#include <time.h>

#include "uvspec.h"
#include "ascii.h"
#include "numeric.h"
#include "fortran_and_c.h"
#include "cloud.h"
/* #include "../src_c/nrutil.h" */

#if HAVE_MYSTIC
#include "mystic.h"
#endif

#if HAVE_SOS
#include "sos.h"
#endif
#include "f77-uscore.h"
#include "solver.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#if HAVE_LIBGSL 
#include <gsl/gsl_math.h>
#include <gsl/gsl_diff.h>
#endif

/* Definitions for numerical recipes functions */
#define NRANSI
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/* internal structures */
typedef struct {
  char    *deltam;
  char    *ground_type;
  char    polscat[15];
  pol_complex ground_index;
  double  albedo;
  double  btemp;
  double  flux;
  double  *gas_extinct;
  double  *height;
  double  mu;
  double  sky_temp;
  double  *temperatures;
  double  wavelength;
  int     *outlevels;
  int     nummu; /* Number of quadrature angles (per hemisphere). */
} polradtran_input;


typedef struct {
  char header[127]; /* A 127- (or less) character header for prints         */

  float accur;      /* Convergence criterion for azimuthal series.          */

  float fbeam;      /* Intensity of incident parallel beam at top boundary. */
                    /*  [same units as PLKAVG (default W/sq m) if thermal   */
                    /*  sources active, otherwise arbitrary units].         */

  float fisot;      /* Intensity of top-boundary isotropic illumination.    */
                    /*  [same units as PLKAVG (default W/sq m) if thermal   */
                    /*  sources active, otherwise arbitrary units].         */
                    /*  Corresponding incident flux is pi (3.14159...)      */
                    /*  times FISOT.                                        */

  float *hl;        /* K = 0 to NSTR.  Coefficients in Legendre-            */
                    /* polynomial expansion of bottom bidirectional         */
                    /* reflectivity                                         */

  int   planck;     /* TRUE,  include thermal emission                      */
                    /* FALSE, ignore thermal emission (saves computer time) */

  float btemp;      /* Temperature of bottom boundary (K)                   */
                    /* (bottom emissivity is calculated from ALBEDO or HL,  */
                    /* so it need not be specified).                        */
                    /* Needed only if PLANK is TRUE.                        */

  float ttemp;      /* Temperature of top boundary (K)                      */
                    /* Needed only if PLANK is TRUE.                        */

  float temis;      /* Emissivity of top boundary.                          */
                    /* Needed only if PLANK is TRUE.                        */

  float *utau;
  float umu0 ;

  int ierror_d[47];
  int ierror_s[33];
  int ierror_t[22];

  int prndis[7];
  int prndis2[5];
  int prntwo[2];

  int ibcnd;        /* 0 : General case.                                    */
                    /* 1 : Return only albedo and transmissivity of the     */
                    /*     entire medium vs. incident beam angle.           */

  int lamber;       /* TRUE, isotropically reflecting bottom boundary.      */
                    /* FALSE, bidirectionally reflecting bottom boundary.   */

  int onlyfl;       /* TRUE, return fluxes, flux divergences, and mean      */
		    /*       intensities.                                   */
                    /* FALSE, return fluxes, flux divergences, mean         */
                    /*        intensities, azimuthally averaged intensities */
                    /*        (at the user angles) AND intensities.         */


  int quiet;
  int usrang;
  int usrtau;

  /* sdisort-specific variables */ 
  int nil;    
  int newgeo; 
  int spher;

  /* qdisort-specific variables */ 
  int gsrc;         /* Flag for general source for qdisort */
  float ***qsrc;  /* The general source, in FORTRAN: REAL*4 qsrc( MXCMU, 0:MXULV, MXPHI )*/

  /* PolRadtran-specific variables */ 
  polradtran_input pol;

} rte_input;



typedef struct {
  float *dfdt;
  float *flup;
  float *rfldir;
  float *rfldn;
  float *uavgso;
  float *uavgdn;
  float *uavgup;
  float *uavg;
  float *heat;
  float *w_zout;
  float **u0u;
  float ***uu;
  	 
  float ***rfldir3d;
  float ***rfldn3d;
  float ***flup3d;
  float ***uavgso3d;
  float ***uavgdn3d;
  float ***uavgup3d;
  float ***abs3d;
  float ***radiance3d;

  float *albmed;
  float *trnmed;

  double **polradtran_down_flux;
  double **polradtran_up_flux;
  double ****polradtran_down_rad_q; /* _q indicates radiances at quadrature angels. */
  double ****polradtran_up_rad_q;
  double ****polradtran_down_rad;
  double ****polradtran_up_rad;
  double *polradtran_mu_values;

} rte_output;


typedef struct{
  float *tauw;
  float *taui;

  float *g1d;
  float *g2d; 
  float *fd; 
  float *g1i;
  float *g2i; 
  float *fi;

  float *ssaw;
  float *ssai;
} save_optprop;
  



/* prototypes of internal functions */
static int reverse_profiles (input_struct input, output_struct *output);
static rte_output *calloc_rte_output (input_struct input, int nzout, int Nxcld, int Nycld, int Nzcld, int *threed);
static int reset_rte_output (rte_output **rte, input_struct input, int nzout, int Nxcld, int Nycld, int Nzcld, int *threed);
static void free_rte_output (rte_output *result, input_struct input, int nzout,
			     int Nxcld, int Nycld, int Nzcld, int *threed);
static int add_rte_output (rte_output *rte, rte_output *add, double factor, input_struct input, 
			   int nzout, int Nxcld, int Nycld, int Nzcld, int *threed);
static int init_rte_input (rte_input *rte, input_struct input, output_struct *output);
static int setup_and_call_solver (input_struct input, output_struct *output, 
				  rte_input *rte_in, rte_output *rte_out, 
				  int iv, int ib, int *threed, int mc_loaddata);
static int call_solver (input_struct input, output_struct *output,
			rte_input *rte_in, int iv, rte_output *rte_out,
			int *threed, int mc_loaddata, int verbose);
static void fourier2azimuth (double**** down_rad_rt3, double**** up_rad_rt3,
			     double**** down_rad, double**** up_rad, 
			     int nzout, int aziorder,
			     int nstr, int numu, int nstokes,
			     int nphi, float* phi);

static int cp_3d_to_rte_output (rte_output *rte, 
				float *rfldir3d, float *rfldn3d,  float *flup3d, 
 				float *uavgso3d, float *uavgdn3d, float *uavgup3d, 
				float *radiance3d,
				int Nz, int Nx, int Ny);

static int calc_spectral_heating (input_struct input, output_struct *output, float *dz, 
                                  double *p_rho_mass_zout, float *p_k_abs, float *p_k_abs_layer, int *zout_index,
                                  rte_output *rte_out, float *p_heat, float *p_w_zout, int iv);

static float ***calloc_abs3d (int Nx, int Ny, int Nz, int *threed);
static void free_abs3d (float ***abs3d, int Nx, int Ny, int Nz, int *threed); 

static float ****calloc_spectral_abs3d (int Nx, int Ny, int Nz, int nlambda, int *threed);

double get_unit_factor(input_struct input, output_struct *output, int iv);


static int generate_effective_cloud (input_struct input, output_struct *output, save_optprop *save_cloud, 
                                     int iv, int iq, int verbose);

static save_optprop *calloc_save_optprop (int Nlev);

void F77_FCN (swde)(float *g_scaled, float *pref, float *prmuz, float *tau, float *ssa_gas, 
		    float *pre1, float *pre2, float *ptr1, float *ptr2);   

float zbrent_taueff(float mu_eff, float g_scaled, float ssa_gas, 
                    float transmission_cloud,  float transmission_layer,
                    float x1, float x2, float tol);

/*********************************************************************/
/* Main function. Loop over wavelengths or wavelength bands,         */
/* correlated-k quadrature points, and independent pixels.           */
/*********************************************************************/

int solve_rte (input_struct input, output_struct *output)
{
  static int first=1;
  int status=0;
  int ipa=0, is=0, js=0, ks=0, iu=0, iv=0, j=0, lu=0;
  int iq=0, lc=0;
  float *dz=NULL;
  double *rho_mass_zout = NULL;
  float *k_abs = NULL; 
  float *k_abs_layer = NULL;
  float *k_abs_outband = NULL; 
  int mc_loaddata=1;
  double unit_factor=0; 
  double ffactor=0.0, rfactor=0.0, hfactor=1.0;
  /* float heating_rate_emission = 0.0; */

  rte_input  *rte_in      = NULL;
  rte_output *rte_out     = NULL;
  rte_output *rte_outband = NULL;

  save_optprop *save_cloud = NULL;
 
  char function_name[]="solve_rte";
  char file_name[]="solve_rte.c";

  if (first) {
    first=0;

    rte_in      = calloc(1, sizeof(rte_input));

    rte_out     = calloc_rte_output(input, output->atm.nzout,  
				    output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, 
				    output->atm.threed);

    rte_outband = calloc_rte_output(input, output->atm.nzout, 
				    output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, 
				    output->atm.threed);

    save_cloud = calloc_save_optprop (output->atm.nlev);
    

    /* initialize RTE input */
    status = init_rte_input (rte_in, input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d initializing rte input in %s (%s)\n", status, function_name, file_name);
      return status;
    }

    /* allocate memory for the output result structures */
    status = setup_result (input, output, &dz, &rho_mass_zout ,&k_abs_layer, &k_abs, &k_abs_outband);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for the model output in %s (%s)\n", status, function_name, file_name);
      return status;
    }
  }

  /************************/
  /* loop over wavelength */
  /************************/
  for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {

    /* solar zenith angle at this wavelength */
    rte_in->umu0 = cos(output->atm.sza_t[iv]*PI/180.0);

    /* calculate 3D cloud properties for this wavelength */
    if (input.rte.solver == MONTECARLO && (output->wc.occurrence1D || output->wc.threed.nthreed>0)) {
      status = convert_wc_mystic (input, output, &(output->wc.threed), iv);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by convert_wc_mystic() in %s (%s)\n", status, function_name, file_name);
	return status;
      }
    }

    /********************************/
    /* loop over independent pixels */
    /********************************/

    for (ipa=0; ipa<output->nipa; ipa++) {

      if (input.verbose && output->nipa == 1)
        fprintf (stderr, "\n\n*** wavelength: iv = %d, %f nm, albedo = %f \n", iv, output->wl.lambda_t[iv], output->alb.albedo_t[iv]); 

      if (input.verbose && output->nipa > 1)
        fprintf (stderr, "\n\n*** wavelength: iv = %d, %f nm, looking at column %d, albedo = %f\n", iv, output->wl.lambda_t[iv], ipa, output->alb.albedo_t[iv]);  


      /* copy pixel number ipa to 1D cloud data */
      if (input.wcipa) {

	/* copy everything except the single scattering properties */
	status = cp_cld_out (&(output->wc), output->wcipa[ipa], 0, 0, input.quiet);
	if (status!=0) {
	  fprintf (stderr, "Error %d copying output->wcipa[ipa] to output->wc\n", status);
	  return status;
	}

	switch (input.wc.properties) {
	case WCPROP_MIE:
	case WCPROP_FILE:
	case WCPROP_PHASE:
	  
	  /* recalculate the optical properties for the */
	  /* new microphysics profile                   */
	  status = ssprop2optprop (&(output->wc), iv);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d calculating water cloud optical properties\n", status);
	    return status;
	  }
	  break;

	case WCPROP_NONE:
	case WCPROP_HU:
	case WCPROP_ECHAM4:
	case WCPROP_EXPLICIT:
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
	  fprintf (stderr, "supposed to happen and indicates a programm error.\n");
	  
	  return -1;
	  break;
	}

	/* overwrite these properties with user-defined optical thickness, ssa, etc */
	status = apply_user_defined_properties_to_cld (input.wc, 
						       output->atm, 
						       output->wl, 
						       output->alt.altitude,
						       &(output->wc));
	if (status!=0) {
	  fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
		   status);
	  return status;
	}
      }

      if (input.icipa) {

	/* copy everything except the single scattering properties */
	status = cp_cld_out (&(output->ic), output->icipa[ipa], 0, 0, input.quiet);
	if (status!=0) {
	  fprintf (stderr, "Error %d copying output->icipa[ipa] to output->ic in %s (%s)\n", status, function_name, file_name);
	  return status;
	}

	switch (input.ic.properties) {
	case ICPROP_MIE:
	case ICPROP_FILE:

	  /* recalculate the optical properties for the */
	  /* new microphysics profile                    */
	  status = ssprop2optprop (&(output->ic), iv);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d calculating ice cloud optical properties in %s (%s)\n", status, function_name, file_name);
	    return status;
	  }

	case ICPROP_NONE:
	case ICPROP_FU:
	case ICPROP_ECHAM4:
	case ICPROP_KEY:
	case ICPROP_YANG:
	case ICPROP_EXPLICIT:
	  break;

	default:
	  fprintf (stderr, "Error, unknown ice cloud property %d. This is not\n", input.ic.properties);
	  fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	  
	  return -1;
	  break;
	}

	/* overwrite these properties with user-defined optical thickness, ssa, etc */
	status = apply_user_defined_properties_to_cld (input.ic, 
						       output->atm, 
						       output->wl, 
						       output->alt.altitude,
						       &(output->ic));
	if (status!=0) {
	  fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
		   status);
	  return status;
	}
      }


      /* IPA molecular absorption and aerosols (saturation inside clouds) */

      /* these lines also optimise the iq-loop for corr-k schemes, also if there is no ipa */
      switch(input.ck_scheme) {
        case CK_NONE:
        case CK_KATO:
        case CK_KATO2:
        case CK_KATO2_96:
          break;
      
        case CK_FU:
        case CK_AVHRR_KRATZ:
        case CK_LOWTRAN:
        case CK_GENERIC:
          if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
              (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
            if (input.ck_scheme == CK_LOWTRAN)
              output->atm.nq_t[iv] = output->crs_ck.ipa_profile[ipa][0][iv].ngauss;
            for (iq=0; iq<output->atm.nq_t[iv]; iq++)
              output->atm.wght_t[iv][iq] = output->atm.ipa_wght_t[ipa][iv][iq];
	  }
          else
            if (input.ck_scheme == CK_LOWTRAN)
              output->atm.nq_t[iv] = output->crs_ck.profile[0][iv].ngauss;       
            break;
        default:
          fprintf (stderr, "Error: unimplemented correlated-k scheme %d in %s (%s)\n", input.ck_scheme, function_name, file_name);
          return -1;
          break;
      }


      /* If only one subband is used in the LOWTRAN parameterization, */
      /* the photon weights of the three subbands are added; this is  */
      /* necessary because the number of subbands changes with        */
      /* concentration and is therefore not known beforehand.         */
      /* The correct use of mc_photons_file for LOWTRAN is then       */
      /* to always distribute the photons over three subbands;        */
      /* uvspec decides automatically if only one is needed           */      

      if (input.ck_scheme==CK_LOWTRAN) {
	if (output->atm.nq_t[iv] == 1)
	  for (iq=1; iq<LOWTRAN_MAXINT; iq++)
	    output->mc_photons_t[iv][0] += output->mc_photons_t[iv][iq];
	/* fprintf (stderr, "mc_photons = %f\n", output->mc_photons_t[iv][0]); */
      }
      

      if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
          (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {

        if (input.verbose) 
          fprintf (stderr, "... modify mol. absorption and aerosol optical properties according to column %d data\n", ipa);

        /* absorption of cloud column also changed due to saturation inside the cloud */
        for (lu=0; lu<output->atm.nlyr; lu++)
          for (iq=0; iq<output->atm.nq_t[iv]; iq++)
            output->atm.optprop.tau_molabs_t[lu][iv][iq] = output->atm.optprop.tau_molabs_t_ipa[ipa][lu][iv][iq];

        /* aerosol also changes due to saturation inside the cloud */
        /* copy IPA Array element[ipa] to optprop */
        status = cp_optprop (&(output->aer.optprop), &(output->aer.optprop_ipa[ipa]), output->aer.nlev, 1, input.quiet);
        if (status!=0) {
          fprintf (stderr, "Error %d copying aerosol optical properties\n", status);
          return status;
        }

      }
      

      /*  if (input.verbose) { */
      /*    fprintf (stderr, "*** wavelength: iv = %d, %f nm, albedo = %f\n", iv, output->wl.lambda_t[iv], output->alb.albedo_t[iv]); */
      /*    fprintf (stderr, "    atm.nmom + 1 = %d phase function moments\n", output->atm.nmom+1); */
      /*    fprintf (stderr, " --------------------------------------------------------------------------------------------\n"); */
      /*    fprintf (stderr, "   lu |    z[km] |      aerosol      |    water cloud    |    ice cloud      | tau_molecular \n"); */
      /*    fprintf (stderr, "      |          |        dtau  nmom |        dtau  nmom |        dtau  nmom |               \n"); */
      /*    fprintf (stderr, " --------------------------------------------------------------------------------------------\n"); */
      /*    for (lu=0; lu<output->atm.nlyr; lu++) */
      /*      fprintf (stderr, "%5d | %8.2f | %11.6f %5d | %11.6f %5d | %11.6f %5d | %11.6f \n",  */
      /* 	  lu, output->atm.zd[lu+1], */
      /* 	  0.0,0, /\*output->aer.dtau[iv][lu], output->aer.nmom[iv][lu],*\/ */
      /* 	  output->wc.optprop.dtau [iv][lu], output->wc.optprop.nmom[iv][lu], */
      /* 	  output->ic.optprop.dtau [iv][lu], output->ic.optprop.nmom[iv][lu], */
      /*          output->atm.optprop.tau_molabs_t[lu][iv][0]); */
      /*      fprintf (stderr, " ---------------------------------------------------------------------------\n"); */
      /*  } */


      /* need to load data during first call of mystic() for each band/wavelength */
      mc_loaddata=1;  

      /* reset band integral */
      reset_rte_output(&rte_outband, input, output->atm.nzout,
		       output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed);




      /******************************************/
      /* loop over quadrature points (subbands) */
      /******************************************/
      for (iq=0; iq<output->atm.nq_t[iv]; iq++) {
      
        if (input.verbose && input.ck_scheme != CK_NONE)
          fprintf (stderr, "\n*** wavelength: iv = %d, %f nm, looking at column %d, quadrature point nr %d, albedo = %f\n", 
                                  iv, output->wl.lambda_t[iv], ipa, iq, output->alb.albedo_t[iv]);
      
        /* set number of photons for this band */
        output->mc_photons = (int) (output->mc_photons_t[iv][iq] * (double) input.rte.mc.photons + 0.5);
        
        /* run at least MIN_MCPHOTONS for each band */
        if (output->mc_photons < MIN_MCPHOTONS)
          output->mc_photons = MIN_MCPHOTONS;
	
        if (input.verbose)
          fprintf (stderr, "... ck weight %9.7f\n", output->atm.wght_t[iv][iq]);
        
        if (output->cf.nlev>0) {

          /* Save optical properties for wavelength iv. This is necessary because averaged optical properties 
             are calulated for each subband and put into output->wc.optprop.... */
          if(iq==0) {
            for (lc=0;lc <output->atm.nlev-1;lc++) {
              /* optical depth */ 
              save_cloud->tauw[lc]  = output->wc.optprop.dtau [iv][lc];
              save_cloud->taui[lc]  = output->ic.optprop.dtau [iv][lc];
              
              /* asymmetry parameter */
              save_cloud->g1d[lc]   = output->wc.optprop.g1 [iv][lc];
              save_cloud->g2d[lc]   = output->wc.optprop.g2 [iv][lc];
              save_cloud->fd[lc]    = output->wc.optprop.ff [iv][lc];                    
              save_cloud->g1i[lc]   = output->ic.optprop.g1 [iv][lc];
              save_cloud->g2i[lc]   = output->ic.optprop.g2 [iv][lc];
              save_cloud->fi[lc]    = output->ic.optprop.ff [iv][lc];
              
              /* single scattering albedo */
              save_cloud->ssaw[lc]  = output->wc.optprop.ssa [iv][lc];
              save_cloud->ssai[lc]  = output->ic.optprop.ssa [iv][lc];
            }
          }
          
          if (input.verbose)
            fprintf(stderr, "*** ECHAM: generate effective cloud");
          
          status = generate_effective_cloud (input, output, save_cloud, iv, iq, input.verbose); 
          if (status!=0) {
            fprintf (stderr, "Error %d returned by generate_effective_cloud()\n", 
                     status);
            return status;
          }
            
        }

	/* setup optical properties and call the RTE solver */
	status = setup_and_call_solver (input, output, rte_in, rte_out, iv, iq, output->atm.threed, mc_loaddata);
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by setup_and_call_solver()\n", status);
	  return status;
	}
      

        /* verbose output */
        if (input.verbose) {
          fprintf (stderr, 
          "  iv = %d, %f nm, iq = %d, flux_dir[lu=0] = %13.7e, flux_dn[lu=0] = %13.7e, flux_up[lu=0] = %13.7e, weight_t = %13.7e \n", 
                  iv, output->wl.lambda_t[iv], iq, rte_out->rfldir[0], rte_out->rfldn[0], rte_out->flup[0], output->atm.wght_t[iv][iq]);
 	}

	/* data need to be loaded only once per iv */
	mc_loaddata=0;

        if (input.heating != NO_HEAT)
	  calc_spectral_heating (input, output, dz, rho_mass_zout, k_abs, k_abs_layer, 
                                 rte_in->pol.outlevels, rte_out, rte_out->heat, rte_out->w_zout, iv);

	/* add result for the current quadrature point considering quadrature weight */
	add_rte_output (rte_outband, rte_out, output->atm.wght_t[iv][iq], input, output->atm.nzout, 
			output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed);


      }

      /************************************************/
      /* add result for the current independent pixel */
      /************************************************/

      if (input.rte.solver == POLRADTRAN) {
	for(lu=0; lu<output->atm.nzout; lu++) {

	  output->rfldir_t [lu][iv] += output->ipaweight[ipa] * rte_outband->rfldir [lu];
          output->heat_t   [lu][iv] += output->ipaweight[ipa] * rte_outband->heat   [lu];
          output->w_zout_t [lu][iv] += output->ipaweight[ipa] * rte_outband->w_zout [lu];

	  for(is=0; is<input.rte.pol_nstokes; is++) {

	    output->up_flux_t[lu][is][iv]   += 
	      output->ipaweight[ipa] * rte_outband->polradtran_up_flux[lu][is]; 

	    output->down_flux_t[lu][is][iv] += 
	      output->ipaweight[ipa] * rte_outband->polradtran_down_flux[lu][is];

	    for (j=0;j<input.rte.nphi;j++) {
	      for (iu=0;iu<input.rte.numu;iu++) {
                output->down_rad_t[lu][j][iu][is][iv] += 
		  output->ipaweight[ipa] * rte_outband->polradtran_down_rad[lu][j][iu][is];
                output->up_rad_t[lu][j][iu][is][iv] += 
		  output->ipaweight[ipa] * rte_outband->polradtran_up_rad[lu][j][iu][is];
	      }
	    }
	  }
	}
      }
      else {
	for(lu=0; lu<output->atm.nzout; lu++) {
	  output->rfldir_t [lu][iv] += output->ipaweight[ipa] * rte_outband->rfldir [lu];
	  output->rfldn_t  [lu][iv] += output->ipaweight[ipa] * rte_outband->rfldn  [lu];
	  output->flup_t   [lu][iv] += output->ipaweight[ipa] * rte_outband->flup   [lu];
	  output->uavg_t   [lu][iv] += output->ipaweight[ipa] * rte_outband->uavg   [lu];
	  output->uavgdn_t [lu][iv] += output->ipaweight[ipa] * rte_outband->uavgdn [lu];
	  output->uavgso_t [lu][iv] += output->ipaweight[ipa] * rte_outband->uavgso [lu];
	  output->uavgup_t [lu][iv] += output->ipaweight[ipa] * rte_outband->uavgup [lu];
          output->heat_t   [lu][iv] += output->ipaweight[ipa] * rte_outband->heat   [lu];
          output->w_zout_t [lu][iv] += output->ipaweight[ipa] * rte_outband->w_zout [lu];	  

	  /* intensities */
	  for (iu=0;iu<input.rte.numu;iu++) {
	    output->u0u_t[lu][iu][iv] += output->ipaweight[ipa]*rte_outband->u0u[lu][iu];
	    
	    for (j=0;j<input.rte.nphi;j++)
	      output->uu_t[lu][j][iu][iv] += output->ipaweight[ipa]*rte_outband->uu[j][lu][iu];
	  }
	  
	  /* 3D fields */
	  if (input.rte.mc.passback3D) {
	    for (is=0; is<input.rte.mc.Nx_sample; is++) 
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		
		output->rfldir3d_t   [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->rfldir3d   [lu][is][js];

		output->rfldn3d_t    [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->rfldn3d    [lu][is][js];

		output->flup3d_t     [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->flup3d     [lu][is][js];

		output->uavgso3d_t   [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->uavgso3d   [lu][is][js];

		output->uavgdn3d_t   [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->uavgdn3d   [lu][is][js];

		output->uavgup3d_t   [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->uavgup3d   [lu][is][js];

		output->radiance3d_t [lu][is][js][iv] += 
		  output->ipaweight[ipa] * rte_outband->radiance3d [lu][is][js];
	      }
	  } 
	}

	/* 3D absorption */
	if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE) 
	  for (ks=0; ks<output->atm.Nzcld; ks++)
	    if (output->atm.threed[ks])    /* only for 3D layers, BM07122005 */
	      for (is=0; is<output->atm.Nxcld; is++)
		for (js=0; js<output->atm.Nycld; js++) 
		  output->abs3d_t [ks][is][js][iv] += 
		    output->ipaweight[ipa] * rte_outband->abs3d [ks][is][js];

      } /* for (iq=0; iq<output->atm.nq_t[iv]; iq++) { */

      /* verbose output */
      if (input.verbose) {
        fprintf (stderr, 
          "  iv = %d, %f nm, sum iq, flux_dir[lu=0] = %13.7e, flux_dn[lu=0] = %13.7e, flux_up[lu=0] = %13.7e \n", 
                    iv, output->wl.lambda_t[iv], output->rfldir_t[0][iv], output->rfldn_t[0][iv], output->flup_t[0][iv]);
      }

    } /* for (ipa=0; ipa<output->nipa; ipa++) independent pixel */ 



    /* change unit of the solar spectrum [e.g. W/(m2 nm)] or terrestral spectrum [e.g. W/(m2 cm-1)]  */  
    /* to output units wanted by the user: 'output per_nm', 'output per_cm', or 'output per_ck_band' */
    /* but only, when dealing with unit (not transmission or reflectivity)                           */
    /* UH 2006-03                                                                                    */

    unit_factor = get_unit_factor(input, output, iv);
    if(unit_factor <= 0) {
      fprintf(stderr,"Error, calculating unit_factor = %f\n", unit_factor);
      return -1;
    }

    switch (input.source) {
    case SRC_THERMAL:
      ffactor = unit_factor;
      rfactor = unit_factor;
      break;

    case SRC_SOLAR:
      switch (input.processing) {
      case PROCESS_INT:
      case PROCESS_SUM:
      case PROCESS_RGB:
      case PROCESS_RGBNORM:
	ffactor = unit_factor;
	rfactor = unit_factor;
	break;

      case PROCESS_NONE:
	switch (input.calibration) {
	case OUTCAL_ABSOLUTE:
	  ffactor = unit_factor;
	  rfactor = unit_factor;
	  break;
	  
	case OUTCAL_TRANSMITTANCE:
	  ffactor = 1.0;
	  rfactor = 1.0;
	  break;
	  
	case  OUTCAL_REFLECTIVITY:
	  ffactor = 1.0;
	  rfactor = 1.0;
	  break;

	default:
	  fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
	  return -1;
	}

	break;

      default:
	fprintf (stderr, "Error, unknown output processing %d\n", input.processing);
	return -1;
      }

      break;

    default:
      fprintf (stderr, "Error, unknown source %d\n", input.source);
      return -1;
    }

    hfactor = unit_factor;

    /*****************************************************************************************/
    /* now scale irradiances with ffactor, radiances with rfactor, heating rate with hfactor */
    /*****************************************************************************************/

    status = scale_output (input,
			   &(output->rfldir_t), &(output->rfldn_t),  &(output->flup_t),
			   &(output->uavgso_t), &(output->uavgdn_t), &(output->uavgup_t),
			   &(output->uavg_t), &(output->u0u_t), &(output->uu_t), &(output->heat_t), &(output->w_zout_t),
			   &(output->down_flux_t), &(output->up_flux_t), &(output->down_rad_t), &(output->up_rad_t),
			   &(output->rfldir3d_t), &(output->rfldn3d_t), &(output->flup3d_t), &(output->uavgso3d_t),
			   &(output->uavgdn3d_t), &(output->uavgup3d_t), &(output->radiance3d_t), 
			   output->atm.nzout, output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed, 
			   &(output->abs3d_t),
			   ffactor, rfactor, hfactor, iv);  /* in ancillary.c */
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_output()\n", status);
      return status;
    }

    /* free 3D cloud properties */
    if (input.rte.solver == MONTECARLO && (output->wc.occurrence1D || output->wc.threed.nthreed>0)) {
      status = free_wc_mystic (input, &(output->wc.threed));
      if (status!=0) {
	fprintf (stderr, "Error %d returned by free_wc_mystic()\n", status);
	return status;
      }
    }
  } /* for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) */
  

  /* free temporary memory */

  /* ??? why is that commented out ??? 
  free(rte_in);
  free_rte_output (rte_out, input);
  free_rte_output (rte_outband, input);
  */
  if (input.rte.solver == POLRADTRAN) {
    free (rte_in->pol.height);
    free (rte_in->pol.temperatures);
    free (rte_in->pol.gas_extinct);
    free (rte_in->pol.outlevels);
  }
  
  if (input.heating != NO_HEAT) {
    free (dz);
    free (k_abs_layer);
    free (k_abs);
  }
  if (output->cf.nlev>0)
    free (save_cloud);
  
  return 0;
}


/* small function to get factor for unit convertion */
double get_unit_factor (input_struct input, output_struct *output, int iv)
{
  double unit_factor=0.0;

  char function_name[]="get_unit_factor";
  char file_name[]="solve_rte.c";

  switch(output->spectrum_unit) {
  case UNIT_PER_NM:
    switch(input.output_unit) {
    case UNIT_PER_NM:
      unit_factor = 1.0;
      break;
    case UNIT_PER_CMm1:
      /* unit_factor = (lambda/k) */  /* (lambda/k) = (lambda**2) / 1.0e+7 */  /* 1.0e+7 == cm -> nm; */
      unit_factor = (output->wl.lambda_t[iv]*output->wl.lambda_t[iv]) / 1.0e+7;
      break;
    case UNIT_PER_CK_BAND:
      /* unit_factor = delta_lambda */  /* lambda_max = 1.0e+7 / k_lower;  lambda_min = 1.0e+7 / k_upper */
      unit_factor =  1.0e+7/output->wl.wvnmlo_t[iv] - 1.0e+7/output->wl.wvnmhi_t[iv];
      break;
    case UNIT_NOT_DEFINED:
      unit_factor = 1.0;
      break;
    default:
      fprintf (stderr, "Error: Program bug, unsupported output unit %d in %s (%s). \n", input.output_unit, function_name, file_name);
      return -1;
    }
    break;
  case UNIT_PER_CMm1:
    switch(input.output_unit) {
    case UNIT_PER_NM:
      /* unit_factor = (k/lambda) */  /* (k/lambda)= 1.0e+7 / (lambda**2) */  /* 1.0e+7 == cm -> nm; */
      /* k wavenumber in 1/cm**-1, lambda in nm */
      unit_factor = 1.0e+7 / (output->wl.lambda_t[iv]*output->wl.lambda_t[iv]); 
      break;
    case UNIT_PER_CMm1:
      unit_factor = 1.0;
      break;
    case UNIT_PER_CK_BAND:
      /* unit_factor = delta_k */
      unit_factor = output->wl.wvnmhi_t[iv]-output->wl.wvnmlo_t[iv];
      break;
    case UNIT_NOT_DEFINED:
      unit_factor = 1.0;
      break;
    default:
      fprintf (stderr, "Error: Program bug, unsupported output unit %d in %s (%s). \n", input.output_unit, function_name, file_name);
      return -1;
    }
    break;
  case UNIT_PER_CK_BAND:
    switch(input.output_unit) {
    case UNIT_PER_NM:
      /* unit_factor = 1 / delta_lambda */  /* lambda_max = 1.0e+7 / k_lower;  lambda_min = 1.0e+7 / k_upper */
      unit_factor = 1.0 / (1.0e7/output->wl.wvnmlo_t[iv] - 1.0e7/output->wl.wvnmhi_t[iv]);
      break;
    case UNIT_PER_CMm1:
      /* unit_factor = 1 / delta_k */
      unit_factor = 1.0 / (output->wl.wvnmhi_t[iv] - output->wl.wvnmlo_t[iv]);
      break;
    case UNIT_PER_CK_BAND:
      unit_factor = 1.0;
      break;
    case UNIT_NOT_DEFINED: /* not defined */
      unit_factor = 1.0;
      break;
    default:
      fprintf (stderr, "Error, program bug, unsupported output unit %d\n", input.output_unit);
      return -1;
    }
    break;
  case UNIT_NOT_DEFINED:
    switch(input.output_unit) {
    case UNIT_PER_NM:
    case UNIT_PER_CMm1:
    case UNIT_PER_CK_BAND:
      fprintf (stderr, "Error, can not convert undefined solar spectrum to output with units\n");
      fprintf (stderr, "       please use 'solar_file filename unit' in order to specify the unit of the spectrum\n");
      return -1;
      break;
    case UNIT_NOT_DEFINED: /* not defined */
      unit_factor = 1.0;
      break;
    default:
      fprintf (stderr, "Error, program bug, unsupported output unit %d\n", input.output_unit);
      return -1;
    }
    break;
  default:
    fprintf (stderr, "Error: Program bug, unsupported unit of solar_file %d\n", output->spectrum_unit);
    return -1;
  }

  return unit_factor; 

}


int setup_result (input_struct input, output_struct *output, float **p_dz, 
                  double **p_rho_mass_zout, float **p_k_abs_layer, float **p_k_abs, float **p_k_abs_outband)
{
  int status=0;
  int nlambda=0;
  int lc=0;

  if ((status = ASCII_calloc_float (&output->flup_t,   output->atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir_t, output->atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn_t,  output->atm.nzout, output->wl.nlambda_t)) !=0)
    return status;
  
  if ((status = ASCII_calloc_float (&output->uavg_t,   output->atm.nzout, output->wl.nlambda_t)) !=0)
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn_t, output->atm.nzout, output->wl.nlambda_t)) !=0)
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso_t, output->atm.nzout, output->wl.nlambda_t)) !=0)
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup_t, output->atm.nzout, output->wl.nlambda_t)) !=0)
    return status;

  if ((status = ASCII_calloc_float (&output->heat_t,   output->atm.nzout, output->wl.nlambda_t)) !=0)
      return status;

  if ((status = ASCII_calloc_float (&output->w_zout_t, output->atm.nzout, output->wl.nlambda_t)) !=0)
      return status;
  
  /* variables in order to calculate heating rates (by actinic flux) */
  if (input.heating != NO_HEAT) {

    *p_dz = (float *) calloc (output->atm.nlyr,  sizeof(float)); /* dz in m for all (nlyr) layers */
    if (*p_dz == NULL)
      return -1;

    /* Initialisation */
    for (lc=0;lc<output->atm.nlyr;lc++)
      (*p_dz)[lc] = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1000.0; /* km -> m */

    *p_rho_mass_zout = (double *) calloc (output->atm.nzout,  sizeof(double));
    if (*p_rho_mass_zout==NULL)
      return -1;    

    *p_k_abs_layer   = (float *) calloc (output->atm.nlyr,  sizeof(float));
    if (*p_k_abs_layer==NULL)
      return -1;

    *p_k_abs         = (float *) calloc (output->atm.nzout,  sizeof(float));
    if((*p_k_abs)==NULL)
      return -1;

    *p_k_abs_outband = (float *) calloc (output->atm.nzout,  sizeof(float));
    if (*p_k_abs_outband==NULL)
      return -1;

  }

  if (input.rte.numu>0)
    if ((status = ASCII_calloc_float_3D (&output->u0u_t, output->atm.nzout, input.rte.numu,
					 output->wl.nlambda_t)) != 0)
      return status;
  
  if (input.rte.numu>0 && input.rte.nphi>0)
    if ((status = ASCII_calloc_float_4D (&output->uu_t,  output->atm.nzout, input.rte.nphi, 
					 input.rte.numu, output->wl.nlambda_t)) !=0)
      return status;
  
  if (input.rte.mc.passback3D) {

    if (!input.quiet)
      fprintf (stderr, " ... allocating %d x %d x %d x %d = %d bytes for 3D output\n",
	       output->atm.nzout, input.rte.mc.Nx_sample,
	       input.rte.mc.Ny_sample, output->wl.nlambda_t,
	       output->atm.nzout * input.rte.mc.Nx_sample * input.rte.mc.Ny_sample * output->wl.nlambda_t*4);
	     

    if ((status = ASCII_calloc_float_4D (&output->rfldir3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->rfldn3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->flup3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->uavgso3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->uavgdn3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->uavgup3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;


    if ((status = ASCII_calloc_float_4D (&output->radiance3d_t, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, output->wl.nlambda_t)) != 0)
      return status;
    


    /* 3d absorption */
    if (input.rte.mc.absorption!=MCABS_NONE)
      if ((output->abs3d_t = calloc_spectral_abs3d (output->atm.Nxcld, 
						    output->atm.Nycld, 
						    output->atm.Nzcld, 
						    output->wl.nlambda_t, 
						    output->atm.threed)) == NULL) {
	fprintf (stderr, "Error allocating memory for abs3d\n");
	return -1;
      }    
  }    


  /* need to allocate enough memory for both output->wl.nlambda_s */
  /* and output->wl.nlambda_h, hence using whichever is larger    */
  nlambda = (output->wl.nlambda_h > output->wl.nlambda_s ?
	     output->wl.nlambda_h : output->wl.nlambda_s);

  if ((status = ASCII_calloc_float (&output->flup,   output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn,  output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir, output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavg,   output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn, output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso, output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup, output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->heat,   output->atm.nzout, nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->w_zout, output->atm.nzout, nlambda)) != 0) 
    return status;
  
  if ((status = ASCII_calloc_float_3D (&output->down_flux, output->atm.nzout, input.rte.pol_nstokes, 
				       nlambda)) != 0) 
    return status;
    
  if ((status = ASCII_calloc_float_3D (&output->down_flux_t,   output->atm.nzout, input.rte.pol_nstokes, 
				       output->wl.nlambda_t)) !=0)
    return status;
    
  if ((status = ASCII_calloc_float_3D (&output->up_flux, output->atm.nzout, input.rte.pol_nstokes, 
				       nlambda)) != 0) 
    return status;

  if ((status = ASCII_calloc_float_3D (&output->up_flux_t,   output->atm.nzout, input.rte.pol_nstokes, 
				       output->wl.nlambda_t)) !=0)
    return status;
    
  if (input.rte.nphi>0) {
    if ((status = ASCII_calloc_float_5D (&output->down_rad, output->atm.nzout, input.rte.nphi,
					 input.rte.numu, input.rte.pol_nstokes, 
					 nlambda)) !=0)
      return status;

    if ((status = ASCII_calloc_float_5D (&output->down_rad_t, output->atm.nzout, input.rte.nphi,
					 input.rte.numu, input.rte.pol_nstokes, 
					 output->wl.nlambda_t)) !=0)
      return status;
    
    if ((status = ASCII_calloc_float_5D (&output->up_rad, output->atm.nzout, input.rte.nphi,
					 input.rte.numu, input.rte.pol_nstokes, 
					 nlambda)) !=0)
      return status;

    if ((status = ASCII_calloc_float_5D (&output->up_rad_t, output->atm.nzout, input.rte.nphi,
					 input.rte.numu, input.rte.pol_nstokes, 
					 output->wl.nlambda_t)) !=0)
      return status;
  }

  if (input.rte.numu>0)
    if ((status = ASCII_calloc_float_3D (&output->u0u, output->atm.nzout, input.rte.numu, 
					 nlambda)) != 0) 
      return status;

  if (input.rte.numu>0 && input.rte.nphi>0)
    if ((status = ASCII_calloc_float_4D (&output->uu, output->atm.nzout, input.rte.nphi, 
					 input.rte.numu, nlambda)) != 0) 
      return status;


  if (input.rte.mc.passback3D) {
    if ((status = ASCII_calloc_float_4D (&output->rfldir3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->rfldn3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->flup3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;

    if ((status = ASCII_calloc_float_4D (&output->uavgso3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->uavgdn3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->uavgup3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    
    if ((status = ASCII_calloc_float_4D (&output->radiance3d, 
					 output->atm.nzout, input.rte.mc.Nx_sample,
					 input.rte.mc.Ny_sample, nlambda)) != 0)
      return status;
    

    /* 3D absorption */
    if (input.rte.mc.absorption!=MCABS_NONE)
      if ((output->abs3d = calloc_spectral_abs3d (output->atm.Nxcld, 
						  output->atm.Nycld, 
						  output->atm.Nzcld, 
						  nlambda, 
						  output->atm.threed)) == NULL) {
	fprintf (stderr, "Error allocating memorry for abs3\n"); 
	return -1;
      }    
  }


  if ((output->sza_h = (float *) calloc (output->wl.nlambda_h, sizeof(float))) == NULL)
    return -1;
  
  return 0;
}

static int reverse_profiles (input_struct input, output_struct *output)
{
  int lu=0, iu=0;
  double tmp=0;

  for (lu=0; lu<output->atm.nlyr/2; lu++) {
    
    /* reverse profile of optical depth dtauc */
    tmp               = output->dtauc[output->atm.nlyr-lu-1];
    output->dtauc[output->atm.nlyr-lu-1] = output->dtauc[lu];
    output->dtauc[lu] = tmp;
    
    /* reverse profile of single scattering albedo ssalb */
    tmp               = output->ssalb[output->atm.nlyr-lu-1];
    output->ssalb[output->atm.nlyr-lu-1] = output->ssalb[lu];
    output->ssalb[lu] = tmp;
    
    /* reverse profile of phase function pmom */
    for (iu=0;iu<=input.rte.nstr;iu++) {
      tmp                  =  output->pmom[output->atm.nlyr-lu-1][0][iu];
      output->pmom[output->atm.nlyr-lu-1][0][iu] = output->pmom[lu][0][iu];
      output->pmom[lu][0][iu] = tmp;
    }
  }
  return 0;
}


/* allocate temporary memory optical properties */
static save_optprop *calloc_save_optprop (int Nlev)
{
  save_optprop *result = calloc(1, sizeof(save_optprop));
  
  result->tauw  = (float *) calloc (Nlev, sizeof(float));
  result->taui  = (float *) calloc (Nlev, sizeof(float));
  result->g1d   = (float *) calloc (Nlev, sizeof(float));
  result->g2d   = (float *) calloc (Nlev, sizeof(float));
  result->fd    = (float *) calloc (Nlev, sizeof(float));
  result->g1i   = (float *) calloc (Nlev, sizeof(float));
  result->g2i   = (float *) calloc (Nlev, sizeof(float));
  result->fi    = (float *) calloc (Nlev, sizeof(float));
  result->ssaw  = (float *) calloc (Nlev, sizeof(float));
  result->ssai  = (float *) calloc (Nlev, sizeof(float));

  return result; 
}


/* allocate temporary memory for the RTE solvers */
static rte_output *calloc_rte_output (input_struct input, int nzout, int Nxcld, int Nycld, int Nzcld, int *threed)
{
  int status=0;


  rte_output *result = calloc(1, sizeof(rte_output));
  
  result->albmed = (float *) calloc (input.rte.maxumu, sizeof(float));
  result->trnmed = (float *) calloc (input.rte.maxumu, sizeof(float));
	     
  result->dfdt   = (float *) calloc (nzout,  sizeof(float));
  result->flup   = (float *) calloc (nzout,  sizeof(float));
  result->rfldir = (float *) calloc (nzout,  sizeof(float));
  result->rfldn  = (float *) calloc (nzout,  sizeof(float));
  result->uavg   = (float *) calloc (nzout,  sizeof(float));
  result->uavgdn = (float *) calloc (nzout,  sizeof(float));
  result->uavgso = (float *) calloc (nzout,  sizeof(float));
  result->uavgup = (float *) calloc (nzout,  sizeof(float));
  result->heat   = (float *) calloc (nzout,  sizeof(float));
  result->w_zout = (float *) calloc (nzout,  sizeof(float));
  
  if (input.rte.maxumu>0)
    if ((status = ASCII_calloc_float (&(result->u0u), 
				      nzout, input.rte.maxumu)) != 0) 
      return NULL;
  
  if (input.rte.maxphi>0 && input.rte.maxumu>0)
    if ((status = ASCII_calloc_float_3D (&(result->uu), 
					 input.rte.maxphi, nzout, input.rte.maxumu)) != 0) 
      return NULL;
  

  /* PolRadtran-specific */
  if (input.rte.solver == POLRADTRAN) {
    result->polradtran_mu_values = (double *) calloc (input.rte.nstr/2+input.rte.numu, sizeof(double));
    
    if ((status = ASCII_calloc_double (&(result->polradtran_up_flux), 
				       nzout, input.rte.pol_nstokes)) != 0)
      return NULL;
    
    
    if ((status = ASCII_calloc_double (&(result->polradtran_down_flux), 
				       nzout, input.rte.pol_nstokes)) != 0)
      return NULL;


    if ((status = ASCII_calloc_double_4D (&(result->polradtran_up_rad), 
					  nzout, input.rte.nphi,
					  input.rte.nstr/2+input.rte.numu,
                                          input.rte.pol_nstokes)) != 0)
      return NULL;

    if ((status = ASCII_calloc_double_4D (&(result->polradtran_down_rad), 
					  nzout, input.rte.nphi,
					  input.rte.nstr/2+input.rte.numu, 
                                          input.rte.pol_nstokes)) != 0)
      return NULL;


    if ((status = ASCII_calloc_double_4D (&(result->polradtran_up_rad_q), 
					  nzout, input.rte.pol_aziorder+1,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes)) != 0)
      return NULL;

    if ((status = ASCII_calloc_double_4D (&(result->polradtran_down_rad_q), 
					  nzout, input.rte.pol_aziorder+1,
					  input.rte.nstr/2+input.rte.numu, input.rte.pol_nstokes)) != 0)
      return NULL;

  }


  /* 3d fields */
  if (input.rte.mc.passback3D) {

    if ((status = ASCII_calloc_float_3D (&(result->rfldir3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->rfldn3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->flup3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->uavgso3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->uavgdn3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->uavgup3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	

    if ((status = ASCII_calloc_float_3D (&(result->radiance3d), 
					 nzout, 
					 input.rte.mc.Nx_sample, 
					 input.rte.mc.Ny_sample)) != 0)
      return NULL;
	


    if (input.rte.mc.absorption!=MCABS_NONE)
      if ((result->abs3d=calloc_abs3d (Nxcld, Nycld, Nzcld, threed)) == NULL)  
	return NULL;

  }   
  return result;
}


/* reset rte_output structure */
static int reset_rte_output (rte_output **rte, input_struct input, int nzout, int Nxcld, int Nycld, int Nzcld, int *threed) {
  free_rte_output(*rte, input, nzout, Nxcld, Nycld, Nzcld, threed);
  *rte = calloc_rte_output (input, nzout, Nxcld, Nycld, Nzcld, threed);

  return 0; /* if o.k. */
}


static int add_rte_output (rte_output *rte, rte_output *add, double factor, input_struct input, 
			   int nzout, int Nxcld, int Nycld, int Nzcld, int *threed)
{
  int j=0, iu=0, is=0, js=0, ks=0, lu=0;
  
  for(lu=0; lu<nzout; lu++) {
    
    rte->rfldir [lu] += factor*add->rfldir [lu];
    rte->rfldn  [lu] += factor*add->rfldn  [lu];
    rte->flup   [lu] += factor*add->flup   [lu];
    rte->uavg   [lu] += factor*add->uavg   [lu];
    rte->uavgdn [lu] += factor*add->uavgdn [lu];
    rte->uavgso [lu] += factor*add->uavgso [lu];
    rte->uavgup [lu] += factor*add->uavgup [lu];
    rte->dfdt   [lu] += factor*add->dfdt   [lu];
    rte->heat   [lu] += factor*add->heat   [lu];
    rte->w_zout [lu] += factor*add->w_zout [lu];    

    for (iu=0;iu<input.rte.numu;iu++) {
      rte->u0u[lu][iu] += factor*add->u0u[lu][iu];
      
      for (j=0;j<input.rte.nphi;j++)
	rte->uu[j][lu][iu] += factor*add->uu[j][lu][iu];
    }
  }

  for (iu=0;iu<input.rte.numu;iu++) {
    rte->albmed[iu] += factor*add->albmed[iu];
    rte->trnmed[iu] += factor*add->trnmed[iu];
  }


  /* PolRadtran-specific */
  if (input.rte.solver == POLRADTRAN) {
    for(lu=0; lu<nzout; lu++) {
      for(is=0; is<input.rte.pol_nstokes; is++) {
	
	rte->polradtran_up_flux[lu][is]   += factor*add->polradtran_up_flux[lu][is];
	rte->polradtran_down_flux[lu][is] += factor*add->polradtran_down_flux[lu][is];
	
	for (j=0;j<input.rte.nphi;j++) {
	  for (iu=0;iu<input.rte.nstr/2+input.rte.numu;iu++) {
	    rte->polradtran_up_rad[lu][j][iu][is]   += factor*add->polradtran_up_rad[lu][j][iu][is];
	    rte->polradtran_down_rad[lu][j][iu][is] += factor*add->polradtran_down_rad[lu][j][iu][is];
          }
	}
      }
    }
  }


  if (input.rte.mc.passback3D) {
    for (ks=0; ks<nzout; ks++) 
      for (is=0; is<input.rte.mc.Nx_sample; is++) 
	for (js=0; js<input.rte.mc.Ny_sample; js++) {
	  rte->rfldir3d   [ks][is][js] += factor * add->rfldir3d   [ks][is][js];
	  rte->rfldn3d    [ks][is][js] += factor * add->rfldn3d    [ks][is][js];
	  rte->flup3d     [ks][is][js] += factor * add->flup3d     [ks][is][js];
	  rte->uavgso3d   [ks][is][js] += factor * add->uavgso3d   [ks][is][js];
	  rte->uavgdn3d   [ks][is][js] += factor * add->uavgdn3d   [ks][is][js];
	  rte->uavgup3d   [ks][is][js] += factor * add->uavgup3d   [ks][is][js];

	  rte->radiance3d [ks][is][js] += factor * add->radiance3d [ks][is][js];
	}
    
    if (input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<Nzcld; ks++) 
	if (threed[ks])    /* only for 3D layers, BM07122005 */
	  for (is=0; is<Nxcld; is++) 
	    for (js=0; js<Nycld; js++) 
	      rte->abs3d [ks][is][js] += factor * add->abs3d [ks][is][js];
  }    
    
  return 0;  /* if o.k. */
}



static int cp_3d_to_rte_output (rte_output *rte, 
				float *rfldir3d, float *rfldn3d,  float *flup3d,
				float *uavgso3d, float *uavgdn3d, float *uavgup3d, 
				float *radiance3d,
				int Nz, int Nx, int Ny)
{
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, rfldir3d,   rte->rfldir3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, rfldn3d,    rte->rfldn3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, flup3d,     rte->flup3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, uavgso3d,   rte->uavgso3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, uavgdn3d,   rte->uavgdn3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, uavgup3d,   rte->uavgup3d);
  fortran2c_3D_float_ary_noalloc (Nz, Nx, Ny, radiance3d, rte->radiance3d);

  return 0;
}


/* free temporary memory for the RTE solvers */
static void free_rte_output (rte_output *result, input_struct input, 
			     int nzout, int Nxcld, int Nycld, int Nzcld, int *threed)
{

  if (result->albmed != NULL) free(result->albmed);

  if (result->trnmed != NULL) free(result->trnmed);
         
  if (result->dfdt != NULL)   free(result->dfdt);

  if (result->flup != NULL)   free(result->flup);

  if (result->rfldir != NULL) free(result->rfldir);

  if (result->rfldn != NULL)  free(result->rfldn);

  if (result->uavg != NULL)   free(result->uavg);

  if (result->uavgdn != NULL) free(result->uavgdn);

  if (result->uavgso != NULL) free(result->uavgso);

  if (result->uavgup != NULL) free(result->uavgup);

  if (result->heat != NULL)   free(result->heat);

  if (result->w_zout != NULL) free(result->w_zout);

  if (result->u0u != NULL)
    ASCII_free_float (result->u0u, nzout);


  if (result->uu != NULL)
    ASCII_free_float_3D (result->uu, input.rte.nphi, nzout);


  if (input.rte.solver == POLRADTRAN) {

    if (result->polradtran_mu_values != NULL)
      free (result->polradtran_mu_values);
    
    if (result->polradtran_up_flux != NULL)
      ASCII_free_double(result->polradtran_up_flux,   nzout);

    if (result->polradtran_down_flux != NULL)
      ASCII_free_double(result->polradtran_down_flux, nzout);
    
    if (result->polradtran_up_rad != NULL)
      ASCII_free_double_4D(result->polradtran_up_rad, 
			   nzout, input.rte.nphi, input.rte.nstr/2+input.rte.numu);
    
    if (result->polradtran_down_rad != NULL) 
      ASCII_free_double_4D(result->polradtran_down_rad, 
			   nzout, input.rte.nphi, input.rte.nstr/2+input.rte.numu);

    if (result->polradtran_up_rad_q != NULL)
      ASCII_free_double_4D(result->polradtran_up_rad_q, 
			   nzout, input.rte.pol_aziorder+1, input.rte.nstr/2+input.rte.numu);
    
    if (result->polradtran_down_rad_q != NULL)
      ASCII_free_double_4D(result->polradtran_down_rad_q, 
			   nzout, input.rte.pol_aziorder+1, input.rte.nstr/2+input.rte.numu);
  }


  /* free 3d fields */
  if (input.rte.mc.passback3D) {

    if (result->rfldir3d != NULL)
      ASCII_free_float_3D(result->rfldir3d, 
			  nzout, input.rte.mc.Nx_sample);

    if (result->rfldn3d != NULL)
      ASCII_free_float_3D(result->rfldn3d, 
			  nzout, input.rte.mc.Nx_sample);

    if (result->flup3d != NULL)
      ASCII_free_float_3D(result->flup3d, 
			  nzout, input.rte.mc.Nx_sample);
    
    if (result->uavgso3d != NULL)
      ASCII_free_float_3D(result->uavgso3d, 
			  nzout, input.rte.mc.Nx_sample);

    if (result->uavgdn3d != NULL)
      ASCII_free_float_3D(result->uavgdn3d, 
			  nzout, input.rte.mc.Nx_sample);

    if (result->uavgup3d != NULL)
      ASCII_free_float_3D(result->uavgup3d, 
			  nzout, input.rte.mc.Nx_sample);

    if (result->radiance3d != NULL)
      ASCII_free_float_3D(result->radiance3d, 
			  nzout, input.rte.mc.Nx_sample);
    
    if (input.rte.mc.absorption!=MCABS_NONE)
      if (result->abs3d != NULL)
	free_abs3d (result->abs3d, Nxcld, Nycld, Nzcld, threed);
  }
  free(result);
}





/******************************************************/
/* setup optical properties, do some initializations, */
/* and call the RTE solver.                           */
/******************************************************/

static int setup_and_call_solver (input_struct input, output_struct *output, 
				  rte_input *rte_in, rte_output *rte_out, 
				  int iv, int ib, int *threed, int mc_loaddata)
{
  int status=0, lu=0;
  static int first = 1;
  static int verbose = 0;

  verbose = input.verbose;

  if (input.verbose && input.ck_scheme==CK_LOWTRAN)
    verbose=0; /* no verbose output for the following call of optical properties */

  /* calculate optical properties from model input */
  status = optical_properties (input, output, iv, ib, verbose); /* in ancillaries.c */

  if (status!=0) {
    fprintf (stderr, "Error %d returned by optical_properties()\n", 
	     status);
    return status;
  }


  if (input.ck_scheme==CK_LOWTRAN) {

    /* this is a little inefficient as we need first the scattering optical */
    /* properties to calculate the absorption coefficients and then         */
    /* recalculate the optical properties; the reason is that SBDART        */
    /* requires the total scattering cross section as input, and this       */
    /* quantity is only available after the call to optical_properties      */

    /* ??? where to get the mixing ratio of N2 from ??? */
    /* ??? setting this value to -1                 ??? */
    status = sbdart_profile (output->atm.microphys.temper, output->atm.microphys.press, output->atm.zd, 
			     output->atm.microphys.dens[MOL_H2O], output->atm.microphys.dens[MOL_O3], 
			     -1.0, 
			     output->mixing_ratio[MX_O2], 
			     output->mixing_ratio[MX_CO2], 
			     output->mixing_ratio[MX_CH4], 
			     output->mixing_ratio[MX_N2O], 
			     output->mixing_ratio[MX_NO2], 
			     input.ck_o4abs, input.ck_n2abs, input.ck_coabs, 
			     input.ck_so2abs, input.ck_nh3abs, input.ck_noabs, input.ck_hno3abs,
			     output->dtauc, output->ssalb, output->atm.nlev, 
			     output->atm.sza_t[iv], output->wl.lambda_t[iv],
		             iv, output->crs_ck.profile[0]);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by sbdart_profile()\n", status);
      return status;
    }


    /* allocate memory for absorption coefficient profile */
    if (first) {
      status  = ASCII_calloc_float (&output->kabs.kabs, output->atm.nlev, 
				    output->wl.nlambda_t);
      if (status!=0) { 
	fprintf (stderr, "Error %d allocating memory for output->kabs.kabs\n", status); 
	return status; 
      } 

      first=0;
    }
    
    /* set absorption coefficient */
    for (lu=0; lu<output->atm.nlyr; lu++)
      output->kabs.kabs[lu][iv]  = 
	output->crs_ck.profile[0][iv].crs[lu][ib];

    /* calculate optical properties from model input */
    status = optical_properties (input, output, iv, ib, input.verbose); /* in ancillaries.c */
    
    if (status!=0) {
	fprintf (stderr, "Error %d returned by optical_properties() in ancillaries.c\n", 
		 status);
	return status;
    }
  }


  /* translate output altitudes (zout) to optical depths (utau) */
  if ( input.rte.solver != MONTECARLO && input.rte.solver != NULLSOLVER ) {

    /* "old" version */
    /*
    status = set_out (output->atm.zd, output->dtauc, output->atm.nlyr, 
		      input.atm.zout, output->atm.nzout, rte_in->utau);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by set_out()\n", status);
      return status;
    }
    */

    /* "very old" version, recycled */ 
    F77_FCN  (setout) (output->dtauc, 
		       &(output->atm.nlyr), 
		       &(output->atm.nzout), 
		       rte_in->utau,
		       output->atm.zd, 
		       output->atm.zout_sur);
   
  }


  /* reverse profiles, if required */
  if (input.rte.reverse) {
    status = reverse_profiles (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d reversing profiles\n", status);
      return status;
    }
  }
  

  /* call RTE solver */
  status = call_solver (input, output, rte_in, iv, rte_out, threed, mc_loaddata, input.verbose);
  if (status!=0) {
    fprintf (stderr, "Error %d calling solver\n", status);
    return status;
  }
  
  return 0;  /* if o.k. */
}





static int call_solver (input_struct input, output_struct *output,
			rte_input *rte_in, int iv, rte_output *rte_out,
			int *threed, int mc_loaddata, int verbose)
{
  int status=0;
  int lev=0;
  int rte_solver = NOT_DEFINED_INTEGER;

  double start=0, end=0;

  int lu=0, is=0, k=0;

  float radius= 6370.0; /* Earth's radius in km */

  float *twostr_gg    = NULL;
  float *sdisort_beta = NULL;
  float *sdisort_sig  = NULL;
  float **sdisort_denstab = NULL;
  float *tosdisort_denstab = NULL;
  float *qsrc = NULL;  /* General temporary source for qdisort */
  float *disort_pmom  = NULL, *disort2_pmom  = NULL, *sss_pmom  = NULL;

  int planck_tempoff = 0;

#if HAVE_POLRADTRAN
  int iu=0; 
 
  double *polradtran_down_flux = NULL;
  double *polradtran_up_flux   = NULL;
  double *polradtran_down_rad  = NULL;
  double *polradtran_up_rad    = NULL;
#endif

  /* temporary Fortran arrays */
  float *disort_u0u = NULL, *disort_uu = NULL;
  float *tzs_u0u = NULL, *tzs_uu = NULL;
  float *sss_u0u = NULL, *sss_uu = NULL;

#if HAVE_MYSTIC
  int il=0;
  int source=0;
  int thermal_photons=0;
  int nres3d=0;

  /* temporary MC output (for thermal MC calculations, two calls */
  /* to mc() are required, one for the surface and one for the   */
  /* atmospheric contribution                                    */
  rte_output *tmp_out=NULL;

  float *tmp_rfldir3d=NULL, *tmp_rfldn3d=NULL,  *tmp_flup3d=NULL; 
  float *tmp_uavgso3d=NULL, *tmp_uavgdn3d=NULL, *tmp_uavgup3d=NULL; 
  float *tmp_radiance3d=NULL;
 
  /* rpv arrays */
  float *rpv_rho0=NULL;
  float *rpv_k=NULL;
  float *rpv_theta=NULL;
  float *rpv_scale=NULL;

  int write_files=0;
  float **pmom_sos = NULL; 
  
  /* write MYSTIC monochromatic output only if monochromatic, non-ck uvspec calculation */
  write_files = (output->wl.nlambda_t * output->atm.nq_t[output->wl.nlambda_rte_lower] > 1 ? 0 : 1);

#endif

  if (verbose)
    start = clock();

  /* change rte solver to NULL solver for                                   */
  /* solar simulations with sza > 90 degrees done by plane paralell solvers */

  rte_solver = input.rte.solver;

  switch (input.source) {
  case SRC_THERMAL:
    /* do not change solver */
    break;
  case SRC_SOLAR:
    if (output->atm.sza_t[iv] >= 90.0) {
      switch (input.rte.solver) {
      
      /* plane parallel solvers */
      case DISORT:
      case DISORT2:
      case QDISORT:
      case TWOSTRPP:
      case POLRADTRAN:
      case SSS:
      case SSSI:
      case SSS2:
        /* change solver to NULLSOLVER, as output is 0 anyway */
        rte_solver = NULLSOLVER;
        if (input.verbose)
          fprintf (stderr, " ... solar calculation with SZA>90 and pp solver, switch solver to NULL-solver \n");
        break;

      /* sperical solvers */
      case SDISORT:
      case SPSDISORT:
      case TWOSTR:
      case SUCORD:
        /* do not change solver as these solvers should be   */
	/* able to do solar calculations for sza > 90 degree */
        break;

      /* special solvers */
      case MONTECARLO:      /* user must know it */
      case TZS:             /* only thermal anyway */
        /* do nothing */
        break;

      case NULLSOLVER:
        break;
      default:
        fprintf (stderr, "Error, unknown RTE solver %d\n", input.rte.solver);
        break;
      }
    }
    break;
  default:
    fprintf (stderr, "Error, unknown source %d\n", input.source);
    return -1;
  }


  switch (rte_solver) {  /* this is the ONLY use of rte_solver, everywhere else still input.rte.solver is used */
  case MONTECARLO:
#if HAVE_MYSTIC

    /* allocate memory for 3D result fields */
    if (input.rte.mc.passback3D) {
      nres3d = input.rte.mc.Nx_sample * input.rte.mc.Ny_sample * output->atm.nzout;
      
      tmp_rfldir3d   = calloc (nres3d, sizeof(float));
      tmp_rfldn3d    = calloc (nres3d, sizeof(float));
      tmp_flup3d     = calloc (nres3d, sizeof(float));
      tmp_uavgso3d   = calloc (nres3d, sizeof(float));
      tmp_uavgdn3d   = calloc (nres3d, sizeof(float));
      tmp_uavgup3d   = calloc (nres3d, sizeof(float));
      tmp_radiance3d = calloc (nres3d, sizeof(float));

      if (tmp_rfldir3d==NULL || tmp_rfldn3d==NULL  || tmp_flup3d==NULL   ||
	  tmp_uavgso3d==NULL || tmp_uavgdn3d==NULL || tmp_uavgup3d==NULL ||
	  tmp_radiance3d==NULL) {
	fprintf (stderr, "Error allocating memory for 3D result fields!\n");
	return -1;
      }
    }

    /* create rpv arrays */
    if (output->surf.n > 0) {
      rpv_rho0  = calloc (output->surf.n, sizeof(float));
      rpv_k     = calloc (output->surf.n, sizeof(float));
      rpv_theta = calloc (output->surf.n, sizeof(float));
      rpv_scale = calloc (output->surf.n, sizeof(float));

      for (il=0; il<output->surf.n; il++) {
	rpv_rho0 [il] = output->surf.rpv[il].rho0_t [iv];
	rpv_k    [il] = output->surf.rpv[il].k_t    [iv];
	rpv_theta[il] = output->surf.rpv[il].theta_t[iv];
	rpv_scale[il] = output->surf.rpv[il].scale_t[iv];
      }
    }
    else {
      rpv_rho0  = calloc (1, sizeof(float));
      rpv_k     = calloc (1, sizeof(float));
      rpv_theta = calloc (1, sizeof(float));
      rpv_scale = calloc (1, sizeof(float));

      rpv_rho0 [0] = output->rpv.rho0_t [iv];
      rpv_k    [0] = output->rpv.k_t    [iv];
      rpv_theta[0] = output->rpv.theta_t[iv];
      rpv_scale[0] = output->rpv.scale_t[iv];
    }

    switch (input.source) {
    case SRC_SOLAR: /* solar source */

      source = MCSRC_SOLAR;

      status = mystic (&output->atm.nlyr,
		       output->mc.dtair, output->mc.dtmol, 
		       output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
		       output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
		       output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
		       output->mc.refind,
		       &(output->rayleigh_depol[iv]),
		       &(output->wc.threed),
		       &(output->ic.threed),
		       output->mc.recld, output->mc.reice, 
		       output->mc.temper,
		       output->mc.z,
		       &(output->surface_temperature),
		       output->mc.momaer, output->mc.nmomaer,
		       &output->alb.albedo_t[iv], 
		       &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
		       output->mc_photons,
		       &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
		       &(output->wl.lambda_t[iv]),
		       output->atm.zout_sur, &(output->atm.nzout),
		       rpv_rho0, rpv_k, rpv_theta,
		       &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
		       &(input.rte.mc.ipa), &(input.rte.mc.absorption),
		       &mc_loaddata,
		       &(output->mc.sample),
		       input.rte.mc.filename[FN_MC_BASENAME], 
		       input.rte.mc.filename[FN_MC_ALBEDO], 
		       input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
		       input.rte.mc.filename[FN_MC_AMBRALS], 
		       input.rte.mc.filename[FN_MC_ELEVATION], 
		       &(input.rte.mc.delta_scaling_mucut), /*TZ ds*/
		       &(input.rte.mc.truncate),
		       &(input.rte.mc.reflectalways),
		       input.rte.mc.surfaceparallel,
		       &(input.quiet),
		       rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		       rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
		       tmp_rfldir3d, tmp_rfldn3d,  tmp_flup3d,
		       tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
		       tmp_radiance3d,
		       rte_out->abs3d, 
		       output->atm.Nxcld,
		       output->atm.Nycld,
		       output->atm.Nzcld,
		       input.filename[FN_PATH],
		       write_files);


      if (status!=0) {
        fprintf (stderr, "Error %d returned by mystic()\n", status);
        return status;
      }

      if (input.rte.mc.passback3D) {

        status = cp_3d_to_rte_output (rte_out,
                                      tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d,
                                      tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d,
                                      tmp_radiance3d,
                                      output->atm.nzout,
                                      input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

        if (status!=0) {
          fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
          return status;
        }
      }

      break;

    case SRC_THERMAL: /* thermal source */

      /* normal atmospheric + surface emission */
      switch (input.rte.mc.absorption) {   
      case MCABS_NONE:
      case MCABS_ABSORPTION:
      case MCABS_HEATING:
	if (!input.rte.mc.backward.yes) {/*TZ bt*/

	  thermal_photons = 0.5 * output->mc_photons;
	  
	  /* thermal emission of the atmosphere */
	  fprintf (stderr, " ... thermal emission of the atmosphere\n");

	}
	else {/*TZ bt*/

	  thermal_photons = 1.0 * output->mc_photons;/*TZ bt*/
	  
	  /* thermal emission into the atmosphere TZ bt*/
	  fprintf (stderr, " ... thermal backward emission into atmosphere\n");/*TZ*/

	}/*TZ bt*/
		
	if (!input.rte.mc.backward.yes) /*TZ bt*/
	  source = MCSRC_THERMAL_ATMOSPHERE;
	else /*TZ bt*/
	  source = MCSRC_THERMAL_BACKWARD;/*TZ bt*/

	status = mystic (&output->atm.nlyr,
			 output->mc.dtair, output->mc.dtmol, 
			 output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
			 output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			 output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			 output->mc.refind,
			 &(output->rayleigh_depol[iv]),
			 &(output->wc.threed),
			 &(output->ic.threed),
			 output->mc.recld, output->mc.reice,
			 output->mc.temper,
			 output->mc.z,
			 &(output->surface_temperature),
			 output->mc.momaer, output->mc.nmomaer,
			 &output->alb.albedo_t[iv], 
			 &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			 thermal_photons,
			 &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			 &(output->wl.lambda_t[iv]),
			 output->atm.zout_sur, &(output->atm.nzout),
			 rpv_rho0, rpv_k, rpv_theta,
			 &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
			 &(input.rte.mc.ipa), &(input.rte.mc.absorption),
			 &mc_loaddata,
			 &(output->mc.sample),
			 input.rte.mc.filename[FN_MC_BASENAME], 
			 input.rte.mc.filename[FN_MC_ALBEDO], 
			 input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			 input.rte.mc.filename[FN_MC_AMBRALS], 
			 input.rte.mc.filename[FN_MC_ELEVATION], 
			 &(input.rte.mc.delta_scaling_mucut), /*TZ ds*/
			 &(input.rte.mc.truncate),
			 &(input.rte.mc.reflectalways),
			 input.rte.mc.surfaceparallel,
			 &(input.quiet),
			 rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			 rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
			 tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, 
			 tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
			 tmp_radiance3d,
			 rte_out->abs3d,
			 output->atm.Nxcld,
			 output->atm.Nycld,
			 output->atm.Nzcld,
			 input.filename[FN_PATH],
			 write_files);
      
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
      
	if (input.rte.mc.passback3D) {
	  status = cp_3d_to_rte_output (rte_out, 
					tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d,
					tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
					tmp_radiance3d,
					output->atm.nzout, 
					input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	    return status;
	  }
	}
      
	/* thermal emission of the surface */
	if (!input.rte.mc.backward.yes) {/*TZ bt, no separate surface emission needed*/
	  fprintf (stderr, " ... thermal emission of the surface\n");
	  
	  /* data have already been loaded during the last MYSTIC call */
	  mc_loaddata=0;
	  
	  tmp_out = calloc_rte_output (input, output->atm.nzout,
				       output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, 
				       output->atm.threed);

	  source = MCSRC_THERMAL_SURFACE;

	  status = mystic (&output->atm.nlyr,
			   output->mc.dtair, output->mc.dtmol, 
			   output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer,
			   output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			   output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			   output->mc.refind,
			   &(output->rayleigh_depol[iv]),
			   &(output->wc.threed),
			   &(output->ic.threed),
			   output->mc.recld, output->mc.reice,
			   output->mc.temper,
			   output->mc.z,
			   &(output->surface_temperature),
			   output->mc.momaer, output->mc.nmomaer,
			   &output->alb.albedo_t[iv], 
			   &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			   thermal_photons,
			   &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			   &(output->wl.lambda_t[iv]),
			   output->atm.zout_sur, &(output->atm.nzout),
			   rpv_rho0, rpv_k, rpv_theta,
			   &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
			   &(input.rte.mc.ipa), &(input.rte.mc.absorption),
			   &mc_loaddata,
			   &(output->mc.sample),
			   input.rte.mc.filename[FN_MC_BASENAME], 
			   input.rte.mc.filename[FN_MC_ALBEDO], 
			   input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			   input.rte.mc.filename[FN_MC_AMBRALS], 
			   input.rte.mc.filename[FN_MC_ELEVATION], 
			   &(input.rte.mc.delta_scaling_mucut), /*TZ ds*/
			   &(input.rte.mc.truncate),
			   &(input.rte.mc.reflectalways),
			   input.rte.mc.surfaceparallel,
			   &(input.quiet),
			   tmp_out->rfldir, tmp_out->rfldn, tmp_out->flup, 
			   tmp_out->uavgso, tmp_out->uavgdn, tmp_out->uavgup,
			   tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, 
			   tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
			   tmp_radiance3d,
			   tmp_out->abs3d,
			   output->atm.Nxcld,
			   output->atm.Nycld,
			   output->atm.Nzcld,
			   input.filename[FN_PATH],
			   write_files);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by mc()\n", status);
	    return status;
	  }
	  
	  if (input.rte.mc.passback3D) {
	    status = cp_3d_to_rte_output (tmp_out, 
					  tmp_rfldir3d, tmp_rfldn3d,  tmp_flup3d, 
					  tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
					  tmp_radiance3d,
					  output->atm.nzout, 
					  input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);
	    
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	      return status;
	    }
	  }

	  /* add atmosphere and surface contributions to get total (rte_out) */
	  status = add_rte_output (rte_out, tmp_out, 1.0, input, output->atm.nzout, 
				   output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed);
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by add_rte_output()\n", status);
	    return status;
	  }
      
	  free_rte_output (tmp_out, input, output->atm.nzout, output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed);
	}/*TZ bt, no surface emission needed*/

	break;

      case MCABS_EMISSION:
	
	thermal_photons = 0;
	
	/* 3D emission field */
	fprintf (stderr, " ... 3D emission field\n");
	
	source = MCSRC_THERMAL_ATMOSPHERE;
	
	status = mystic (&output->atm.nlyr,
			 output->mc.dtair, output->mc.dtmol, 
			 output->mc.dtaer, output->mc.omaer, output->mc.g1aer, output->mc.g2aer, output->mc.ffaer, 
			 output->mc.dtcld, output->mc.omcld, output->mc.g1cld, output->mc.g2cld, output->mc.ffcld, 
			 output->mc.dtice, output->mc.omice, output->mc.g1ice, output->mc.g2ice, output->mc.ffice,
			 output->mc.refind,
			 &(output->rayleigh_depol[iv]),
			 &(output->wc.threed),
			 &(output->ic.threed),
			 output->mc.recld, output->mc.reice,
			 output->mc.temper,
			 output->mc.z,
			 &(output->surface_temperature),
			 output->mc.momaer, output->mc.nmomaer,
			 &output->alb.albedo_t[iv], 
			 &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
			 thermal_photons,
			 &source, &(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]), 
			 &(output->wl.lambda_t[iv]),
			 output->atm.zout_sur, &(output->atm.nzout),
			 rpv_rho0, rpv_k, rpv_theta,
			 &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
			 &(input.rte.mc.ipa), &(input.rte.mc.absorption),
			 &mc_loaddata,
			 &(output->mc.sample),
			 input.rte.mc.filename[FN_MC_BASENAME], 
			 input.rte.mc.filename[FN_MC_ALBEDO], 
			 input.rte.mc.filename[FN_MC_RPV], output->surf.label, &(output->surf.n),
			 input.rte.mc.filename[FN_MC_AMBRALS], 
			 input.rte.mc.filename[FN_MC_ELEVATION], 
			 &(input.rte.mc.delta_scaling_mucut), /*TZ ds*/
			 &(input.rte.mc.truncate),
			 &(input.rte.mc.reflectalways),
			 input.rte.mc.surfaceparallel,
			 &(input.quiet),
			 rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			 rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
			 tmp_rfldir3d, tmp_rfldn3d, tmp_flup3d, 
			 tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
			 tmp_radiance3d,
			 rte_out->abs3d,
			 output->atm.Nxcld, 
			 output->atm.Nycld, 
			 output->atm.Nzcld,
			 input.filename[FN_PATH],
			 write_files);
      
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
	
	if (input.rte.mc.passback3D) {
	  status = cp_3d_to_rte_output (rte_out, 
					tmp_rfldir3d, tmp_rfldn3d,  tmp_flup3d, 
					tmp_uavgso3d, tmp_uavgdn3d, tmp_uavgup3d, 
					tmp_radiance3d,
					output->atm.nzout, 
					input.rte.mc.Nx_sample, input.rte.mc.Ny_sample);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by cp_3d_to_rte_output()\n", status);
	    return status;
	  }
	}

	break;

      default:
	fprintf (stderr, "Error, unknown absorption type %d\n", input.rte.mc.absorption);
	return -1;
      }

      break;

    default:
      fprintf (stderr, "Error, unknown source\n");
      return -1;
    }


    /* free memory of 3D results */
    if (input.rte.mc.passback3D) {
      free (tmp_rfldir3d);
      free (tmp_rfldn3d);
      free (tmp_flup3d);
      free (tmp_uavgso3d);
      free (tmp_uavgdn3d);
      free (tmp_uavgup3d);

      if (tmp_radiance3d != NULL)
	free (tmp_radiance3d);
    }

    /* free RPV arrays */
    free (rpv_rho0);
    free (rpv_k);
    free (rpv_theta);
    free (rpv_scale);


    break;

#else
    fprintf (stderr, "Error: MYSTIC solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Please contact bernhard.mayer@dlr.de\n");
    return -1;
#endif	

  case DISORT:
  case DISORT2:

    disort_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    if (input.rte.solver == DISORT) {

      status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
				&input.rte.numu, &input.rte.nphi); 
	
      if (status!=0)
	return status;


      disort_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1, input.rte.nstr+1,  output->pmom);

      if (((input.source == SRC_SOLAR) && (rte_in->umu0 > 0)) || (input.source == SRC_THERMAL)) { 
                              /* no need to call disort otherwise as sun below horizon */
	F77_FCN  (disort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
			  output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
			  &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &output->atm.nzout, rte_in->utau,
			  &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
			  &input.rte.nphi, input.rte.phi, &(rte_in->ibcnd), &(rte_in->fbeam), 
			  &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->fisot), &(rte_in->lamber), 
			  &output->alb.albedo_t[iv], rte_in->hl, &(rte_in->btemp), &(rte_in->ttemp), 
			  &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
			  &(rte_in->onlyfl), &(rte_in->accur), 
			  rte_in->prndis, rte_in->header, &output->atm.nlyr, 
			  &output->atm.nzout, &input.rte.maxumu, &input.rte.nstr,
			  &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			  rte_out->dfdt, rte_out->uavg, 
			  disort_uu, disort_u0u, rte_out->albmed, rte_out->trnmed, 
			  rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
			  &(input.quiet));
 
        for (lev=0;lev<output->atm.nzout;lev++) 
             rte_out->uavg[lev] = rte_out->uavgso[lev] + rte_out->uavgdn[lev] + rte_out->uavgup[lev]; 

      }
    }
    else if (input.rte.solver == DISORT2) {
      
      status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
				 &input.rte.numu, &input.rte.nphi); 
      
      if (status!=0)
	return status;
      

      /* BRDF or Lambertian albedo */
      if (input.brdf != BRDF_NONE)
	rte_in->lamber = 0;
      else
	rte_in->lamber = 1;
	
      disort2_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1,  output->atm.nmom+1, output->pmom);

      if (((input.source == SRC_SOLAR) && (rte_in->umu0 > 0)) || (input.source == SRC_THERMAL)) { 
                  /* no need to call disort2 otherwise, as sun below horizon */
	F77_FCN  (disort2)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
			   disort2_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
			   &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &output->atm.nzout, rte_in->utau,
			   &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
			   &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
			   &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
			   &output->alb.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
			   &rte_in->temis, &rte_in->planck,
			   &rte_in->onlyfl, &rte_in->accur,
			   rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
			   &output->atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
			   rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
			   rte_out->dfdt, rte_out->uavg, 
			   disort_uu, disort_u0u, rte_out->albmed, rte_out->trnmed, 
			   rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
			   &(input.brdf),
			   &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]), 
			   &input.rpv.sigma, &input.rpv.t1, &input.rpv.t2, &(output->rpv.scale_t[iv]), 
			   &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), &(input.quiet));
        for (lev=0;lev<output->atm.nzout;lev++) 
             rte_out->uavg[lev] = rte_out->uavgso[lev] + rte_out->uavgdn[lev] + rte_out->uavgup[lev]; 
      }
    }    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(disort_pmom);
    free(disort2_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;

  case TZS:

    tzs_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    tzs_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;
    
    
    /* call RTE solver */
    F77_FCN  (tzs) (&output->atm.nlyr, output->dtauc, output->ssalb, 
		    output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		    &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &output->atm.nzout, rte_in->utau,
		    &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		    &input.rte.nphi, input.rte.phi, 
		    &output->alb.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		    &rte_in->temis, &rte_in->planck,
		    rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		    &output->atm.nzout, &input.rte.maxumu, &input.rte.maxphi, 
		    rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		    rte_out->dfdt, rte_out->uavg, 
		    tzs_uu, rte_out->albmed, rte_out->trnmed, 
		    rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup);

    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, tzs_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, tzs_uu, rte_out->uu);

    free(tzs_u0u);
    free(tzs_uu);

    break;

  case SSS:

    sss_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    sss_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by dcheck()\n", status);
      return status;
    }
    
    /* call RTE solver */
    sss_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1, output->atm.nmom+1, output->pmom);
    F77_FCN  (sss)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
		   sss_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		   &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &output->atm.nzout, rte_in->utau,
		   &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		   &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
		   &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
		   &output->alb.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		   &rte_in->temis, &rte_in->planck,
		   &rte_in->onlyfl, &rte_in->accur,
		   rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
		   &output->atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
		   rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		   rte_out->dfdt, rte_out->uavg, 
		   sss_uu, rte_out->albmed, rte_out->trnmed, 
		   rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		   &(input.brdf),
		   &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
		   &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal));
    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }


    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, sss_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, sss_uu, rte_out->uu);

    free(sss_pmom);
    free(sss_u0u);
    free(sss_uu);

    break;

  case SSSI:

#if HAVE_SSSI

    /* get cloud reflectivity from lookup table */
    status = read_isccp_reflectivity (output->sssi.type, output->atm.sza_t[iv], output->sssi.tautot, 
				      input.filename[FN_PATH], input.quiet, &(output->sssi.ref));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_isccp_reflectivity()\n", status);
      return status;
    }

    if (input.verbose) {
      fprintf (stderr, "*** SSSI cloud properties:\n");
      fprintf (stderr, "    top level:          zd[%d] = %f\n", 
	       output->sssi.lctop, output->atm.zd[output->sssi.lctop]);
      switch (output->sssi.type) {
      case ISCCP_WATER: 
	fprintf (stderr, "    type:               water\n");
	break;
      case ISCCP_ICE: 
	fprintf (stderr, "    type:               ice\n");
	break;
      default:
	fprintf (stderr, "Error, unknown ISCCP cloud type %d\n", output->sssi.type);
	return -1;
      }
      fprintf (stderr, "    optical thickness:  %f\n", 
	       output->sssi.tautot);
      fprintf (stderr, "    reflectivity:       %f\n", 
	       output->sssi.ref);
    }
    fflush (stderr);

    sss_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    sss_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by dcheck()\n", status);
      return status;
    }
    
    /* call RTE solver */
    sss_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1, output->atm.nmom+1, output->pmom);
    F77_FCN  (sssi)(&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
		    sss_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
		    &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &output->atm.nzout, rte_in->utau,
		    &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
		    &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
		    &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
		    &output->alb.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
		    &rte_in->temis, &rte_in->planck,
		    &rte_in->onlyfl, &rte_in->accur,
		    rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
		    &output->atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
		    rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		    rte_out->dfdt, rte_out->uavg, 
		    sss_uu, rte_out->albmed, rte_out->trnmed, 
		    rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		    &(input.brdf),
		    &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
		    &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
		    &(output->sssi.ref), &(output->sssi.lctop));
    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }


    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, sss_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, sss_uu, rte_out->uu);

    free(sss_pmom);
    free(sss_u0u);
    free(sss_uu);
#else
  fprintf (stderr, "Error, SSSI solver not available!\n");
  return -1;
#endif

    break;

  case SSS2:

#if HAVE_SSS2

    /* get cloud reflectivity from lookup table */
    status = read_isccp_reflectivity (output->sssi.type, output->atm.sza_t[iv], output->sssi.tautot, 
				      input.filename[FN_PATH], input.quiet, &(output->sssi.ref));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_isccp_reflectivity()\n", status);
      return status;
    }

    if (input.verbose) {
      fprintf (stderr, "*** SSS2 cloud properties:\n");
      fprintf (stderr, "    top level:          zd[%d] = %f\n", 
	       output->sssi.lctop, output->atm.zd[output->sssi.lctop]);
      switch (output->sssi.type) {
      case ISCCP_WATER: 
	fprintf (stderr, "    type:               water\n");
	break;
      case ISCCP_ICE: 
	fprintf (stderr, "    type:               ice\n");
	break;
      default:
	fprintf (stderr, "Error, unknown ISCCP cloud type %d\n", output->sssi.type);
	return -1;
      }
      fprintf (stderr, "    optical thickness:  %f\n", 
	       output->sssi.tautot);
      fprintf (stderr, "    reflectivity:       %f\n", 
	       output->sssi.ref);
    }
    fflush (stderr);

    sss_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    sss_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by dcheck()\n", status);
      return status;
    }
    
    /* call RTE solver */
    sss_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1,  output->atm.nmom+1, output->pmom);
    sss2 (&output->atm.nlyr, output->dtauc, output->ssalb, &output->atm.nmom,
	  sss_pmom, output->atm.microphys.temper, &output->wl.wvnmlo_t[iv],
	  &output->wl.wvnmhi_t[iv], &rte_in->usrtau, &output->atm.nzout, rte_in->utau,
	  &input.rte.nstr, &rte_in->usrang, &input.rte.numu, input.rte.umu, 
	  &input.rte.nphi, input.rte.phi, &rte_in->ibcnd, &rte_in->fbeam, 
	  &rte_in->umu0, &output->atm.phi0_t[iv], &rte_in->fisot, &rte_in->lamber, 
	  &output->alb.albedo_t[iv], &rte_in->btemp, &rte_in->ttemp, 
	  &rte_in->temis, &rte_in->planck,
	  &rte_in->onlyfl, &rte_in->accur,
	  rte_in->prndis2, rte_in->header, &output->atm.nlyr, 
	  &output->atm.nzout, &input.rte.maxumu, &input.rte.maxphi, &output->atm.nmom,
	  rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
	  rte_out->dfdt, rte_out->uavg, 
	  sss_uu, rte_out->albmed, rte_out->trnmed, 
	  rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
	  &(input.brdf),
	  &(output->rpv.rho0_t[iv]), &(output->rpv.k_t[iv]), &(output->rpv.theta_t[iv]),
	  &(input.cm.u10), &(input.cm.pcl), &(input.cm.sal), 
	  &(output->sssi.ref), &(output->sssi.lctop));
    
    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }


    /* convert temporary Fortran arrays to permanent result */
    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, sss_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, sss_uu, rte_out->uu);

    free(sss_pmom);
    free(sss_u0u);
    free(sss_uu);
#else
  fprintf (stderr, "Error, SSS2 solver not available!\n");
  return -1;
#endif

    break;

  case POLRADTRAN:
#if HAVE_POLRADTRAN
    rte_in->pol.nummu    = input.rte.nstr/2+input.rte.numu;
    rte_in->pol.albedo   = (double) output->alb.albedo_t[iv];
    rte_in->pol.btemp    = (double) (rte_in->btemp);
    rte_in->pol.flux     = (double) (rte_in->fbeam*rte_in->umu0); /* Polradtran wants flux 
                                                                     on horizontal surface */
    rte_in->pol.mu       = (double) (rte_in->umu0);
    rte_in->pol.sky_temp = (double) (rte_in->ttemp);
    rte_in->pol.wavelength = 0.0; /* Not used for solar source */
    for (lu=0; lu<output->atm.nlyr; lu++) {
      /* rte_in->pol.gas_extinct[lu] = (double) (1.0-output->ssalb[lu])*output->dtauc[lu]; 
      ????? CE: Something wrong here?? If not set to zero, results for radiances are totally wrong */ 
      rte_in->pol.gas_extinct[lu] = 0.0;
    }

    polradtran_down_flux = (double *) calloc (output->atm.nzout*input.rte.pol_nstokes, sizeof(double));
    polradtran_up_flux   = (double *) calloc (output->atm.nzout*input.rte.pol_nstokes, sizeof(double));
    polradtran_down_rad  = (double *) calloc (output->atm.nzout*(input.rte.pol_aziorder+1)*
                         		      (input.rte.nstr/2+input.rte.numu)*input.rte.pol_nstokes,
					      sizeof(double));
    polradtran_up_rad    = (double *) calloc (output->atm.nzout*(input.rte.pol_aziorder+1)*
					      (input.rte.nstr/2+input.rte.numu)*input.rte.pol_nstokes,
					      sizeof(double));
    for (iu=0; iu<rte_in->pol.nummu; iu++) 
      rte_out->polradtran_mu_values[iu] = (double) fabs(output->mu_values[iu]);
    /* Take fabs since radtran will calculate both up_rad and down_rad for the value of mu.  */
    /* The user wants one of these. That is sorted out in the output section of uvspec_lex.l */

    F77_FCN (radtran) (&input.rte.pol_nstokes, &(rte_in->pol.nummu), 
		       &input.rte.pol_aziorder, &input.rte.pol_max_delta_tau,
		       &input.rte.pol_src_code, input.rte.pol_quad_type,
		       rte_in->pol.deltam, &(rte_in->pol.flux), &(rte_in->pol.mu),
		       &(rte_in->pol.btemp), (rte_in->pol.ground_type),
		       &(rte_in->pol.albedo), &(rte_in->pol.ground_index),
		       &(rte_in->pol.sky_temp), &(rte_in->pol.wavelength),
		       &output->atm.nlyr, (rte_in->pol.height), 
		       (rte_in->pol.temperatures), (rte_in->pol.gas_extinct),
		       output->atm.pol_scat_files, &output->atm.nzout,
		       (rte_in->pol.outlevels), rte_out->polradtran_mu_values,
		       polradtran_up_flux, polradtran_down_flux,
		       polradtran_up_rad, polradtran_down_rad);
    
    fortran2c_2D_double_ary_noalloc (output->atm.nzout, 
				     input.rte.pol_nstokes, 
				     polradtran_up_flux,
				     rte_out->polradtran_up_flux);
    
    fortran2c_2D_double_ary_noalloc (output->atm.nzout, 
				     input.rte.pol_nstokes, 
				     polradtran_down_flux,
				     rte_out->polradtran_down_flux);

    
    if (input.rte.pol_aziorder > 0) {
      fortran2c_4D_double_ary_noalloc (output->atm.nzout, 
				       input.rte.pol_aziorder+1,
				       input.rte.nstr/2+input.rte.numu, 
				       input.rte.pol_nstokes, 
				       polradtran_down_rad,
				       rte_out->polradtran_down_rad_q);
      
      fortran2c_4D_double_ary_noalloc (output->atm.nzout, 
				       input.rte.pol_aziorder+1,
				       input.rte.nstr/2+input.rte.numu, 
				       input.rte.pol_nstokes, 
				       polradtran_up_rad,
				       rte_out->polradtran_up_rad_q);

      fourier2azimuth (rte_out->polradtran_down_rad_q, rte_out->polradtran_up_rad_q,
		       rte_out->polradtran_down_rad, rte_out->polradtran_up_rad, 
		       output->atm.nzout, input.rte.pol_aziorder,
		       input.rte.nstr, input.rte.numu, input.rte.pol_nstokes,
		       input.rte.nphi, input.rte.phi);
    }


    free(polradtran_down_flux);
    free(polradtran_up_flux);

    free(polradtran_down_rad);
    free(polradtran_up_rad);

#else
    fprintf (stderr, "Error: RTE polradtran solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Get solver and rebuild uvspec.\n");
    return -1;
#endif	
    break;

  case QDISORT:
#if HAVE_QDISORT
    fprintf (stderr, "Here qdisort will come\n");
#endif
    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom  = c2fortran_3D_float_ary(output->atm.nlyr, 1, input.rte.nstr+1, output->pmom);
    disort_u0u   = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu    = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));
    sdisort_sig  = (float *) calloc (output->atm.nlyr+1, sizeof(float));

    rte_in->gsrc = 1;
    if ( rte_in->gsrc ) 
       if ((status = ASCII_calloc_float_3D (&(rte_in->qsrc), input.rte.maxphi,
					 output->atm.nlyr+1, input.rte.nstr)) != 0) return status;
    qsrc         = c2fortran_3D_float_ary(input.rte.maxphi, output->atm.nlyr+1, input.rte.nstr, rte_in->qsrc);

    if ((status  = ASCII_calloc_float(&sdisort_denstab, output->atm.microphys.nsza_denstab, output->atm.nlyr+1)) !=0)
      return status;
    if (output->atm.microphys.denstab_id > 0) {
      for (lu=0; lu<=output->atm.nlyr; lu++) {
	sdisort_sig[lu] = output->crs.crs_amf[iv][lu];
	for (is=0; is<output->atm.microphys.nsza_denstab; is++)
	  sdisort_denstab[is][lu] = output->atm.microphys.denstab_amf[is][lu];
      }

      tosdisort_denstab = c2fortran_2D_float_ary (output->atm.microphys.nsza_denstab, 
						  output->atm.nlyr+1, sdisort_denstab);
    }

    F77_FCN (qdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &output->atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->alb.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &output->atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg,
		      disort_uu, disort_u0u, &input.rte.nscat,
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		      &input.rte.nrefrac, &input.rte.ichap, output->atm.microphys.refind,
		      &output->atm.microphys.nsza_denstab, output->atm.microphys.sza_denstab,
 		      sdisort_sig, tosdisort_denstab,
		      output->dtauc_md, &(rte_in->gsrc), qsrc );

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);
    if ( rte_in->gsrc )     free(qsrc);

    break;

  case SDISORT:

    status = F77_FCN (dcheck) (&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			       &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;

    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom  = c2fortran_3D_float_ary(output->atm.nlyr, 1, input.rte.nstr+1, output->pmom);
    disort_u0u   = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu    = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));
    sdisort_sig  = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    if ((status  = ASCII_calloc_float(&sdisort_denstab, output->atm.microphys.nsza_denstab, output->atm.nlyr+1)) !=0)
      return status;
    if (output->atm.microphys.denstab_id > 0) {
      for (lu=0; lu<=output->atm.nlyr; lu++) {
	sdisort_sig[lu] = output->crs.crs_amf[iv][lu];
	for (is=0; is<output->atm.microphys.nsza_denstab; is++)
	  sdisort_denstab[is][lu] = output->atm.microphys.denstab_amf[is][lu];
      }

      tosdisort_denstab = c2fortran_2D_float_ary (output->atm.microphys.nsza_denstab, 
						  output->atm.nlyr+1, sdisort_denstab);
    }

    F77_FCN (sdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &output->atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->alb.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &output->atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg,
		      disort_uu, disort_u0u, &input.rte.nscat,
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup,
		      &input.rte.nrefrac, &input.rte.ichap, output->atm.microphys.refind,
		      &output->atm.microphys.nsza_denstab, output->atm.microphys.sza_denstab,
 		      sdisort_sig, tosdisort_denstab, output->dtauc_md);

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;
  case SPSDISORT:

    status = F77_FCN (dcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			       &input.rte.numu, &input.rte.nphi); 
	
    if (status!=0)
      return status;

    sdisort_beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
    disort_pmom = c2fortran_3D_float_ary(output->atm.nlyr, 1,  input.rte.nstr+1, output->pmom);
    disort_u0u  = (float *) calloc (output->atm.nzout*input.rte.maxumu, sizeof(float));
    disort_uu   = (float *) calloc (output->atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

    F77_FCN (spsdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, disort_pmom, 
		      output->atm.microphys.temper, &(output->wl.wvnmlo_t[iv]), 
		      &(output->wl.wvnmhi_t[iv]), &(rte_in->usrtau), &output->atm.nzout, rte_in->utau,
		      &input.rte.nstr, &(rte_in->usrang), &input.rte.numu, input.rte.umu, 
		      &input.rte.nphi, input.rte.phi, &(rte_in->fbeam), sdisort_beta, &(rte_in->nil),
		      &(rte_in->umu0), &output->atm.phi0_t[iv], &(rte_in->newgeo), output->atm.zd,
		      &(rte_in->spher), &radius, &(rte_in->fisot), &output->alb.albedo_t[iv], 
		      &(rte_in->btemp), &(rte_in->ttemp), &(rte_in->temis), &input.rte.deltam, &(rte_in->planck),
		      &(rte_in->onlyfl), &(rte_in->accur), &(rte_in->quiet), rte_in->ierror_s, 
		      rte_in->prndis, rte_in->header, &output->atm.nlyr, 
		      &output->atm.nzout, &input.rte.maxumu, &input.rte.nstr,
		      &input.rte.maxphi, rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		      rte_out->dfdt, rte_out->uavg, 
		      disort_uu, disort_u0u, 
		      rte_out->uavgdn, rte_out->uavgso, rte_out->uavgup);

    /* convert temporary Fortran arrays to permanent result */

    fortran2c_2D_float_ary_noalloc (output->atm.nzout, 
				    input.rte.maxumu, disort_u0u, rte_out->u0u);
    
    fortran2c_3D_float_ary_noalloc (input.rte.maxphi, output->atm.nzout, 
				    input.rte.maxumu, disort_uu, rte_out->uu);

    free(sdisort_beta);
    free(disort_pmom);
    free(disort_u0u);
    free(disort_uu);

    break;
  case TWOSTR:
  case TWOSTRPP:

    status = F77_FCN (tcheck)(&output->atm.nlyr, &output->atm.nzout, &input.rte.nstr, 
			      &input.rte.numu, &input.rte.nphi); 
    
    if (status!=0)
      return status;

    twostr_gg = (float *) calloc (output->atm.nlyr, sizeof(float));
    for (lu=0; lu<output->atm.nlyr; lu++) 
      twostr_gg[lu]=output->pmom[lu][0][1];
		

    /* ??? no need for thermal below 2um;                 ??? */
    /* ??? need to do that to avoid numerical underflow;  ??? */
    /* ??? however, this could be done without actually   ??? */
    /* ??? calling the solver                             ??? */
    if (rte_in->planck && output->wl.lambda_t[iv] < 2000.0) {   
      rte_in->planck = 0;
      planck_tempoff=1;
    }

    F77_FCN  (twostr)(&output->alb.albedo_t[iv], &(rte_in->btemp), &input.rte.deltam,
		      output->dtauc, &(rte_in->fbeam), &(rte_in->fisot), twostr_gg, rte_in->header, 
		      rte_in->ierror_t, &output->atm.nlyr, &output->atm.nzout, &(rte_in->newgeo), 
		      &output->atm.nlyr, &(rte_in->planck), &output->atm.nzout,
		      rte_in->prntwo, &(rte_in->quiet), &radius, &(rte_in->spher), 
		      output->ssalb, &(rte_in->temis),  
		      output->atm.microphys.temper, &(rte_in->ttemp), &(rte_in->umu0), 
		      &(rte_in->usrtau), rte_in->utau, &(output->wl.wvnmlo_t[iv]),
		      &(output->wl.wvnmhi_t[iv]), output->atm.zd, rte_out->dfdt, rte_out->flup, 
		      rte_out->rfldir, rte_out->rfldn, rte_out->uavg);
    free(twostr_gg);

    for (lev=0;lev<output->atm.nzout;lev++) {
      rte_out->uavgso[lev] = NAN;
      rte_out->uavgdn[lev] = NAN;
      rte_out->uavgup[lev] = NAN;
    }

    if (planck_tempoff) {
      rte_in->planck = 1;
      planck_tempoff=0;
    }

    break;
  case SUCORD:
#if HAVE_SOS
    ASCII_calloc_float (&pmom_sos, output->atm.nlyr, output->atm.nmom+1);
    
    for(lu=0; lu<output->atm.nlyr; lu++)
      for(k=0; k<output->atm.nmom+1; k++)
        pmom_sos[lu][k]=output->pmom[lu][0][k];
    
    
    status = sos (output->atm.nlyr, rte_in->newgeo, input.rte.nstr, 
		  input.rte.nscat, output->alb.albedo_t[iv],
		  radius, output->atm.zd, output->ssalb, pmom_sos,
		  output->dtauc, 
		  output->atm.sza_t[iv], output->atm.nzout, 
		  input.rte.numu, input.rte.umu,
		  rte_in->utau, 
		  rte_out->rfldir, rte_out->rfldn, rte_out->flup, 
		  rte_out->uavgso, rte_out->uavgdn, rte_out->uavgup,
		  rte_out->u0u);
      
    if (status!=0) {
      fprintf (stderr, "Error %d returned by sos()\n", status);
      return status;
    }
    break;
#else
    fprintf (stderr, "Error: sos solver not included in uvspec build.\n");
    fprintf (stderr, "Error: Please contact arve.kylling@nilu.no\n");
    return -1;
#endif	

  case NULLSOLVER:
    /* do nothing */
    break;

  default:
    fprintf (stderr, "Error: RTE solver %d not yet implemented, call_solver (solve_rte.c)\n", input.rte.solver);
    return -1;
  }

  if (verbose) {
    end = clock();
    fprintf (stderr, "*** last solver call: %f seconds\n", 
	     ((double) (end - start)) / CLOCKS_PER_SEC);
  }
      
  return 0;    /* if o.k. */
}


static int init_rte_input (rte_input *rte, input_struct input, output_struct *output)
{
  int i=0, found=0;
  int ip=0, is=0, lu=0;
  int status = 0;

  strcpy(rte->header, "");
  rte->accur = 1.0e-3;


  switch (input.source) {
  case SRC_NONE:
    rte->planck = 0; 
    rte->fbeam  = 0.0;
    rte->fisot  = 0.0;
    break;

  case SRC_SOLAR:
    rte->planck=0; 

    if (input.rte.fisot > 0) { 
      rte->fbeam = 0.0; 
      rte->fisot = 1.0; 
    }
    else { 
      rte->fbeam = 1.0; 
      rte->fisot = 0.0; 
    }

    break;

  case SRC_THERMAL:
    rte->planck = 1; 
    rte->fbeam  = 0.0;
    rte->fisot  = 0.0;
    
    rte->btemp = output->surface_temperature;
    rte->ttemp = output->atm.microphys.temper[0];

    rte->temis = 0.0;
    
    break;
  default:
    fprintf (stderr, "Error, unknown source\n");
    return -1;
  }

  rte->umu0  = 0.0;

  rte->ierror_d[0] = 0;
  rte->ierror_s[0] = 0;
  rte->ierror_t[0] = 0;

  for (i=0; i<7; i++)
    rte->prndis[i]  = 0;

  for (i=0; i<5; i++)
    rte->prndis2[i] = 0;

  for (i=0; i<2; i++)
    rte->prntwo[i]  = 0;

  rte->ibcnd=0;
  rte->lamber=1;
  rte->newgeo=1;
  rte->nil=0;
  rte->onlyfl=1;
  rte->quiet = 0;
  rte->usrang = 0;
  rte->usrtau = 1;

  if (input.rte.solver==TWOSTRPP) /* plane-parallel twostr */
    rte->spher = 0;
  else 
    rte->spher = 1;

  rte->hl   = (float *) calloc (input.rte.maxumu+1, sizeof(float));
  rte->utau = (float *) calloc (output->atm.nzout,    sizeof(float));


  if (input.rte.numu > 0) { 
    rte->onlyfl = 0; 
    rte->usrang = 1; 
  }

  /* PolRadtran */
  rte->pol.deltam       = "Y";
  rte->pol.ground_type  = "L";
  strcpy(rte->pol.polscat, "");
  rte->pol.albedo       = 0.0;
  rte->pol.btemp        = 0.0;
  rte->pol.flux         = 1.0;
  rte->pol.gas_extinct  = NULL;
  rte->pol.height       = NULL;
  rte->pol.mu           = 1.0;
  rte->pol.sky_temp     = 0.0;
  rte->pol.temperatures = NULL;
  rte->pol.wavelength   = 0.0;
  rte->pol.outlevels    = NULL;
  rte->pol.nummu        = 0; 
  
  rte->pol.ground_index.re = 0.0;
  rte->pol.ground_index.im = 0.0;

  /* get the indices of the zout levels in the z-scale */
  rte->pol.outlevels    = (int *) calloc (output->atm.nzout, sizeof(int));
  found = 0;
  for (i=0; i<output->atm.nzout; i++) 
    for (lu=0; lu<output->atm.nlev; lu++) 
      if (fabs(output->atm.zout_sur[i] - output->atm.zd[lu]) <= 0)
	rte->pol.outlevels[found++] = lu+1;

  switch (input.rte.solver) {
  case MONTECARLO:
    /* set number of photons */
    output->mc_photons = input.rte.mc.photons;
    break;
  case SDISORT:
  case SPSDISORT:
  case QDISORT:
  case DISORT:
  case SSS:
  case SSSI:
  case TZS:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      (rte->prndis)[input.rte.prndis[ip]-1] = 1;
    break;
  case DISORT2:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      (rte->prndis2)[input.rte.prndis[ip]-1] = 1;
    break;
  case POLRADTRAN:
    /* Need to do a little checking of polradtran specific input stuff here...*/
    if (found != output->atm.nzout) {
      fprintf (stderr, "*** zout does not correspond to atmosphere file levels.\n");
      fprintf (stderr, "*** zout must do so for the polradtran solver.\n");
      status--;
    }

    if (input.rte.deltam == 0) 
      rte->pol.deltam = "N";
    else if (input.rte.deltam == 1) 
      rte->pol.deltam = "Y";
    
    if ((output->mu_values = (float *) calloc (input.rte.nstr/2+input.rte.numu, sizeof(float))) == NULL)
      return -1;

    if (strncmp(input.rte.pol_quad_type,"E",1) == 0)
      for (i=0; i<input.rte.numu; i++) 
        output->mu_values[input.rte.nstr/2+i] = input.rte.umu[i];
    

    rte->pol.gas_extinct  = (double *) calloc (output->atm.nlyr+1, sizeof(double));
    rte->pol.height       = (double *) calloc (output->atm.nlyr+1, sizeof(double));
    rte->pol.temperatures = (double *) calloc (output->atm.nlyr+1, sizeof(double));

    output->atm.pol_scat_files= (char *) calloc (64*output->atm.nlyr, sizeof(char));      

    for (is=0;is<64*output->atm.nlyr;is++)
      output->atm.pol_scat_files[is] = ' ';

    for (lu=0;lu<output->atm.nlyr;lu++) {
      is = lu*64;
      sprintf(rte->pol.polscat,".scat_file_%03d",lu);
      strcpy(&output->atm.pol_scat_files[is], rte->pol.polscat);
    }

    for (lu=0;lu<=output->atm.nlyr;lu++) {
      rte->pol.height[lu] = (double) lu;  /*  Weird hey???? Well, the story goes as 
					      follows: polradtran wants extinction and
					      scattering in terms of 1/(layerthickness).
					      However, we feed it optical depth. Hence,
					      we need to make delta_height of each layer
					      equal one. One simple remedy is the one
					      used. Arve 15.03.2000 */
      rte->pol.temperatures[lu] = (double) output->atm.microphys.temper[lu];
    }
    break;
  case TWOSTR:
  case TWOSTRPP:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      if (input.rte.prndis[ip]==1 || input.rte.prndis[ip]==2) 
	(rte->prntwo)[input.rte.prndis[ip]-1] = 1;
    break;
  case SUCORD:
  case NULLSOLVER:
    break;
  default:
    fprintf (stderr, "Error: RTE solver %d not yet implemented, init_rte_input (solve_rte.c)\n", input.rte.solver);
    return -1;
    break;
  }    
  
  return status;
}

static void fourier2azimuth (double**** down_rad_rt3, double**** up_rad_rt3,
			     double**** down_rad, double**** up_rad, 
			     int nzout, int aziorder, int nstr, int numu, int nstokes,
			     int nphi, float* phi)
{
  /* For each azimuth and polar angle sum the Fourier azimuth series appropriate
     for the particular Stokes parameter to produce the radiance.
     Only used for the polradtran solver 
  */
  int i=0, j=0, je=0, k=0, lu=0, m=0;
  double sumd=0, sumu=0;
  float phir;
  for (lu=0;lu<nzout;lu++) {
    for (k=0;k<nphi;k++) {
      phir = PI*phi[k]/180.0;
      
      /* Up- and downwelling irradiances at user angles only*/
      /*      for (j=0;j<nstr/2+numu;j++) {  */
      for (j=0;j<numu;j++) {
	je = j+nstr/2;
	for (i=0;i<nstokes;i++) {
	  sumd=0.0;
	  sumu=0.0;
	  for (m=0;m<=aziorder;m++) {
	    if (i < 2) {
	      sumd += cos(m*phir)*down_rad_rt3[lu][m][je][i];
	      sumu += cos(m*phir)*up_rad_rt3[lu][m][je][i];
	    }
	    else {
	      sumd += sin(m*phir)*down_rad_rt3[lu][m][je][i];
	      sumu += sin(m*phir)*up_rad_rt3[lu][m][je][i];
	    }
	  }
          down_rad[lu][k][j][i] = sumd;
	  up_rad[lu][k][j][i] = sumu;
	}
      }
    }
  }
}


/***************************************************************/
/* calc_spectral_heating calculates the divergence of the flux */
/* either by differences of the flux or                        */
/* with the help of the actinic flux                           */
/***************************************************************/

static int calc_spectral_heating (input_struct input, output_struct *output,
				  float *dz, double *rho_mass_zout, float *k_abs, float *k_abs_layer, 
				  int *zout_index, rte_output *rte_out, float *heat, float *w_zout, int iv)
{

  int status  = 0;
  int lz      = NOT_DEFINED_INTEGER;
  int lc      = NOT_DEFINED_INTEGER;
  int nzout   = NOT_DEFINED_INTEGER;
  /* int z_index = NOT_DEFINED_INTEGER; */
  int nlev    = NOT_DEFINED_INTEGER;
  int nlyr    = NOT_DEFINED_INTEGER;
  /* float *Fup = NULL; */
  /* float *Fdn = NULL; */
  float *F_net = NULL;
  float  dFdz = 0;
  float *c_p = 0;
  float  dtheta_dz = 0;
  float *dFdz_array = NULL;
  int lz1=NOT_DEFINED_INTEGER, lz2=NOT_DEFINED_INTEGER, n_lz=NOT_DEFINED_INTEGER;
  float M_AIR  = mol_mass_air/1000.0;  /* molecular weight of air (kg mol-1) */
  float planck_radiance = 0.0;

  float *z_center=NULL;
  float *zout_in_m=NULL;
  int outside=0;
  int start=0;
  int z_index=0;

/*   int additional_verbose_output=FALSE; */
  
  nlev  = output->atm.nlev;
  nlyr  = nlev-1;
  nzout = output->atm.nzout;

/*   /\* additional verbose output *\/ */
/*   if (additional_verbose_output) { */
/*     if (((Fup) = (float *) calloc (nzout, sizeof(float))) == NULL) { */
/*       fprintf (stderr, "Error allocating memory for (Fup) in calc_spectral_heating (solve_rte.c)\n"); */
/*       return -1; */
/*     } */

/*     if (((Fdn) = (float  *) calloc (nzout, sizeof(float))) == NULL) { */
/*       fprintf (stderr, "Error allocating memory for (Fdn) in calc_spectral_heating (solve_rte.c)\n"); */
/*       return -1; */
/*     } */
/*   } */

  if (((F_net) = (float *) calloc (nzout, sizeof(float))) == NULL) {
    fprintf (stderr, "Error allocating memory for (dF) in calc_spectral_heating (solve_rte.c)\n");
    return -1;
  }

  if (((dFdz_array) = (float *) calloc (nzout, sizeof(float))) == NULL) {
    fprintf (stderr, "Error allocating memory for (dFdz_array) in calc_spectral_heating (solve_rte.c)\n");
    return -1;
  }

  if (((c_p) = (float *) calloc (nzout, sizeof(float))) == NULL) {
    fprintf (stderr, "Error allocating memory for (dFdz_array) in calc_spectral_heating (solve_rte.c)\n");
    return -1;
  }

  switch(input.rte.solver) {
  case  DISORT:
  case  SDISORT:
  case  TWOSTR:
  case  TWOSTRPP:
  case  SUCORD:
  case  MONTECARLO:
  case  DISORT2:
  case  SPSDISORT:
  case  QDISORT:
  case  TZS:
  case  SSS:
  case  SSSI:
  case  SSS2:
  case  NULLSOLVER:
    for (lz=0; lz < nzout; lz++) {
/*       if (additional_verbose_output) { */
/*         Fdn[lz] = rte_out->rfldir[lz] + rte_out->rfldn[lz]; */
/*         Fup[lz] = rte_out->flup[lz]; */
/*       } */
      F_net[lz] = (rte_out->rfldir[lz] + rte_out->rfldn[lz]) - rte_out->flup[lz];
    }
    break;
  case POLRADTRAN:  
    for (lz=0; lz < nzout; lz++) {
/*       if (additional_verbose_output) { */
/*         Fdn[lz] = rte_out->polradtran_down_flux[lz][0]; */
/*         Fup[lz] = rte_out->polradtran_up_flux[lz][0]; */
/*       } */
      F_net[lz]  = rte_out->polradtran_down_flux[lz][0] - rte_out->polradtran_up_flux[lz][0];
    }
    break;
  default:
    fprintf (stderr, "Error: unknown solver %d in calc_spectral_heating \n", input.rte.solver);
    return -1;
    break;
  }

  /* calculate normalized (1/incident_flux) spectral heating rate */
  switch (input.heating) {
  case HEAT_LAYER_CD:
  case HEAT_LAYER_FD:

    if (input.heating == HEAT_LAYER_CD) n_lz = nzout;
    if (input.heating == HEAT_LAYER_FD) n_lz = nzout-1;

    for (lz=0; lz<n_lz; lz++) {

      if (input.heating == HEAT_LAYER_CD) {
        if (lz == 0) {
	  lz1 = lz;
	  lz2 = lz+1;}
        else if (lz == output->atm.nzout-1) {
	  lz1 = lz-1;
	  lz2 = lz;}
        else {
          lz1 = lz-1;
          lz2 = lz+1;
        }

        if (lz != 0 && lz != output->atm.nzout-1) {
          /* centered difference */ /* 1.0e+6: convert from cm-3 to m-3 */
          rho_mass_zout[lz] = output->atm.microphys.dens_zout[MOL_AIR][lz] * 1.0e+6 * M_AIR / avogadro;

          /* mass weighted mean of specific heating rates */
          c_p[lz] = output->atm.microphys.c_p[lz];
	}
        else {
          /* boundary, no centered difference possible, (log) average density for forward difference */
          rho_mass_zout[lz] = log_average(output->atm.microphys.dens_zout[MOL_AIR][lz1],
					    output->atm.microphys.dens_zout[MOL_AIR][lz2]) * 1.0e+6 * M_AIR / avogadro;
         
          /* mass weighted mean of specific heating rates, assuming exponential change of density and linear change of c_p */
          c_p[lz] = mass_weighted_average( output->atm.microphys.c_p               [lz1], output->atm.microphys.c_p               [lz2],
                                           output->atm.microphys.dens_zout[MOL_AIR][lz1], output->atm.microphys.dens_zout[MOL_AIR][lz2] );
        }
      }
      else if (input.heating == HEAT_LAYER_FD) {
        /* forward difference */
        lz1 = lz;
        lz2 = lz+1;

        /* effective density for one layer (logarithmic) */ /* 1.0e+6: convert from cm-3 to m-3 */
        rho_mass_zout[lz] = log_average(output->atm.microphys.dens_zout[MOL_AIR][lz1],
                                               output->atm.microphys.dens_zout[MOL_AIR][lz2]) * 1.0e+6 * M_AIR / avogadro;

        /* mass weighted mean of specific heating rates, assuming exponential change of density and linear change of c_p */
        c_p[lz] = mass_weighted_average( output->atm.microphys.c_p               [lz1], output->atm.microphys.c_p               [lz2],
                                         output->atm.microphys.dens_zout[MOL_AIR][lz1], output->atm.microphys.dens_zout[MOL_AIR][lz2] );
      }
      else
        fprintf (stderr, "Error, unknown processing scheme in calc_spectral_heating %d\n", input.processing);

      /* Compute the derivative dF/dz using a two-point formula */

      if (fabs((F_net[lz1]-F_net[lz2])/F_net[lz1]) > 1.0E-6 || fabs((output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2])*rho_mass_zout[lz]) > 1.0E-6) { 
        dFdz      =                       (F_net[lz1]-F_net[lz2])                                 / (output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2]);
      }
      else 
        dFdz = NAN;

      dFdz_array[lz] = 0.001 * dFdz;      /* 1/1000 = km -> m, for dz */
      dtheta_dz = (output->atm.microphys.theta_zout[lz1]-output->atm.microphys.theta_zout[lz2]) / (output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2]);
      dtheta_dz      = 0.001 * dtheta_dz; /* 1/1000 = km -> m, for dz */

/*       if (additional_verbose_output) { */
/*        if (lz==0) { */
/*          fprintf (stderr, " ... calling calc_spectral_heating()\n"); */
/*          if (input.heating == HEAT_LAYER_CD)  fprintf (stderr, " ... calculate heating_rate with centered differences \n"); */
/*          if (input.heating == HEAT_LAYER_FD)  fprintf (stderr, " ... calculate heating_rate with forward differences \n"); */
/*          fprintf (stderr, "\n#--------------------------------------------------------------------------------------------------------------------------------\n"); */
/*          fprintf (stderr, "#lvl      z    l1  l2    z(l1)    z(l2)    E(l1)     E(l2)      dE/dz      |dE/E|        Edn       Eup      Edn/dz        Eup/dz   \n"); */
/*          fprintf (stderr, "#         km              km       km      W/m2      W/m2      W/(m2 m)                  W/m2      W/m2    W/(m2 km)     W/(m2 km) \n"); */
/*          fprintf (stderr, "#----------------------------------------------------------------------------------------------------------------------------------\n"); */
/*        } */

/*        fprintf (stderr, "%3d %9.3f %3d %3d %8.3f %8.3f %9.4f %9.4f %12.5e %10.5e %9.4f %9.4f %12.5e %12.5e\n", */
/*           lz, output->atm.zout_sur[lz], lz1, lz2, output->atm.zout_sur[lz1],output->atm.zout_sur[lz2], */
/*           F_net[lz1],F_net[lz2],dFdz_array[lz],fabs((F_net[lz1]-F_net[lz2])/F_net[lz1]),Fdn[lz],Fup[lz], */
/*           (Fdn[lz1]-Fdn[lz2])/(output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2]),(Fup[lz1]-Fup[lz2])/(output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2])); */
/*       } */

      heat[lz] = dFdz_array[lz] / ( rho_mass_zout[lz] * c_p[lz] );
      w_zout[lz] = (output->atm.microphys.theta_zout[lz]/output->atm.microphys.temper_zout[lz]) * 1.0/dtheta_dz * heat[lz];
    }

    break;
  case HEAT_LOCAL:

    if (((z_center) = (float  *) calloc (nlyr, sizeof(float))) == NULL) {
      fprintf (stderr, "Error allocating memory for (z_center) in calc_spectral_heating (solve_rte.c)\n");
      return -1;
    }

    /* z_center == layer midpoints im m of the atmosphere z-layers above surface */
    /* k_abs_layer == absorption coefficient representative for one layer (layer midpoint is z_center) */
    for (lc=0; lc<nlyr; lc++) {
      k_abs_layer[lc] = (1.0 - output->ssalb[lc]) * output->dtauc[lc] / dz[lc];
      z_center[lc] = output->atm.zd[lc]*1000.0 - dz[lc]/2;  /* 1000 == km -> m */
/*       fprintf (stderr, " k_abs_layer =%e (z_center[%3d] = %6.2f) \n", k_abs_layer[lc], lc, z_center[lc]); */
    }

    /* zout_in_m == levels (layer boundaries) of zout levels in m above surface */
    if (((zout_in_m) = (float  *) calloc (nzout, sizeof(float))) == NULL) {
      fprintf (stderr, "Error allocating memory for (nzout) in calc_spectral_heating (solve_rte.c)\n");
      return -1;
    }

    for (lz=0; lz < nzout; lz++) {
      zout_in_m[lz] = output->atm.zout_sur[lz]*1000.0;
/*       fprintf (stderr, " zout_in_m[%3d] = %7.0f) \n", lz, zout_in_m[lz] ); */
    }

    /* spline interpolation of all level EXEPT LOWEST AND HIGHEST LEVEL */
    /* they are outside of the range of layer midpoints and must therefor be extrapolated !!!! */


    /* if uppermost zout level is above the uppermost layer midpoint (e.g. zout TOA),  */
    /* than extrapolate k_abs from the uppermost 2 layers                              */
    if ( zout_in_m[nzout-1] > z_center[0] ) {
/*       fprintf (stderr, " upper extrapolation: zout[%d]=%8.3f km,  z(layer_midpoint)[%d]=%8.3f km \n", nzout-1, zout_in_m[nzout-1]/1000., 0, z_center[0]/1000.  ); */
      outside = 1;
/*       /\* linear extrapolation *\/ */
/*       k_abs[nzout-1] =  k_abs_layer[0] + dz[0] * (k_abs_layer[1] - k_abs_layer[0]) / */
/* 	                         ((output->atm.zd[2] - output->atm.zd[0])*1000.0); /\* 1000 = km -> m *\/ */
/*                                                            /\* canceled 2/2 in (dz/2)/((z[2]-z[0])/2) *\/ */

      /* exponentiell extrapolation */
      k_abs[nzout-1] = k_abs_layer[0] * pow ( k_abs_layer[0]/k_abs_layer[1] , dz[0]/((output->atm.zd[0] - output->atm.zd[2])*1000.0) );


      fprintf (stderr, " k_abs_layer[     1]=%e, k_abs_layer[     0]=%e, k_abs[nzout-1]=%e \n\n",
               k_abs_layer[1], k_abs_layer[0], k_abs[nzout-1] );
    }


    /* if lowermost zout level is below the lowest layer midpoint (e.g. zout surface),  */
    /* than extrapolate k_abs from the lowermost 2 layers                              */
    if ( zout_in_m[0] < z_center[nlyr-1] ) {
/*       fprintf (stderr, " lower extrapolation: zout[%d]=%8.3f km,  z(layer_midpoint)[%d]=%8.3f km \n", 0, zout_in_m[0]/1000., nlyr-1, z_center[nlyr-1]/1000.  ); */
      outside = outside+1;
      start = 1;
      /* linear extrapolation */
      k_abs[0] = k_abs_layer[nlyr-1] - dz[nlyr-2] *
                          (k_abs_layer[nlyr-1] - k_abs_layer[nlyr-2]) /
                       ((output->atm.zd[nlyr] - output->atm.zd[nlyr-2])*1000.0);  /* 1000 = km -> m */
                                                          /* canceled 2/2 in (dz/2)/((z[2]-z[0])/2) */ 

/*       /\* exponentiell extrapolation *\/ */
/*       k_abs[0] = k_abs_layer[nlyr-1] * pow (k_abs_layer[nlyr-2]/k_abs_layer[nlyr-1],dz[nlyr-2]/((output->atm.zd[nlyr] - output->atm.zd[nlyr-2])*1000.0)); */

/*       fprintf (stderr, " k_abs_layer[nlyr-2]=%e, k_abs_layer[nlyr-1]=%e, k_abs[      0]=%e \n\n", k_abs_layer[nlyr-2], k_abs_layer[nlyr-1], k_abs[0] );  */
    }

/*     /\* interpolate the rest, attention pointer arithmetic, last argument 1 means descending order of z_center *\/ */
/*     status = arb_wvn(nlyr, z_center, k_abs_layer, nzout-outside, zout_in_m+start, k_abs+start, INTERP_METHOD_SPLINE, 1 ); */
/*     if (status != 0) { */
/*       fprintf (stderr, " Error, interpolation of k_abs in calc_spectral_heat (solve_rte.c) \n"); */
/*       return -1; */
/*     } */

    /* interpolate the rest, attention pointer arithmetic, last argument 1 means descending order of z_center */
    status = arb_wvn(nlyr, z_center, k_abs_layer, nzout-outside, zout_in_m+start, k_abs+start, INTERP_METHOD_LINEAR, 1 );
    if (status != 0) {
      fprintf (stderr, " Error, interpolation of k_abs in calc_spectral_heat (solve_rte.c) \n");
      return -1;
    }

/*     for (lz=0; lz < nzout; lz++) */
/*       fprintf (stderr, "start: zout[%4d]=%7.0f, k_abs=%e \n", lz, zout_in_m[lz], k_abs[lz] ); */
    

    /* corrections for possible overshooting of the spline function */
    for (lz=0; lz < nzout; lz++) {

      z_index = zout_index[lz]-1;
/*       fprintf (stderr, "\n zout[%4d]=%7.0f \n", lz, zout_in_m[lz] ); */

      /* check 2 layer below and above if there is a relative change over 150% -> spline overshoot risk -> correction algorithm */
      /* if k_abs is negative, this is obviously an overshoot */

      for ( lc = z_index-2; lc < z_index+3; lc++ ) {
        if ( 0 < lc && lc < nlyr ) {
/*           fprintf (stderr, " check k_abs for: z1=%f, z2=%f \n", z_center[lc-1], z_center[lc] ); */

          if ( k_abs[lz] < 0.0 || k_abs_layer[lc-1] / k_abs_layer[lc] > (1.+1.5) || k_abs_layer[lc-1] / k_abs_layer[lc] < 1/(1.+1.5) ) {

/*             if ( k_abs_layer[lc-1] / k_abs_layer[lc] > (1.+1.5) || k_abs_layer[lc-1] / k_abs_layer[lc] < 1/(1.+1.5) ) */
/*               fprintf (stderr, " relative change over 150 per cent: %e <-> %e\n", k_abs_layer[lc-1], k_abs_layer[lc] ); */
/*             if ( k_abs[lz] < 0.0 )  */
/*               fprintf (stderr, " negative absorption coefficient: %e \n", k_abs[lz] ); */

            if ( z_index == 0 ) {
              /* top of atmosphere: make only correction, if k_abs negative, otherwise it's OK */
              if ( k_abs[lz] < 0.0 )
                k_abs[nzout-1] = k_abs_layer[z_index];
            }
            else if ( z_index-1 == nlyr-1 ) {
              /* surface: take value from lowest layer as level value */
/*               fprintf (stderr, " surface: k_abs_old =%e, k_abs_new=%e  \n", */
/*                          k_abs[0], k_abs_layer[nlyr-1] ); */
              k_abs[0]=k_abs_layer[nlyr-1];
            }
            else {
              /* somewhere inside the profile not at the boundaries */
              /* in a layer nearby, there is probably a cloud       */
              /* change to linear interpolation for this level      */
              k_abs[lz] = (k_abs_layer[z_index-1]*dz[z_index] + k_abs_layer[z_index]*dz[z_index-1]) / (dz[z_index] + dz[z_index-1]);
/*               fprintf (stderr, " k_abs_layer[%d]=%e (zd=%7.0f) k_abs_layer[%d]=%e (zd=%7.0f)  \n", */
/*                        z_index-1, k_abs_layer[z_index-1], z_center[z_index-1], */
/*                        z_index,   k_abs_layer[z_index]  , z_center[z_index] ); */
            }
            break;
          }
        }
      }
    }


/*     for (lz=0; lz < nzout; lz++) */
/*       fprintf (stderr, "changed: lz=%4d, %7.0f, %e \n", lz, zout_in_m[lz], k_abs[lz] ); */


    for (lz=0; lz < nzout; lz++) {

      /* level (local) property */                                          /* 1.0e+6: convert from cm-3 to m-3 */
      rho_mass_zout[lz] = output->atm.microphys.dens_zout[MOL_AIR][lz] * 1.0e+6 * M_AIR / avogadro;

      /* correction of the emission term of the heating rate, when calculated with actinic flux */
      if (input.source == SRC_THERMAL) {
        F77_FCN (cplkavg) (&(output->wl.wvnmlo_t[iv]), &(output->wl.wvnmhi_t[iv]),
                           &(output->atm.microphys.temper_zout[lz]),&(planck_radiance));
      }
      /* (else (in the solar case) planck_radiance == 0) */

      
      heat[lz] = k_abs[lz] * 4.0 * PI * (rte_out->uavg[lz]-planck_radiance) / (rho_mass_zout[lz] * output->atm.microphys.c_p[lz] );

      /* calculate dtheta_dz */
      /* centered differences (like HEAT_LAYER_CD) */
      if (lz == 0) { /* top */
	lz1 = lz;
	lz2 = lz+1;}
      else if (lz == output->atm.nzout-1) { /* bottom */
	lz1 = lz-1;
	lz2 = lz;}
      else { /* normal case */
        lz1 = lz-1;
        lz2 = lz+1;
      }

      dtheta_dz = (output->atm.microphys.theta_zout[lz1]-output->atm.microphys.theta_zout[lz2]) / 
                              (output->atm.zout_sur[lz1]-output->atm.zout_sur[lz2]);
      dtheta_dz      = 0.001 * dtheta_dz; /* 1/1000 = km -> m, for dz */

      w_zout[lz] = (output->atm.microphys.theta_zout[lz]/output->atm.microphys.temper_zout[lz]) * 1.0/dtheta_dz * heat[lz];

    }

    break;
  default:
    fprintf (stderr, "Error, wrong or unknown processing scheme %d in calc_spectral heating \n", input.processing);
    return -1;
  }

/*   if (additional_verbose_output) { */
/*     free(Fup); */
/*     free(Fdn); */
/*   } */

  free(F_net);
  free(dFdz_array);
  free(c_p);

  return status;
}



static float ***calloc_abs3d (int Nx, int Ny, int Nz, int *threed) 
{
  int lc=0, lx=0;
  float ***tmp;
  
  tmp = calloc (Nz, sizeof(float **));
  if (tmp==NULL)
    return NULL;

  for (lc=0; lc<Nz; lc++)
    if (threed[lc]) {
      tmp[lc] = calloc (Nx, sizeof (float *));
      if (tmp[lc]==NULL)
	return NULL;

      for (lx=0; lx<Nx; lx++) {
	tmp[lc][lx] = calloc (Ny, sizeof (float));
	if (tmp[lc][lx]==NULL)
	  return NULL;
      }
    }
  
  return tmp;
}
      


static void free_abs3d (float ***abs3d, int Nx, int Ny, int Nz, int *threed) 
{
  int lc=0, lx=0;
  
  for (lc=0; lc<Nz; lc++)
    if (threed[lc]) {
      for (lx=0; lx<Nx; lx++)
	free (abs3d[lc][lx]);
      free (abs3d[lc]);
    }

  free (abs3d);
}
      

static float ****calloc_spectral_abs3d (int Nx, int Ny, int Nz, int nlambda, int *threed) 
{
  int lx=0, ly=0, lc=0;
  float ****tmp=NULL;


  if ((tmp = (float ****) calloc (Nz, sizeof (float ***))) == NULL)  
    return NULL;

    
  for (lc=0; lc<Nz; lc++) {
    if (threed[lc]) {
      if ((tmp[lc] = (float ***) calloc (Nx, sizeof (float **))) == NULL)
	return NULL;
      
      for (lx=0; lx<Nx; lx++) {
	if ((tmp[lc][lx] = (float **) calloc (Ny, sizeof (float *))) == NULL) 
	  return NULL;
	
	for (ly=0; ly<Ny; ly++)
	  if ((tmp[lc][lx][ly] = (float *) calloc (nlambda, sizeof (float))) == NULL) 
	    return NULL;
      }    
    }
  }

  return tmp;
}


/** 
 * *generate_effective_cloud*: Generates an effective cloud assuming random overlap (as in ECHAM) 
 * given cloud profiles and cloud fraction. 
 * 
 * @param input uvspec input structure
 * @param output uvspec output structure 
 * @param save_cloud unmodified cloud properties for given wavelength
 * @param iv wavelength index
 * @param iq subband index
 * @param verbose flag for verbose output
  * 
 */
static int generate_effective_cloud (input_struct input, output_struct *output, save_optprop *save_cloud, 
                                     int iv, int iq, int verbose)
{
   
  int lc=0, i=0;
  float tauw=0.0, taui=0.0, taua=0.0, taum=0.0, taur=0.0, tau_clear=0.0, tau_cloud=0.0; 
  float g1d=0.0, g2d=0.0, fd=0.0, g1i=0.0, g2i=0.0, fi=0.0, gi=0.0, gw=0.0;
  float g1a=0.0, g2a=0.0, fa=0.0, ga=0.0;
  float ssai=0.0, ssaw=0.0, ssaa=0.0;
  float *ssa=NULL, *tau=NULL, *g=NULL;

  /* Reflectivity of underlying layer, should be zero, because here
     we calculate just the effective optical thickness of one
     layer. This optical thickness is used in the RTE solver, where
     then of course the reflectivity of the neighbour layers is
     considered. */
  float pref=0.0;
  /* Effective solar zenith angle, accounts for the decrease of the
     direct solar beam and the corresponding increase of the diffuse
     part of the radiation. */
  float mu_eff=0.0; 
  /* Diffusivity factor, 1.66 in ECHAM */
  float r = 1.66;
  /* Effective cloudyness */
  float C_eff=0.0, product=1.0; 
  /* solar zenith angle */
  double mu =0.0;
  /* Output variables of swde */
  float pre1=0.0, pre2=0.0, ptr2=0.0, dummy = 0.0;
  /* Transmission of clear and cloudy parts and of the layer */ 
  float transmission_clear=0.0, transmission_cloud=0.0, transmission_layer=0.0;
  /* Variables for iteratiom */
  float taueff=0.0;

  /* Set to 1 if you want to include absorption, aerosol, or rayleigh */
  int include_absorption = 1;
  int include_aerosol    = 1;
  int include_rayleigh   = 1;

  ssa = (float *) calloc (output->atm.nlev-1, sizeof(float));
  tau = (float *) calloc (output->atm.nlev-1, sizeof(float));
  g   = (float *) calloc (output->atm.nlev-1, sizeof(float));
      
  mu = cos( output->atm.sza_t[iv] * PI / 180);
  
  /*fprintf (stderr, "...Calculate effective optical depth...\n");*/ 
  
  /* scattering properties */
  for (lc=0; lc<output->atm.nlev-1; lc++) {
    
    /* Ice and water clouds */
    tauw  = save_cloud->tauw[lc];
    taui  = save_cloud->taui[lc];
       
    g1d   = save_cloud->g1d[lc]; 
    g2d   = save_cloud->g2d[lc]; 
    fd    = save_cloud->fd[lc]; 
    gw    = g1d*fd+(1.0-fd)*g2d;

    g1i   = save_cloud->g1i[lc];
    g2i   = save_cloud->g2i[lc];
    fi    = save_cloud->fi[lc];  
    gi    = g1i*fi+(1.0-fi)*g2i;
    
    ssaw  = save_cloud->ssaw[lc];
    ssai  = save_cloud->ssai[lc];

    
    /* molecular absorption */
    taum  = output->atm.optprop.tau_molabs_t[lc][iv][iq];

    /* aerosol */
    taua  = output->aer.optprop.dtau[iv][lc];
    ssaa  = output->aer.optprop.ssa[iv][lc];

    g1a   = output->aer.optprop.g1[iv][lc];
    g2a   = output->aer.optprop.g2[iv][lc];
    fa    = output->aer.optprop.ff[iv][lc];
    ga    = g1a*fa+(1.0-fa)*g2a;
    
    /*Rayleigh*/
    switch(input.ck_scheme) {
    case CK_FU:
      taur = output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
      break;

    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
    case CK_AVHRR_KRATZ:
    case CK_GENERIC:
    case CK_LOWTRAN:
    case CK_NONE:
      taur = output->atm.optprop.tau_rayleigh_t[lc][iv][0];
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", input.ck_scheme);
      return -1;

      break;
    }
    
    /* Calculate mean optical properties of the layer */
  
    /* total optical thickness */
    tau_cloud = tauw+taui;       /* Water and Ice cloud */
    tau_clear = taum+taua+taur;  /* Molecular, Aerosol, and Rayleigh */
        
    /* Calculate effective Cloudiness */
    tau[lc] = tau_cloud+tau_clear;
    /* fprintf (stderr,"tau [lc] %g, tau_clear %g \n", tau[lc], tau_clear); */
    /* single scattering albedo */
    ssa[lc] =  (tauw*ssaw + taui*ssai)/(tauw+taui+taua+taur+taum);
    /* asymmetry parameter */
    g[lc]=0.0;
    if((tauw || taui) != 0.0)
      g[lc] = (tauw*gw*ssaw + taui*gi*ssai)/(ssaw*tauw + ssai*taui);
   
    /* Calculate effective zenith angle */
    for (i=lc; i>=0; i--)
      product *= 1-output->cf.cf[i]*(1-exp(-((1-ssa[i]*g[i]*g[i])*(tau[lc]))/mu));
    
    C_eff = 1.0-product;
    mu_eff = 1.0/((1.0-C_eff)/mu+r*C_eff);
    
    tau[lc]=tau_cloud+tau_clear;

    if (!include_absorption) {
      tau_clear=0.0; 
      tau[lc]=tau_cloud;
    }
    if (include_aerosol)
      tau_clear += taua;
    if (include_rayleigh)
      tau_clear += taur;


    if((tauw || taui) != 0.0) {
      /*Initialize inputs for swde.*/
      pre1=0.0;
      pre2=0.0;
      transmission_cloud=0.0;
      transmission_clear=0.0; 
      ptr2=0.0;
      dummy=0.0;
    
      /*fprintf (stderr, " g %g, pref %g, prmuz %g, tau[lc] %g tau_clear %g, ptr1 %g, ptr2 %g ssa[lc]%g \n", 
        g[lc], pref, acos(mu_eff)*180/PI, tau[lc], tau_clear, transmission_cloud, ptr2, ssa[lc]);*/
    
      /* Perform radiative transfer (twostream) with scaled optical properties
         to calculate effective optical thickness. */
      /* Scattering + sbsorption + rayleigh + aerosol*/
      F77_FCN (swde)(&g[lc], &pref, &mu_eff, &tau[lc], &ssa[lc], &pre1, &pre2, &transmission_cloud, &ptr2);
      
      ptr2=0.0;
      /* Only absorption, rayleigh, aerosol */
      F77_FCN (swde)(&g[lc], &pref, &mu_eff, &tau_clear, &ssa[lc], &pre1, &pre2, &transmission_clear, &ptr2);
      
      /* Transmission of the layer */
      transmission_layer= output->cf.cf[lc]*transmission_cloud+(1.0- output->cf.cf[lc])*transmission_clear;
      /*fprintf (stderr, "transmission_cloud %g, transmission_clear %g, transmission_layer %g cloud cover %g \n", 
        transmission_cloud, transmission_clear, transmission_layer, output->cf.cf[lc]);*/
      
      /* Calculate effective optical thickness */
      taueff = zbrent_taueff(mu_eff, g[lc], ssa[lc],
                             transmission_cloud, transmission_layer,
                             tau_clear, tau[lc], 0.00001);  
    
      
      /*fprintf (stderr, "test %d %g  %g %g %g %g %g %g %g \n", lc, 
        output->atm.zd[lc+1]+output->alt.altitude, tau[lc], mu_eff, 
        output->cf.cf[lc], transmission_cloud, transmission_clear, transmission_layer, taueff);
        if (taueff > 0 && taueff < 1e-7)
        fprintf (stderr, "taueff: %g \n" , taueff);*/
      
      /* Set the optical properties to be used in rte calculation.*/

      /* wc is now representing both, water and ice clouds */
      output->wc.optprop.dtau [iv][lc] = taueff; 
      output->wc.optprop.g1   [iv][lc]= g[lc]; 
      output->wc.optprop.g2   [iv][lc]= 0.0; 
      output->wc.optprop.ff   [iv][lc]= 1.0;
      output->wc.optprop.ssa  [iv][lc]= ssa[lc];
    
      /* ic is now not needed any more */
      output->ic.optprop.dtau [iv][lc] = 0.0; 
      output->ic.optprop.g1   [iv][lc]= 0.0; 
      output->ic.optprop.g2   [iv][lc]= 0.0; 
      output->ic.optprop.ff   [iv][lc]= 0.0;
      output->ic.optprop.ssa  [iv][lc]= 0.0;
    }
  }
  
  free(ssa);
  free(tau);
  free(g);
  return 0;
}


/** 
 * *zbrent_taueff* returns the effective optical depth of a layer.
 *
 * The function is a slightly modified version of the "zbrent" function in the 
 * "Numerical Recipes in C" (p. 352 ff.) for finding roots of an arbitrary function. 
 * The function is here the ECHAM radiative tarnsfer routine swde.f and the root of it 
 * is the effective optical thickness.
 * 
 * @param mu_eff effective zenith angle
 * @param g asymmetry parameter
 * @param ssa single scattering albedo
 * @param transmission_cloud transmission cloudy part
 * @param transmission_layer total transmission
 * @param x1 clear optical depth
 * @param x2 cloudy optical depth
 * @param tol accuracy
 * 
 * @return effective tau
 */
float zbrent_taueff(float mu_eff, float g, float ssa,
                    float transmission_cloud,  float transmission_layer,
                    float x1, float x2, float tol)
{
  int iter=0;
  float a=x1, b=x2, c=x2, d=0, e=0, min1=0, min2=0;
  float fa=0, fb=0;
  float fc=0, p=0, q=0, r=0, s=0, tol1=0, xm=0;
  int itmax=100;
  float eps=3.0e-8;
  float dummy=0.0;
  float pref=0.0;
 

  F77_FCN (swde)(&g, &pref, &mu_eff, &x1, &ssa, &dummy, &dummy, &transmission_cloud, &dummy);
  fa= transmission_cloud-transmission_layer;
  
  dummy =0.0;
  pref=0.0; 
  F77_FCN (swde)(&g, &pref, &mu_eff, &x2, &ssa, &dummy, &dummy, &transmission_cloud, &dummy);
  fb= transmission_cloud-transmission_layer;
  
  if ((fa > 0.0 && fb > 0.0) || (fa < 0.0 && fb < 0.0))
    fprintf(stderr,"Root must be bracketed in zbrent");
  fc=fb;
  for (iter=1;iter<=itmax;iter++) {
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
      c=a;
      fc=fa;
      e=d=b-a;
    }
    if (fabs(fc) < fabs(fb)) {
      a=b;
      b=c;
      c=a;
      fa=fb;
      fb=fc;
      fc=fa;
    }
    tol1=2.0*eps*fabs(b)+0.5*tol;
    xm=0.5*(c-b);
    if (fabs(xm) <= tol1 || fb == 0.0) return b;
    if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
      s=fb/fa;
      if (a == c) {
	p=2.0*xm*s;
	q=1.0-s;
      } else {
	q=fa/fc;
	r=fb/fc;
	p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
	q=(q-1.0)*(r-1.0)*(s-1.0);
      }
      if (p > 0.0) q = -q;
      p=fabs(p);
      min1=3.0*xm*q-fabs(tol1*q);
      min2=fabs(e*q);
      if (2.0*p < (min1 < min2 ? min1 : min2)) {
	e=d;
	d=p/q;
      } else {
	d=xm;
	e=d;
      }
    } else {
      d=xm;
      e=d;
    }
    a=b;
    fa=fb;
    if (fabs(d) > tol1)
      b += d;
    else
      b += SIGN(tol1,xm);

    dummy =0.0;
    pref = 0.0;
    F77_FCN (swde)(&g, &dummy, &mu_eff, &b, &ssa, &dummy, &dummy, &transmission_cloud, &dummy);
    fb= transmission_cloud-transmission_layer;
    
  }
  fprintf(stderr,"Maximum number of iterations exceeded in zbrent");
  return 0.0;
}

#undef NRANSI

